@@ -1,5 +1,5 @@
 /* Hiearchic listboxes browser dialog commons */
-/* $Id: hierbox.c,v 1.143 2004/01/05 19:44:50 jonas Exp $ */
+/* $Id: hierbox.c,v 1.201.2.6 2005/04/06 09:11:18 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -12,13 +12,16 @@
 #include "bfu/button.h"
 #include "bfu/dialog.h"
 #include "bfu/hierbox.h"
+#include "bfu/inpfield.h"
 #include "bfu/listbox.h"
 #include "bfu/msgbox.h"
 #include "bfu/text.h"
-#include "bookmarks/bookmarks.h"
 #include "config/kbdbind.h"
+#include "dialogs/download.h"
 #include "intl/gettext/libintl.h"
+#include "protocol/uri.h"
 #include "sched/task.h"
+#include "terminal/screen.h"
 #include "terminal/tab.h"
 #include "terminal/terminal.h"
 
@@ -36,72 +39,40 @@ update_hierbox_browser(struct hierbox_browser *browser)
 	struct hierbox_dialog_list_item *item;
 
 	foreach (item, browser->dialogs) {
-		struct widget_data *widget_data =
-			item->dlg_data->widgets_data;
-
-		display_dlg_item(item->dlg_data, widget_data, 1);
+		redraw_from_window(item->dlg_data->win->next);
 	}
 }
 
 
+/* Common backend for listbox adding */
 struct listbox_item *
-do_add_listbox_item(struct listbox_item *root, unsigned char *text, void *data,
-		    int no_empty, enum listbox_item_type type)
+add_listbox_item(struct hierbox_browser *browser, struct listbox_item *root,
+		 enum listbox_item_type type, void *data, int add_position)
 {
 	struct listbox_item *item;
 
-	if (type == BI_FOLDER) {
-		/* Check if we have the folder already. Could be optional if
-		 * we are gonna use this for bookmarks and stuff. --jonas */
-		foreach (item, root->child)
-			if (!strcmp(item->text, text))
-				return item;
+	if (!root) {
+		assertm(browser, "Nowhere to add new list box item");
+		root = &browser->root;
 	}
 
-	item = mem_calloc(1, sizeof(struct listbox_item));
+	item = mem_calloc(1, sizeof(*item));
 	if (!item) return NULL;
 
 	init_list(item->child);
 	item->visible = 1;
 
-	item->text = text;
 	item->udata = data;
 	item->type = type;
 	item->depth = root->depth + 1;
-	item->free_empty_folder = no_empty;
-	if (item->depth > 0) item->root = root;
 
-	/* TODO: Sort? */
-	add_to_list(root->child, item);
+	/* TODO: Possibility to sort by making add_position into a flag */
+	if (add_position < 0)
+		add_to_list_end(root->child, item);
+	else
+		add_to_list(root->child, item);
 
-	return item;
-}
-
-struct listbox_item *
-add_listbox_items(struct hierbox_browser *browser, void *data, int noempty,
-		  unsigned char *text, ...)
-{
-	struct listbox_item *item, *root = &browser->root;
-	unsigned char *label;
-	va_list ap;
-
-	va_start(ap, text);
-
-	while ((label = va_arg(ap, unsigned char *))) {
-		root = do_add_listbox_item(root, text, NULL, noempty, BI_FOLDER);
-		if (!root) {
-			va_end(ap);
-			return NULL;
-		}
-
-		text = label;
-	}
-
-	va_end(ap);
-
-	item = do_add_listbox_item(root, text, data, 0, BI_LEAF);
-
-	update_hierbox_browser(browser);
+	if (browser) update_hierbox_browser(browser);
 
 	return item;
 }
@@ -127,7 +98,7 @@ done_listbox_item(struct hierbox_browser *browser, struct listbox_item *box_item
 {
 	struct listbox_data *box_data;
 
-	assert(box_item);
+	assert(box_item && list_empty(box_item->child));
 
 	/* If we are removing the top or the selected box we have to figure out
 	 * a replacement. */
@@ -143,12 +114,6 @@ done_listbox_item(struct hierbox_browser *browser, struct listbox_item *box_item
 	/* The option dialog needs this test */
 	if (box_item->next) del_from_list(box_item);
 
-	if (box_item->root
-	    && box_item->root->free_empty_folder
-	    && box_item->root->type == BI_FOLDER
-	    && list_empty(box_item->root->child))
-		done_listbox_item(browser, box_item->root);
-
 	mem_free(box_item);
 	update_hierbox_browser(browser);
 }
@@ -179,140 +144,166 @@ test_search(struct listbox_item *item, void *data_, int *offset)
 	return 0;
 }
 
+static t_handler_event_status
+hierbox_ev_kbd(struct dialog_data *dlg_data)
+{
+	struct hierbox_browser *browser = dlg_data->dlg->udata2;
+	struct widget_data *widget_data = dlg_data->widgets_data;
+	struct widget *widget = widget_data->widget;
+	struct listbox_data *box;
+	struct listbox_item *selected;
+	enum menu_action action;
+	struct term_event *ev = dlg_data->term_event;
+
+	/* Check if listbox has something to say to this */
+	if (widget->ops->kbd
+	    && widget->ops->kbd(dlg_data, widget_data)
+	       == EVENT_PROCESSED)
+		return EVENT_PROCESSED;
+
+	box = get_dlg_listbox_data(dlg_data);
+	selected = box->sel;
+	action = kbd_action(KEYMAP_MENU, ev, NULL);
+
+	if (action == ACT_MENU_SELECT) {
+		if (!selected) return EVENT_PROCESSED;
+		if (selected->type != BI_FOLDER)
+			return EVENT_NOT_PROCESSED;
+		selected->expanded = !selected->expanded;
+
+	} else if (action == ACT_MENU_UNEXPAND) {
+		/* Recursively unexpand all folders */
+		if (!selected) return EVENT_PROCESSED;
+
+		/* Special trick: if the folder is already
+		 * folded, jump at the parent folder, so the
+		 * next time when user presses the key, the
+		 * whole parent folder will be closed. */
+		if (list_empty(selected->child)
+		    || !selected->expanded) {
+			struct listbox_item *root = box->ops->get_root(selected);
+
+			if (root) {
+				struct listbox_context ctx;
+
+				memset(&ctx, 0, sizeof(ctx));
+				ctx.box = box;
+				ctx.offset = 1;
+
+				traverse_listbox_items_list(
+						root, box, 0, 1,
+						test_search, &ctx);
+				listbox_sel_move(dlg_data->widgets_data,
+					         ctx.offset);
+			}
+
+		} else if (selected->type == BI_FOLDER) {
+			recursively_set_expanded(selected, 0);
+		}
+
+	} else if (action == ACT_MENU_EXPAND) {
+		/* Recursively expand all folders */
+
+		if (!selected || box->sel->type != BI_FOLDER)
+			return EVENT_PROCESSED;
+
+		recursively_set_expanded(box->sel, 1);
+
+	} else if (action == ACT_MENU_SEARCH) {
+		if (!box->ops->match)
+			return EVENT_NOT_PROCESSED;
+
+		push_hierbox_search_button(dlg_data, NULL);
+		return EVENT_PROCESSED;
+
+	} else {
+		return EVENT_NOT_PROCESSED;
+
+	}
+
+	if (browser->expansion_callback)
+		browser->expansion_callback();
+
+	display_widget(dlg_data, dlg_data->widgets_data);
+
+	return EVENT_PROCESSED;
+}
+
+static t_handler_event_status
+hierbox_ev_init(struct dialog_data *dlg_data)
+{
+	struct hierbox_browser *browser = dlg_data->dlg->udata2;
+	struct hierbox_dialog_list_item *item;
+	struct listbox_item *litem;
+
+	/* If we fail here it only means automatic updating
+	 * will not be possible so no need to panic. */
+	item = mem_alloc(sizeof(*item));
+	if (item) {
+		item->dlg_data = dlg_data;
+		add_to_list(browser->dialogs, item);
+	}
+
+	foreach (litem, browser->root.child) {
+		litem->visible = 1;
+	}
+
+	return EVENT_NOT_PROCESSED;	/* FIXME: is this correct ? --Zas */
+}
+
+static t_handler_event_status
+hierbox_ev_abort(struct dialog_data *dlg_data)
+{
+	struct listbox_data *box = get_dlg_listbox_data(dlg_data);
+	struct hierbox_browser *browser = dlg_data->dlg->udata2;
+	struct hierbox_dialog_list_item *item;
+
+	/* Save state and delete the box structure */
+	/* FIXME: It's workaround for bug 397. Real fix is needed. */
+	if (browser != &download_browser)
+		copy_struct(&browser->box_data, box);
+	del_from_list(box);
+
+	/* Delete the dialog list entry */
+	foreach (item, browser->dialogs) {
+		if (item->dlg_data == dlg_data) {
+			del_from_list(item);
+			mem_free(item);
+			break;
+		}
+	}
+
+	return EVENT_NOT_PROCESSED; /* FIXME: is this correct ? --Zas */
+}
+
+
 /* We install own dialog event handler, so that we can give the listbox widget
  * an early chance to catch the event. Basically, the listbox widget is itself
  * unselectable, instead one of the buttons below is always active. So, we
  * always first let the listbox catch the keypress and handle it, and if it
  * doesn't care, we pass it on to the button. */
-static int
-hierbox_dialog_event_handler(struct dialog_data *dlg_data, struct term_event *ev)
+static t_handler_event_status
+hierbox_dialog_event_handler(struct dialog_data *dlg_data)
 {
+	struct term_event *ev = dlg_data->term_event;
+
 	switch (ev->ev) {
-		case EV_KBD:
-		{
-			struct widget_data *widget_data =
-						dlg_data->widgets_data;
-			struct widget *widget = widget_data->widget;
-			struct listbox_data *box;
-			struct listbox_item *selected;
-			enum keyact action;
+		case EVENT_KBD:
+			return hierbox_ev_kbd(dlg_data);
 
-			/* Check if listbox has something to say to this */
-                        if (widget->ops->kbd
-			    && widget->ops->kbd(widget_data, dlg_data, ev)
-			       == EVENT_PROCESSED)
-				return EVENT_PROCESSED;
+		case EVENT_INIT:
+			return hierbox_ev_init(dlg_data);
 
-			box = get_dlg_listbox_data(dlg_data);
-			selected = box->sel;
-			action = kbd_action(KM_MENU, ev, NULL);
-
-			if (action == ACT_SELECT) {
-				if (!selected) return EVENT_PROCESSED;
-				if (selected->type != BI_FOLDER)
-					return EVENT_NOT_PROCESSED;
-				selected->expanded = !selected->expanded;
-				break;
-			}
-
-			/* Recursively unexpand all folders */
-			if (action == ACT_UNEXPAND) {
-				if (!selected) return EVENT_PROCESSED;
-
-				/* Special trick: if the folder is already
-				 * folded, jump at the parent folder, so the
-				 * next time when user presses the key, the
-				 * whole parent folder will be closed. */
-				if (list_empty(selected->child)
-				    || !selected->expanded) {
-					struct listbox_context ctx;
-
-					if (!selected->root) break;
-
-					memset(&ctx, 0, sizeof(struct listbox_context));
-					ctx.box = box;
-					ctx.offset = 1;
-
-					traverse_listbox_items_list(
-							selected->root, box, 0, 1,
-							test_search, &ctx);
-					box_sel_move(dlg_data->widgets_data,
-						     ctx.offset);
-
-				} else if (selected->type == BI_FOLDER) {
-					recursively_set_expanded(selected, 0);
-				}
-
-				break;
-			}
-
-			/* Recursively expand all folders */
-			if (action == ACT_EXPAND) {
-				if (!selected || box->sel->type != BI_FOLDER)
-					return EVENT_PROCESSED;
-
-				recursively_set_expanded(box->sel, 1);
-				break;
-			}
-
-			return EVENT_NOT_PROCESSED;
-		}
-
-		case EV_INIT:
-		{
-			struct hierbox_browser *browser = dlg_data->dlg->udata2;
-			struct hierbox_dialog_list_item *item;
-			struct listbox_item *litem;
-
-			/* If we fail here it only means automatic updating
-			 * will not be possible so no need to panic. */
-			item = mem_alloc(sizeof(struct hierbox_dialog_list_item));
-			if (item) {
-				item->dlg_data = dlg_data;
-				add_to_list(browser->dialogs, item);
-			}
-
-			foreach (litem, browser->root.child) {
-				litem->visible = 1;
-			}
-		}
-		case EV_RESIZE:
-		case EV_REDRAW:
-		case EV_MOUSE:
+		case EVENT_RESIZE:
+		case EVENT_REDRAW:
+		case EVENT_MOUSE:
 			return EVENT_NOT_PROCESSED;
 
-		case EV_ABORT:
-		{
-			struct listbox_data *box = get_dlg_listbox_data(dlg_data);
-			struct hierbox_browser *browser = dlg_data->dlg->udata2;
-			struct hierbox_dialog_list_item *item;
-
-			/* Save state and delete the box structure */
-			memcpy(&browser->box_data, box, sizeof(struct listbox_data));
-			del_from_list(box);
-
-			/* Delete the dialog list entry */
-			foreach (item, browser->dialogs) {
-				if (item->dlg_data == dlg_data) {
-					del_from_list(item);
-					mem_free(item);
-					break;
-				}
-			}
-			return EVENT_NOT_PROCESSED;
-		}
+		case EVENT_ABORT:
+			return hierbox_ev_abort(dlg_data);
 	}
 
-#ifdef CONFIG_BOOKMARKS
-	/* FIXME - move from here to bookmarks/dialogs.c! */
-	/* We should probably call provided callback function, notifying the
-	 * user of expansion change. --pasky */
-	bookmarks_dirty = 1;
-#endif
-
-	display_dlg_item(dlg_data, dlg_data->widgets_data, 1);
-
-	return EVENT_PROCESSED;
+	return EVENT_NOT_PROCESSED;
 }
 
 
@@ -323,11 +314,11 @@ hierbox_browser(struct hierbox_browser *browser, struct session *ses)
 	struct listbox_data *listbox_data;
 	struct dialog *dlg;
 	int button = browser->buttons_size + 2;
-	int anonymous = get_opt_bool_tree(cmdline_options, "anonymous");
+	int anonymous = get_cmd_opt_bool("anonymous");
 
 	assert(ses);
 
-	dlg = calloc_dialog(button, sizeof(struct listbox_data));
+	dlg = calloc_dialog(button, sizeof(*listbox_data));
 	if (!dlg) return NULL;
 
 	listbox_data = (struct listbox_data *) get_dialog_offset(dlg, button);
@@ -343,19 +334,18 @@ hierbox_browser(struct hierbox_browser *browser, struct session *ses)
 	add_dlg_listbox(dlg, 12, listbox_data);
 
 	for (button = 0; button < browser->buttons_size; button++) {
-		hierbox_button_handler handler = browser->buttons[button].handler;
-		unsigned char *label = browser->buttons[button].label;
+		struct hierbox_browser_button *but = &browser->buttons[button];
 
 		/* Skip buttons that should not be displayed in anonymous mode */
-		if (!browser->buttons[button].anonymous && anonymous) {
+		if (anonymous && !but->anonymous) {
 			anonymous++;
 			continue;
 		}
 
-		add_dlg_button(dlg, B_ENTER, handler, _(label, term), NULL);
+		add_dlg_button(dlg, _(but->label, term), B_ENTER, but->handler, NULL);
 	}
 
-	add_dlg_button(dlg, B_ESC, cancel_dialog, _("Close", term), NULL);
+	add_dlg_button(dlg, _("Close", term), B_ESC, cancel_dialog, NULL);
 
 	/* @anonymous was initially 1 if we are running in anonymous mode so we
 	 * have to subtract one. */
@@ -396,15 +386,15 @@ scan_for_used(struct listbox_item *item, void *info_, int *offset)
 
 static struct listbox_context *
 init_listbox_context(struct listbox_data *box, struct terminal *term,
-			 struct listbox_item *item,
-			 int (*scanner)(struct listbox_item *, void *, int *))
+		     struct listbox_item *item,
+		     int (*scanner)(struct listbox_item *, void *, int *))
 {
 	struct listbox_context *context;
 
-	context = mem_calloc(1, sizeof(struct listbox_context));
+	context = mem_calloc(1, sizeof(*context));
 	if (!context) return NULL;
 
-	context->item = item;;
+	context->item = item;
 	context->term = term;
 	context->box = box;
 
@@ -431,7 +421,7 @@ done_listbox_context(void *context_)
 
 /* Info action */
 
-int
+t_handler_event_status
 push_hierbox_info_button(struct dialog_data *dlg_data, struct widget_data *button)
 {
 	struct listbox_data *box = get_dlg_listbox_data(dlg_data);
@@ -439,61 +429,56 @@ push_hierbox_info_button(struct dialog_data *dlg_data, struct widget_data *butto
 	struct listbox_context *context;
 	unsigned char *msg;
 
-	if (!box->sel) return 0;
-
-	if (box->sel->type == BI_FOLDER) {
-		msg_box(term, NULL, 0,
-		N_("Info"), AL_CENTER,
-		_("Press space to expand this folder.", term),
-		NULL, 1,
-		N_("OK"), NULL, B_ESC | B_ENTER);
-
-		return 0;
-	}
+	if (!box->sel) return EVENT_PROCESSED;
 
 	assert(box->ops);
 
 	context = init_listbox_context(box, term, box->sel, NULL);
-	if (!context) return 0;
+	if (!context) return EVENT_PROCESSED;
 
-	msg = box->ops->get_info(context->item, term, LISTBOX_ALL);
+	msg = box->ops->get_info(context->item, term);
 	if (!msg) {
 		mem_free(context);
-		return 0;
+		if (box->sel->type == BI_FOLDER) {
+			info_box(term, 0, N_("Info"), ALIGN_CENTER,
+				 N_("Press space to expand this folder."));
+		}
+		return EVENT_PROCESSED;
 	}
 
 	box->ops->lock(context->item);
 
 	msg_box(term, getml(context, NULL), MSGBOX_FREE_TEXT /* | MSGBOX_SCROLLABLE */,
-		N_("Info"), AL_LEFT,
+		N_("Info"), ALIGN_LEFT,
 		msg,
 		context, 1,
-		N_("OK"), done_listbox_context, B_ESC | B_ENTER);
+		N_("~OK"), done_listbox_context, B_ESC | B_ENTER);
 
-	return 0;
+	return EVENT_PROCESSED;
 }
 
 
 /* Goto action */
 
 static void
-recursively_goto_listbox(struct session *ses, struct terminal *term,
-			 struct listbox_item *root, struct listbox_data *box)
+recursively_goto_listbox(struct session *ses, struct listbox_item *root,
+			 struct listbox_data *box)
 {
 	struct listbox_item *item;
-	unsigned char *uri;
 
 	foreach (item, root->child) {
+		struct uri *uri;
+
 		if (item->type == BI_FOLDER) {
-			recursively_goto_listbox(ses, term, item, box);
+			recursively_goto_listbox(ses, item, box);
 			continue;
 		}
 
-		uri = box->ops->get_info(item, term, LISTBOX_URI);
+		uri = box->ops->get_uri(item);
 		if (!uri) continue;
 
-		open_url_in_new_tab(ses, uri, 1);
-		mem_free(uri);
+		open_uri_in_new_tab(ses, uri, 1, 0);
+		done_uri(uri);
 	}
 }
 
@@ -505,24 +490,24 @@ goto_marked(struct listbox_item *item, void *data_, int *offset)
 	if (item->marked) {
 		struct session *ses = context->dlg_data->dlg->udata;
 		struct listbox_data *box = context->box;
-		unsigned char *uri;
+		struct uri *uri;
 
 		if (item->type == BI_FOLDER) {
-			recursively_goto_listbox(ses, ses->tab->term, item, box);
+			recursively_goto_listbox(ses, item, box);
 			return 0;
 		}
 
-		uri = box->ops->get_info(item, ses->tab->term, LISTBOX_URI);
+		uri = box->ops->get_uri(item);
 		if (!uri) return 0;
 
-		open_url_in_new_tab(ses, uri, 1);
-		mem_free(uri);
+		open_uri_in_new_tab(ses, uri, 1, 0);
+		done_uri(uri);
 	}
 
 	return 0;
 }
 
-int
+t_handler_event_status
 push_hierbox_goto_button(struct dialog_data *dlg_data,
 			 struct widget_data *button)
 {
@@ -530,13 +515,12 @@ push_hierbox_goto_button(struct dialog_data *dlg_data,
 	struct session *ses = dlg_data->dlg->udata;
 	struct terminal *term = dlg_data->win->term;
 	struct listbox_context *context;
-	unsigned char *uri;
 
 	/* Do nothing with a folder */
-	if (!box->sel) return 0;
+	if (!box->sel) return EVENT_PROCESSED;
 
 	context = init_listbox_context(box, term, box->sel, scan_for_marks);
-	if (!context) return 0;
+	if (!context) return EVENT_PROCESSED;
 
 	if (!context->item) {
 		context->dlg_data = dlg_data;
@@ -545,22 +529,22 @@ push_hierbox_goto_button(struct dialog_data *dlg_data,
 					    goto_marked, context);
 
 	} else if (box->sel->type == BI_FOLDER) {
-		recursively_goto_listbox(ses, term, box->sel, box);
+		recursively_goto_listbox(ses, box->sel, box);
 
 	} else {
-		/* Follow the bookmark */
-		uri = box->ops->get_info(box->sel, term, LISTBOX_URI);
-		if (!uri) return 0;
+		struct uri *uri = box->ops->get_uri(box->sel);
 
-		goto_url(ses, uri);
-		mem_free(uri);
+		if (uri) {
+			goto_uri(ses, uri);
+			done_uri(uri);
+		}
 	}
 
 	mem_free(context);
 
 	/* Close the dialog */
 	delete_window(dlg_data->win);
-	return 0;
+	return EVENT_PROCESSED;
 }
 
 
@@ -588,13 +572,55 @@ print_delete_error(struct listbox_item *item, struct terminal *term,
 		   struct listbox_ops *ops, enum delete_error err)
 {
 	struct string msg;
-	unsigned char *text = delete_messages[(item->type == BI_FOLDER)][err];
+	unsigned char *errmsg;
+	unsigned char *text;
 
-	if (!init_string(&msg)) return;
+	switch (err) {
+	case DELETE_IMPOSSIBLE:
+		if (item->type == BI_FOLDER) {
+			if (ops->messages && ops->messages->cant_delete_folder)
+				errmsg = ops->messages->cant_delete_folder;
+			else
+				errmsg = delete_messages[1][DELETE_IMPOSSIBLE];
+		} else {
+			if (ops->messages && ops->messages->cant_delete_item)
+				errmsg = ops->messages->cant_delete_item;
+			else
+				errmsg = delete_messages[0][DELETE_IMPOSSIBLE];
+		}
+		break;
+
+	case DELETE_LOCKED:
+		if (item->type == BI_FOLDER) {
+			if (ops->messages && ops->messages->cant_delete_used_folder)
+				errmsg = ops->messages->cant_delete_used_folder;
+			else
+				errmsg = delete_messages[1][DELETE_LOCKED];
+		} else {
+			if (ops->messages && ops->messages->cant_delete_used_item)
+				errmsg = ops->messages->cant_delete_used_item;
+			else
+				errmsg = delete_messages[0][DELETE_LOCKED];
+		}
+		break;
+
+	default:
+		INTERNAL("Bad delete error code (%d)!", err);
+		return;
+	}
+
+	text = ops->get_text(item, term);
+
+	if (!text || !init_string(&msg)) {
+		mem_free_if(text);
+		return;
+	}
+
+	add_format_to_string(&msg, _(errmsg, term), text);
+	mem_free(text);
 
-	add_format_to_string(&msg, _(text, term), item->text);
 	if (item->type == BI_LEAF) {
-		unsigned char *info = ops->get_info(item, term, LISTBOX_ALL);
+		unsigned char *info = ops->get_info(item, term);
 
 		if (info) {
 			add_format_to_string(&msg, "\n\n%s", info);
@@ -602,41 +628,29 @@ print_delete_error(struct listbox_item *item, struct terminal *term,
 		}
 	}
 
-	msg_box(term, NULL, MSGBOX_FREE_TEXT,
-		N_("Delete error"), AL_LEFT,
-		msg.source,
-		NULL, 1,
-		N_("OK"), NULL, B_ENTER | B_ESC);
+	info_box(term, MSGBOX_FREE_TEXT, N_("Delete error"), ALIGN_LEFT,
+		 msg.source);
 }
 
 static void
 do_delete_item(struct listbox_item *item, struct listbox_context *info,
-	       int last, int delete)
+	       int last)
 {
 	struct listbox_ops *ops = info->box->ops;
 
 	assert(item);
 
-	if ((!delete && !ops->can_delete(item)) || ops->is_used(item)) {
-		print_delete_error(item, info->term, ops,
-				   delete ? DELETE_LOCKED : DELETE_IMPOSSIBLE);
+	if (!ops->can_delete(item)) {
+		print_delete_error(item, info->term, ops, DELETE_IMPOSSIBLE);
 		return;
 	}
 
-	if (item->type == BI_FOLDER) {
-		struct listbox_item *child = item->child.next;
-
-		while (child != (void *) &item->child) {
-			child = child->next;
-			/* Propagate the ``delete'' property down to children
-			 * since if a parent can be deleted the child should
-			 * just delete itself too. */
-			do_delete_item(child->prev, info, 0, 1);
-		}
+	if (ops->is_used(item)) {
+		print_delete_error(item, info->term, ops, DELETE_LOCKED);
+		return;
 	}
 
-	if (list_empty(item->child))
-		ops->delete(item, last);
+	ops->delete(item, last);
 }
 
 static int
@@ -649,7 +663,7 @@ delete_marked(struct listbox_item *item, void *data_, int *offset)
 		if (!context->item) {
 			context->item = item;
 		} else {
-			do_delete_item(item, context, 0, 0);
+			do_delete_item(item, context, 0);
 		}
 
 		return 1;
@@ -662,6 +676,8 @@ static void
 push_ok_delete_button(void *context_)
 {
 	struct listbox_context *context = context_;
+	struct listbox_item *root;
+	int last = 0;
 
 	if (context->item) {
 		context->box->ops->unlock(context->item);
@@ -672,71 +688,118 @@ push_ok_delete_button(void *context_)
 		if (!context->item) return;
 	}
 
+	root = context->box->ops->get_root(context->item);
+	if (root) {
+		last = context->item == root->child.prev;
+	}
+
 	/* Delete the last one (traversal should save one to delete) */
-	do_delete_item(context->item, context, 1, 0);
+	do_delete_item(context->item, context, 1);
+
+	/* If removing the last item in a folder move focus to previous item in
+	 * the folder or the root. */
+	if (last)
+		listbox_sel_move(context->widget_data, -1);
 }
 
-int
+t_handler_event_status
 push_hierbox_delete_button(struct dialog_data *dlg_data,
 			   struct widget_data *button)
 {
 	struct terminal *term = dlg_data->win->term;
 	struct listbox_data *box = get_dlg_listbox_data(dlg_data);
 	struct listbox_context *context;
+	unsigned char *text;
 	enum delete_error delete;
 
-	if (!box->sel) return 0;
+	if (!box->sel) return EVENT_PROCESSED;
 
 	assert(box->ops && box->ops->can_delete && box->ops->delete);
 
 	context = init_listbox_context(box, term, box->sel, scan_for_marks);
-	if (!context) return 0;
+	if (!context) return EVENT_PROCESSED;
+
+	context->widget_data = dlg_data->widgets_data;
 
 	if (!context->item) {
+		unsigned char *title = N_("Delete marked items");
+		unsigned char *message = N_("Delete marked items?");
+
+		if (box->ops->messages) {
+			if (box->ops->messages->delete_marked_items)
+				message = box->ops->messages->delete_marked_items;
+			if (box->ops->messages->delete_marked_items_title)
+				title = box->ops->messages->delete_marked_items_title;
+		}
+
 		msg_box(term, getml(context, NULL), 0,
-			N_("Delete marked items"), AL_CENTER,
-			N_("Delete marked items?"),
+			title, ALIGN_CENTER,
+			message,
 			context, 2,
-			N_("Yes"), push_ok_delete_button, B_ENTER,
-			N_("No"), done_listbox_context, B_ESC);
-		return 0;
+			N_("~Yes"), push_ok_delete_button, B_ENTER,
+			N_("~No"), done_listbox_context, B_ESC);
+		return EVENT_PROCESSED;
 	}
 
 	delete = box->ops->can_delete(context->item)
-		? DELETE_LOCKED : DELETE_IMPOSSIBLE;
+		 ? DELETE_LOCKED : DELETE_IMPOSSIBLE;
 
 	if (delete == DELETE_IMPOSSIBLE || box->ops->is_used(context->item)) {
 		print_delete_error(context->item, term, box->ops, delete);
 		mem_free(context);
-		return 0;
+		return EVENT_PROCESSED;
+	}
+
+	text = box->ops->get_text(context->item, term);
+	if (!text) {
+		mem_free(context);
+		return EVENT_PROCESSED;
 	}
 
 	if (context->item->type == BI_FOLDER) {
+		unsigned char *title = N_("Delete folder");
+		unsigned char *message = N_("Delete the folder \"%s\" and its content?");
+
+		if (box->ops->messages) {
+			if (box->ops->messages->delete_folder)
+				message = box->ops->messages->delete_folder;
+			if (box->ops->messages->delete_folder_title)
+				title = box->ops->messages->delete_folder_title;
+		}
+
 		box->ops->lock(context->item);
 		msg_box(term, getml(context, NULL), MSGBOX_FREE_TEXT,
-			N_("Delete folder"), AL_CENTER,
-			msg_text(term, N_("Delete the folder \"%s\" and its content?"),
-				 context->item->text),
+			title, ALIGN_CENTER,
+			msg_text(term, message, text),
 			context, 2,
-			N_("Yes"), push_ok_delete_button, B_ENTER,
-			N_("No"), done_listbox_context, B_ESC);
+			N_("~Yes"), push_ok_delete_button, B_ENTER,
+			N_("~No"), done_listbox_context, B_ESC);
 	} else {
+		unsigned char *title = N_("Delete item");
+		unsigned char *message = N_("Delete \"%s\"?\n\n%s");
 		unsigned char *msg;
 
-		msg = box->ops->get_info(context->item, term, LISTBOX_ALL);
+		if (box->ops->messages) {
+			if (box->ops->messages->delete_item)
+				message = box->ops->messages->delete_item;
+			if (box->ops->messages->delete_item_title)
+				title = box->ops->messages->delete_item_title;
+		}
+
+		msg = box->ops->get_info(context->item, term);
 		box->ops->lock(context->item);
 
-		msg_box(term, getml(context, msg, NULL), MSGBOX_FREE_TEXT,
-			N_("Delete item"), AL_LEFT,
-			msg_text(term, N_("Delete \"%s\"?\n\n"
-				"%s"),
-				context->item->text, empty_string_or_(msg)),
+		msg_box(term, getml(context, NULL), MSGBOX_FREE_TEXT,
+			title, ALIGN_LEFT,
+			msg_text(term, message, text, empty_string_or_(msg)),
 			context, 2,
-			N_("Yes"), push_ok_delete_button, B_ENTER,
-			N_("No"), done_listbox_context, B_ESC);
+			N_("~Yes"), push_ok_delete_button, B_ENTER,
+			N_("~No"), done_listbox_context, B_ESC);
+		mem_free_if(msg);
 	}
+	mem_free(text);
 
-	return 0;
+	return EVENT_PROCESSED;
 }
 
 
@@ -749,7 +812,7 @@ delete_unused(struct listbox_item *item, void *data_, int *offset)
 
 	if (context->box->ops->is_used(item)) return 0;
 
-	do_delete_item(item, context, 0, 0);
+	do_delete_item(item, context, 0);
 	return 1;
 }
 
@@ -763,20 +826,22 @@ do_clear_browser(void *context_)
 				    delete_unused, context);
 }
 
-int
+t_handler_event_status
 push_hierbox_clear_button(struct dialog_data *dlg_data,
 			  struct widget_data *button)
 {
 	struct listbox_data *box = get_dlg_listbox_data(dlg_data);
 	struct terminal *term = dlg_data->win->term;
 	struct listbox_context *context;
+	unsigned char *title = N_("Clear all items");
+	unsigned char *message = N_("Do you really want to remove all items?");
 
-	if (!box->sel) return 0;
+	if (!box->sel) return EVENT_PROCESSED;
 
 	assert(box->ops);
 
 	context = init_listbox_context(box, term, NULL, scan_for_used);
-	if (!context) return 0;
+	if (!context) return EVENT_PROCESSED;
 
 	if (context->item) {
 		/* FIXME: If the clear button should be used for browsers where
@@ -785,15 +850,122 @@ push_hierbox_clear_button(struct dialog_data *dlg_data,
 		 * error types. */
 		print_delete_error(context->item, term, box->ops, DELETE_LOCKED);
 		mem_free(context);
-		return 0;
+		return EVENT_PROCESSED;
+	}
+
+	if (box->ops->messages) {
+		if (box->ops->messages->clear_all_items)
+			message = box->ops->messages->clear_all_items;
+		if (box->ops->messages->clear_all_items_title)
+			title = box->ops->messages->clear_all_items_title;
 	}
 
 	msg_box(term, getml(context, NULL), 0,
-		N_("Clear all items"), AL_CENTER,
-		N_("Do you really want to remove all items?"),
+		title, ALIGN_CENTER,
+		message,
 		context, 2,
-		N_("Yes"), do_clear_browser, B_ENTER,
-		N_("No"), NULL, B_ESC);
+		N_("~Yes"), do_clear_browser, B_ENTER,
+		N_("~No"), NULL, B_ESC);
+
+	return EVENT_PROCESSED;
+}
+
+
+/* Search action */
+
+static int
+scan_for_matches(struct listbox_item *item, void *info_, int *offset)
+{
+	struct listbox_context *context = info_;
+	unsigned char *text = (unsigned char *) context->widget_data;
+
+	if (!*text) {
+		item->visible = 1;
+		return 0;
+	}
+
+	switch (context->box->ops->match(item, context->term, text)) {
+	case LISTBOX_MATCH_OK:
+		/* Mark that we have a match by setting the item to non-NULL */
+		context->item = item;
+		item->visible = 1;
+		break;
+
+	case LISTBOX_MATCH_NO:
+		item->visible = 0;
+		break;
+
+	case LISTBOX_MATCH_IMPOSSIBLE:
+		break;
+	}
 
 	return 0;
 }
+
+static int
+mark_visible(struct listbox_item *item, void *xxx, int *offset)
+{
+	item->visible = 1;
+	return 0;
+}
+
+
+static void
+search_hierbox_browser(void *data, unsigned char *text)
+{
+	struct dialog_data *dlg_data = data;
+	struct listbox_data *box = get_dlg_listbox_data(dlg_data);
+	struct terminal *term = dlg_data->win->term;
+	struct listbox_context *context;
+
+	context = init_listbox_context(box, term, NULL, NULL);
+	if (!context) return;
+
+	/* Eeew :/ */
+	context->widget_data = (void *) text;
+
+	traverse_listbox_items_list(box->items->next, box, 0, 0,
+				    scan_for_matches, context);
+
+	if (!context->item && *text) {
+		switch (get_opt_int("document.browse.search.show_not_found")) {
+		case 2:
+			info_box(term, MSGBOX_FREE_TEXT,
+				 N_("Search"), ALIGN_CENTER,
+				 msg_text(term,
+					  N_("Search string '%s' not found"),
+					  text));
+			break;
+
+		case 1:
+			beep_terminal(term);
+
+		default:
+			break;
+		}
+
+		traverse_listbox_items_list(box->items->next, box, 0, 0,
+					    mark_visible, NULL);
+	}
+
+	mem_free(context);
+}
+
+t_handler_event_status
+push_hierbox_search_button(struct dialog_data *dlg_data,
+			   struct widget_data *button)
+{
+	struct terminal *term = dlg_data->win->term;
+	struct listbox_data *box = get_dlg_listbox_data(dlg_data);
+
+	if (!box->sel) return EVENT_PROCESSED;
+
+	assert(box->ops->match);
+
+	input_dialog(term, NULL, N_("Search"), N_("Name"),
+		     dlg_data, NULL,
+		     MAX_STR_LEN, "", 0, 0, NULL,
+		     search_hierbox_browser, NULL);
+
+	return EVENT_PROCESSED;
+}
