@@ -1,4 +1,4 @@
-/* $Id: leds.h,v 1.16 2004/01/01 14:03:51 jonas Exp $ */
+/* $Id: leds.h,v 1.17.2.1 2005/01/29 01:39:15 jonas Exp $ */
 
 #ifndef EL__BFU_LEDS_H
 #define EL__BFU_LEDS_H
@@ -9,6 +9,7 @@
 #include "util/color.h"
 
 struct session;
+struct terminal;
 
 /* TODO: Variable count! */
 #define LEDS_COUNT	6
@@ -27,7 +28,7 @@ struct led {
 	unsigned char value;
 
 	/* Private data. */
-	int __used;
+	int used__;
 };
 
 /* Per-session led panel structure. */
@@ -43,6 +44,8 @@ void init_led_panel(struct led_panel *leds);
 
 void draw_leds(struct session *ses);
 
+void menu_leds_info(struct terminal *term, void *xxx, void *xxxx);
+
 struct led *register_led(struct session *ses, int number);
 void unregister_led(struct led *);
 
