@@ -1,5 +1,5 @@
 /* Listbox widget implementation. */
-/* $Id: listbox.c,v 1.136 2004/01/02 18:37:56 jonas Exp $ */
+/* $Id: listbox.c,v 1.183.4.5 2005/04/06 08:32:41 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -12,30 +12,47 @@
 #include "bfu/dialog.h"
 #include "bfu/hierbox.h"
 #include "bfu/listbox.h"
-#include "bfu/style.h"
 #include "config/kbdbind.h"
 #include "intl/gettext/libintl.h"
 #include "terminal/draw.h"
-#include "terminal/kbd.h"
+#include "terminal/mouse.h"
 #include "terminal/terminal.h"
 #include "util/color.h"
 #include "util/conv.h"
 #include "util/lists.h"
 
 
+#define VERTICAL_LISTBOX_MARGIN 3
+
+void
+add_dlg_listbox(struct dialog *dlg, int height, void *box_data)
+{
+	struct widget *widget;
+
+	widget = &dlg->widgets[dlg->number_of_widgets++];
+	widget->type = WIDGET_LISTBOX;
+	widget->info.listbox.height = height;
+	widget->data = box_data;
+}
+
+struct listbox_data *
+get_listbox_widget_data(struct widget_data *widget_data)
+{
+	assert(widget_data->widget->type == WIDGET_LISTBOX);
+	return ((struct listbox_data *) widget_data->widget->data);
+}
+
 /* Layout for generic boxes */
 void
-dlg_format_box(struct terminal *term, struct widget_data *widget_data,
-	       int x, int *y, int w, int max_height, int *rw,
-	       enum format_align align)
+dlg_format_listbox(struct terminal *term, struct widget_data *widget_data,
+	           int x, int *y, int w, int max_height, int *rw,
+	           enum format_align align)
 {
 	int min, optimal_h;
 
-	widget_data->x = x;
-	widget_data->y = *y;
-	widget_data->w = w;
+	set_box(&widget_data->box, x, *y, w, 1);
 
-	if (rw) int_bounds(rw, widget_data->w, w);
+	if (rw) int_bounds(rw, widget_data->box.width, w);
 
 	/* Height bussiness follows: */
 
@@ -43,21 +60,21 @@ dlg_format_box(struct terminal *term, struct widget_data *widget_data,
 	/* (*y) += widget_data->widget->gid; */
 
 	/* This is only weird heuristic, it could scale well I hope. */
-	optimal_h = max_height * 7 / 10 - 3;
+	optimal_h = max_height * 7 / 10 - VERTICAL_LISTBOX_MARGIN;
 	min = get_opt_int("ui.dialogs.listbox_min_height");
 
-	if (max_height - 3 < min) {
+	if (max_height - VERTICAL_LISTBOX_MARGIN < min) {
 		/* Big trouble: can't satisfy even the minimum :-(. */
-		widget_data->h = max_height - 3;
+		widget_data->box.height = max_height - VERTICAL_LISTBOX_MARGIN;
 	} else if (optimal_h < min) {
-		widget_data->h = min;
+		widget_data->box.height = min;
 	} else {
-		widget_data->h = optimal_h;
+		widget_data->box.height = optimal_h;
 	}
 
 	/* DBG("::%d(%d)::%d::%d::", max_y, term?1:2, widget_data->h, *y); */
 
-	(*y) += widget_data->h;
+	(*y) += widget_data->box.height;
 }
 
 
@@ -132,7 +149,7 @@ traverse_listbox_items_list(struct listbox_item *item, struct listbox_data *box,
 		 * and no dangling children. */
 #define	item_cache(item) \
 	do { \
-		croot = item->root; cprev = item->prev; cnext = item->next; \
+		croot = box->ops->get_root(item); cprev = item->prev; cnext = item->next; \
 	} while (0)
 		struct listbox_item *croot, *cprev, *cnext;
 
@@ -249,7 +266,7 @@ done_down:
 
 /* Takes care about listbox top moving. */
 static int
-box_sel_move_do(struct listbox_item *item, void *data_, int *offset)
+listbox_sel_move_do(struct listbox_item *item, void *data_, int *offset)
 {
 	struct listbox_context *data = data_;
 
@@ -258,7 +275,7 @@ box_sel_move_do(struct listbox_item *item, void *data_, int *offset)
 
 	if (data->dist > 0) {
 		if (data->box->sel_offset
-		    < data->widget_data->h - 1) {
+		    < data->widget_data->box.height - 1) {
 			data->box->sel_offset++;
 		} else {
 			data->box->top =
@@ -281,7 +298,7 @@ box_sel_move_do(struct listbox_item *item, void *data_, int *offset)
 /* Moves the selected item by [dist] items. If [dist] is out of the current
  * range, the selected item is moved to the extreme (ie, the top or bottom) */
 void
-box_sel_move(struct widget_data *widget_data, int dist)
+listbox_sel_move(struct widget_data *widget_data, int dist)
 {
 	struct listbox_data *box = get_listbox_widget_data(widget_data);
 
@@ -301,7 +318,7 @@ box_sel_move(struct widget_data *widget_data, int dist)
 	    != box->sel) {
 		struct listbox_context data;
 
-		memset(&data, 0, sizeof(struct listbox_context));
+		memset(&data, 0, sizeof(data));
 
 		data.box = box;
 		data.widget_data = widget_data;
@@ -310,7 +327,8 @@ box_sel_move(struct widget_data *widget_data, int dist)
 		/* XXX: This is ugly, yes; but we don't want to call the
 		 * callback if we won't move on at all. */
 		box->sel = traverse_listbox_items_list(box->sel, box, dist, 1,
-						       box_sel_move_do, &data);
+						       listbox_sel_move_do,
+						       &data);
 	}
 }
 
@@ -320,19 +338,12 @@ static int
 display_listbox_item(struct listbox_item *item, void *data_, int *offset)
 {
 	struct listbox_context *data = data_;
-	unsigned char *text = item->text;
 	unsigned char *stylename;
 	int len; /* Length of the current text field. */
 	struct color_pair *color;
 	int depth = item->depth + 1;
 	int d;
-	int y;
-
-	if (item->translated)
-		text = _(text, data->term);
-
-	len = strlen(text);
-	int_upper_bound(&len, int_max(0, data->widget_data->w - depth * 5));
+	int x, y;
 
 	stylename = (item == data->box->sel) ? "menu.selected"
 		  : ((item->marked)	     ? "menu.marked"
@@ -340,7 +351,7 @@ display_listbox_item(struct listbox_item *item, void *data_, int *offset)
 
 	color = get_bfu_color(data->term, stylename);
 
-	y = data->widget_data->y + data->offset;
+	y = data->widget_data->box.y + data->offset;
 	for (d = 0; d < depth - 1; d++) {
 		struct listbox_item *root = item;
 		struct listbox_item *child = item;
@@ -348,11 +359,11 @@ display_listbox_item(struct listbox_item *item, void *data_, int *offset)
 
 		for (i = depth - d; i; i--) {
 			child = root;
-			if (root) root = root->root;
+			if (root) root = data->box->ops->get_root(root);
 		}
 
 		/* XXX */
-		x = data->widget_data->x + d * 5;
+		x = data->widget_data->box.x + d * 5;
 		draw_text(data->term, x, y, "     ", 5, 0, color);
 
 		if (root ? root->child.prev == child
@@ -365,11 +376,16 @@ display_listbox_item(struct listbox_item *item, void *data_, int *offset)
 	if (depth) {
 		enum border_char str[5] =
 			{ 32, BORDER_SRTEE, BORDER_SHLINE, BORDER_SHLINE, 32 };
-		int i, x;
+		int i;
 
-		if (item->type == BI_LEAF) {
-			if (item->root) {
-				if (item == item->root->child.prev) {
+		switch (item->type) {
+		case BI_LEAF:
+		case BI_SEPARATOR:
+		{
+			struct listbox_item *root = data->box->ops->get_root(item);
+
+			if (root) {
+				if (item == root->child.prev) {
 					str[1] = BORDER_SDLCORNER;
 				}
 			} else {
@@ -381,36 +397,63 @@ display_listbox_item(struct listbox_item *item, void *data_, int *offset)
 					str[1] = BORDER_SDLCORNER;
 				}
 			}
-		} else {
+			break;
+		}
+		case BI_FOLDER:
 			str[0] = '[';
 			str[1] = (item->expanded) ? '-' : '+';
 			str[2] = ']';
+			break;
+		default:
+			INTERNAL("Unknown item type");
+			break;
 		}
 
 		if (item->marked) str[4] = '*';
 
-		x = data->widget_data->x + (depth - 1) * 5;
+		x = data->widget_data->box.x + (depth - 1) * 5;
 		for (i = 0; i < 5; i++) {
 			draw_border_char(data->term, x + i, y, str[i], color);
 		}
 	}
 
-	if (data->box->ops && data->box->ops->draw) {
-		int x = data->widget_data->x + depth * 5;
-		int width = data->widget_data->w - depth * 5;
+	x = data->widget_data->box.x + depth * 5;
+
+	if (item->type == BI_SEPARATOR) {
+		int i;
+		int width = data->widget_data->box.width - depth * 5;
+
+		for (i = 0; i < width; i++) {
+			draw_border_char(data->term, x + i, y, BORDER_SHLINE, color);
+		}
+
+	} else if (data->box->ops && data->box->ops->draw) {
+		int width = data->widget_data->box.width - depth * 5;
 
 		data->box->ops->draw(item, data, x, y, width);
+
 	} else {
-		draw_text(data->term, data->widget_data->x + depth * 5, y,
-			  text, len, 0, color);
+		unsigned char *text;
+		struct listbox_ops *ops = data->box->ops;
+
+		assert(ops && ops->get_info);
+
+		text = ops->get_text(item, data->term);
+		if (!text) return 0;
+
+		len = strlen(text);
+		int_upper_bound(&len, int_max(0, data->widget_data->box.width - depth * 5));
+
+		draw_text(data->term, x, y, text, len, 0, color);
+
+		mem_free(text);
 	}
 
-
 	if (item == data->box->sel) {
-		int x = data->widget_data->x;
+		x = data->widget_data->box.x;
 
 		/* For blind users: */
-		set_cursor(data->term, x, y, 0);
+		set_cursor(data->term, x, y, 1);
 		set_window_ptr(data->dlg_data->win, x, y);
 	}
 
@@ -420,30 +463,27 @@ display_listbox_item(struct listbox_item *item, void *data_, int *offset)
 }
 
 /* Displays a dialog box */
-static void
-display_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
-		int sel)
+static t_handler_event_status
+display_listbox(struct dialog_data *dlg_data, struct widget_data *widget_data)
 {
 	struct terminal *term = dlg_data->win->term;
 	struct listbox_data *box = get_listbox_widget_data(widget_data);
 	struct listbox_context data;
 	/* Add one to offset to get the actual height of the selected item */
-	int moves = box->sel_offset + 1 - widget_data->h;
+	int moves = box->sel_offset + 1 - widget_data->box.height;
 
 	if (moves > 0) {
 		/* Move selected listbox to visible and update box->top while we're
 		 * at it. Fixes bug 58. */
-		box_sel_move(widget_data, -moves);
+		listbox_sel_move(widget_data, -moves);
 
 	} else if (!list_empty(*box->items)) {
 		if (!box->top) box->top = box->items->next;
 		if (!box->sel) box->sel = box->top;
 	}
 
-	draw_area(term, widget_data->x, widget_data->y,
-		  widget_data->w, widget_data->h, ' ', 0,
-		  get_bfu_color(term, "menu.normal"));
-
+	draw_box(term, &widget_data->box, ' ', 0,
+		 get_bfu_color(term, "menu.normal"));
 
 	/* We want to have these visible if possible. */
 	if (box->top && !box->top->visible) {
@@ -453,14 +493,16 @@ display_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
 		box->sel = box->top;
 	}
 
-	memset(&data, 0, sizeof(struct listbox_context));
+	memset(&data, 0, sizeof(data));
 	data.term = term;
 	data.widget_data = widget_data;
 	data.box = box;
 	data.dlg_data = dlg_data;
 
-	traverse_listbox_items_list(box->top, box, widget_data->h,
+	traverse_listbox_items_list(box->top, box, widget_data->box.height,
 				    1, display_listbox_item, &data);
+
+	return EVENT_PROCESSED;
 }
 
 static int
@@ -479,17 +521,16 @@ check_old_state(struct listbox_item *item, void *info_, int *offset)
 	return 0;
 }
 
-static void
-init_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
-	     struct term_event *ev)
+static t_handler_event_status
+init_listbox(struct dialog_data *dlg_data, struct widget_data *widget_data)
 {
 	struct hierbox_browser *browser = dlg_data->dlg->udata2;
-	struct listbox_data *box =
-		(struct listbox_data *) widget_data->widget->data;
+	struct listbox_data *box = get_listbox_widget_data(widget_data);
 
 	/* Try to restore the position from last time */
-	memcpy(box, &browser->box_data, sizeof(struct listbox_data));
-	if (box->items) {
+	if (!list_empty(browser->root.child) && browser->box_data.items) {
+		copy_struct(box, &browser->box_data);
+
 		traverse_listbox_items_list(browser->root.child.next, box, 0, 0,
 					    check_old_state, box);
 
@@ -503,61 +544,61 @@ init_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
 	box->items = &browser->root.child;
 
 	add_to_list(browser->boxes, box);
+	return EVENT_PROCESSED;
 }
 
-static int
-mouse_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
-	      struct term_event *ev)
+static t_handler_event_status
+mouse_listbox(struct dialog_data *dlg_data, struct widget_data *widget_data)
 {
 #ifdef CONFIG_MOUSE
 	struct listbox_data *box = get_listbox_widget_data(widget_data);
+	struct term_event *ev = dlg_data->term_event;
 
 	if (!list_empty(*box->items)) {
 		if (!box->top) box->top = box->items->next;
 		if (!box->sel) box->sel = box->top;
 	}
 
-	if ((ev->b & BM_ACT) == B_DOWN) {
+	if (check_mouse_action(ev, B_DOWN)) {
 		struct widget_data *dlg_item = dlg_data->widgets_data;
 
-		switch (ev->b & BM_BUTT) {
+		switch (get_mouse_button(ev)) {
 			case B_WHEEL_DOWN:
-				box_sel_move(dlg_item, 1);
-				display_dlg_item(dlg_data, dlg_item, 1);
+				listbox_sel_move(dlg_item, 1);
+				display_widget(dlg_data, dlg_item);
 				return EVENT_PROCESSED;
 
 			case B_WHEEL_UP:
-				box_sel_move(dlg_item, -1);
-				display_dlg_item(dlg_data, dlg_item, 1);
+				listbox_sel_move(dlg_item, -1);
+				display_widget(dlg_data, dlg_item);
 				return EVENT_PROCESSED;
 		}
 	}
 
-	if ((ev->b & BM_ACT) == B_UP) {
-		if ((ev->b & BM_BUTT) < B_WHEEL_UP &&
-		    (ev->y >= widget_data->y && ev->y < widget_data->y + widget_data->h) &&
-		    (ev->x >= widget_data->x && ev->x <= widget_data->x + widget_data->w)) {
-			/* Clicked in the box. */
-			int offset = ev->y - widget_data->y;
+	if (check_mouse_wheel(ev))
+		return EVENT_NOT_PROCESSED;
 
-			box->sel_offset = offset;
-			if (offset)
-				box->sel = traverse_listbox_items_list(box->top,
-								       box,
-								       offset, 1,
-								       NULL, NULL);
-			else box->sel = box->top;
+	if (check_mouse_position(ev, &widget_data->box)) {
+		/* Clicked in the box. */
+		int offset = ev->info.mouse.y - widget_data->box.y;
 
-			if (box->sel && box->sel->type == BI_FOLDER) {
-				int xdepth = widget_data->x + box->sel->depth * 5;
+		box->sel_offset = offset;
+		box->sel = offset ? traverse_listbox_items_list(box->top, box,
+								offset, 1,
+								NULL, NULL)
+				  : box->top;
 
-			       	if (ev->x >= xdepth && ev->x <= xdepth + 2)
-					box->sel->expanded = !box->sel->expanded;
-			}
+		if (box->sel && box->sel->type == BI_FOLDER) {
+			int xdepth = widget_data->box.x + box->sel->depth * 5;
+			int x = ev->info.mouse.x;
 
-			display_dlg_item(dlg_data, widget_data, 1);
-			return EVENT_PROCESSED;
+			if (x >= xdepth && x <= xdepth + 2)
+				box->sel->expanded = !box->sel->expanded;
 		}
+
+		display_widget(dlg_data, widget_data);
+
+		return EVENT_PROCESSED;
 	}
 
 #endif /* CONFIG_MOUSE */
@@ -565,78 +606,78 @@ mouse_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
 	return EVENT_NOT_PROCESSED;
 }
 
-static int
-kbd_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
-	    struct term_event *ev)
+static t_handler_event_status
+kbd_listbox(struct dialog_data *dlg_data, struct widget_data *widget_data)
 {
 	struct widget_data *dlg_item = dlg_data->widgets_data;
+	struct term_event *ev = dlg_data->term_event;
 
 	/* Not a pure listbox, but you're not supposed to use this outside of
 	 * the listbox browser anyway, so what.. */
 
 	switch (ev->ev) {
-		enum keyact action;
+		enum menu_action action;
 
-		case EV_KBD:
-			action = kbd_action(KM_MENU, ev, NULL);
+		case EVENT_KBD:
+			action = kbd_action(KEYMAP_MENU, ev, NULL);
 
 			/* Moving the box */
-			if (action == ACT_DOWN) {
-				box_sel_move(dlg_item, 1);
-				display_dlg_item(dlg_data, dlg_item, 1);
+			if (action == ACT_MENU_DOWN) {
+				listbox_sel_move(dlg_item, 1);
+				display_widget(dlg_data, dlg_item);
 
 				return EVENT_PROCESSED;
 			}
 
-			if (action == ACT_UP) {
-				box_sel_move(dlg_item, -1);
-				display_dlg_item(dlg_data, dlg_item, 1);
+			if (action == ACT_MENU_UP) {
+				listbox_sel_move(dlg_item, -1);
+				display_widget(dlg_data, dlg_item);
 
 				return EVENT_PROCESSED;
 			}
 
-			if (action == ACT_PAGE_DOWN) {
-				box_sel_move(dlg_item, dlg_item->h / 2);
-				display_dlg_item(dlg_data, dlg_item, 1);
+			if (action == ACT_MENU_PAGE_DOWN) {
+				listbox_sel_move(dlg_item, dlg_item->box.height / 2);
+				display_widget(dlg_data, dlg_item);
 
 				return EVENT_PROCESSED;
 			}
 
-			if (action == ACT_PAGE_UP) {
-				box_sel_move(dlg_item, -dlg_item->h / 2);
-				display_dlg_item(dlg_data, dlg_item, 1);
+			if (action == ACT_MENU_PAGE_UP) {
+				listbox_sel_move(dlg_item, -dlg_item->box.height / 2);
+				display_widget(dlg_data, dlg_item);
 
 				return EVENT_PROCESSED;
 			}
 
-			if (action == ACT_HOME) {
-				box_sel_move(dlg_item, -MAXINT);
-				display_dlg_item(dlg_data, dlg_item, 1);
+			if (action == ACT_MENU_HOME) {
+				listbox_sel_move(dlg_item, -INT_MAX);
+				display_widget(dlg_data, dlg_item);
 
 				return EVENT_PROCESSED;
 			}
 
-			if (action == ACT_END) {
-				box_sel_move(dlg_item, MAXINT);
-				display_dlg_item(dlg_data, dlg_item, 1);
+			if (action == ACT_MENU_END) {
+				listbox_sel_move(dlg_item, INT_MAX);
+				display_widget(dlg_data, dlg_item);
 
 				return EVENT_PROCESSED;
 			}
 
-			if (action == ACT_MARK_ITEM) {
+			if (action == ACT_MENU_MARK_ITEM) {
 				struct listbox_data *box;
 
 				box = get_listbox_widget_data(dlg_item);
 				if (box->sel) {
 					box->sel->marked = !box->sel->marked;
-					box_sel_move(dlg_item, 1);
+					listbox_sel_move(dlg_item, 1);
 				}
-				display_dlg_item(dlg_data, dlg_item, 1);
+				display_widget(dlg_data, dlg_item);
 
 				return EVENT_PROCESSED;
 			}
 
-			if (action == ACT_DELETE) {
+			if (action == ACT_MENU_DELETE) {
 				struct listbox_data *box;
 
 				box = get_listbox_widget_data(dlg_item);
@@ -652,11 +693,11 @@ kbd_listbox(struct widget_data *widget_data, struct dialog_data *dlg_data,
 			/* Selecting a button; most probably ;). */
 			break;
 
-		case EV_INIT:
-		case EV_RESIZE:
-		case EV_REDRAW:
-		case EV_MOUSE:
-		case EV_ABORT:
+		case EVENT_INIT:
+		case EVENT_RESIZE:
+		case EVENT_REDRAW:
+		case EVENT_MOUSE:
+		case EVENT_ABORT:
 			break;
 	}
 
@@ -669,4 +710,5 @@ struct widget_ops listbox_ops = {
 	mouse_listbox,
 	kbd_listbox,
 	NULL,
+	NULL,
 };
