@@ -1,16 +1,14 @@
-/* $Id: bookmarks.h,v 1.30 2004/01/02 18:37:56 jonas Exp $ */
+/* $Id: bookmarks.h,v 1.36 2004/07/15 00:45:33 jonas Exp $ */
 
 #ifndef EL__BOOKMARKS_BOOKMARKS_H
 #define EL__BOOKMARKS_BOOKMARKS_H
 
-#ifdef CONFIG_BOOKMARKS
-
-/* #include "bfu/listbox.h" */
-struct listbox_item;
-struct terminal;
-
 #include "modules/module.h"
 #include "util/lists.h"
+#include "util/object.h"
+
+struct listbox_item;
+struct terminal;
 
 /* Bookmark record structure */
 
@@ -25,7 +23,7 @@ struct bookmark {
 
 	unsigned char *title;   /* title of bookmark */
 	unsigned char *url;     /* Location of bookmarked item */
-	int refcount;		/* No direct access, use provided macros for that. */
+	struct object object;	/* No direct access, use provided macros for that. */
 
 	struct list_head child;
 };
@@ -34,8 +32,6 @@ struct bookmark {
 
 extern struct list_head bookmarks; /* struct bookmark */
 
-extern int bookmarks_dirty;
-
 /* The bookmarks module */
 
 extern struct module bookmarks_module;
@@ -46,12 +42,16 @@ void read_bookmarks(void);
 void write_bookmarks(void);
 
 /* Bookmarks manipulation */
+void bookmarks_set_dirty(void);
+void bookmarks_unset_dirty(void);
+int bookmarks_are_dirty(void);
 
-int delete_bookmark(struct bookmark *);
+void delete_bookmark(struct bookmark *);
 struct bookmark *add_bookmark(struct bookmark *, int, unsigned char *, unsigned char *);
+struct bookmark *get_bookmark(unsigned char *url);
 void bookmark_terminal_tabs(struct terminal *term, unsigned char *foldername);
+void bookmark_auto_save_tabs(struct terminal *term);
 int update_bookmark(struct bookmark *, unsigned char *, unsigned char *);
 void open_bookmark_folder(struct session *ses, unsigned char *foldername);
 
 #endif
-#endif
