@@ -1,5 +1,5 @@
 /* Command line processing */
-/* $Id: cmdline.c,v 1.30 2004/01/02 19:54:51 pasky Exp $ */
+/* $Id: cmdline.c,v 1.107.2.3 2005/04/06 11:07:48 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -8,12 +8,18 @@
 #include <stdio.h>
 #include <string.h>
 #include <sys/types.h>
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h> /* OS/2 needs this after sys/types.h */
+#endif
 #include <sys/stat.h> /* OS/2 needs this after sys/types.h */
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
+#endif
 
 /* We need to have it here. Stupid BSD. */
+#ifdef HAVE_NETINET_IN_H
 #include <netinet/in.h>
+#endif
 #ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
 #endif
@@ -26,14 +32,22 @@
 #include "config/opttypes.h"
 #include "intl/gettext/libintl.h"
 #include "lowlevel/dns.h"
+#include "protocol/uri.h"
+#include "sched/session.h"
+#include "util/error.h"
+#include "util/file.h"
+#include "util/lists.h"
 #include "util/memory.h"
+#include "util/string.h"
 
 
-static unsigned char *
-_parse_options(int argc, unsigned char *argv[], struct option *opt)
+/* Hack to handle URL extraction for -remote commands */
+static unsigned char *remote_url;
+
+static int
+parse_options_(int argc, unsigned char *argv[], struct option *opt,
+               struct list_head *url_list)
 {
-	unsigned char *location = NULL;
-
 	while (argc) {
 		argv++, argc--;
 
@@ -41,6 +55,7 @@ _parse_options(int argc, unsigned char *argv[], struct option *opt)
 			struct option *option;
 			unsigned char *argname = &argv[-1][1];
 			unsigned char *oname = stracpy(argname);
+			unsigned char *err;
 
 			if (!oname) continue;
 
@@ -65,43 +80,41 @@ _parse_options(int argc, unsigned char *argv[], struct option *opt)
 
 			mem_free(oname);
 
-			if (!option) {
+			if (!option) goto unknown_option;
+
+			if (!option_types[option->type].cmdline)
 				goto unknown_option;
+
+			err = option_types[option->type].cmdline(option, &argv, &argc);
+
+			if (err) {
+				if (err[0])
+					usrerror(gettext("Cannot parse option %s: %s"), argv[-1], err);
+
+				return 1;
+
+			} else if (remote_url) {
+				if (url_list) add_to_string_list(url_list, remote_url, -1);
+				mem_free(remote_url);
+				remote_url = NULL;
 			}
 
-			if (option_types[option->type].cmdline
-			    && !(option->flags & OPT_HIDDEN)) {
-				unsigned char *err;
-
-				err = option_types[option->type].cmdline(option, &argv, &argc);
-
-				if (err) {
-					if (err[0])
-						ERROR(gettext("Cannot parse option %s: %s"), argv[-1], err);
-
-					return NULL;
-				}
-			} else {
-				goto unknown_option;
-			}
-
-		} else if (!location) {
-			location = argv[-1];
-
-		} else {
-unknown_option:
-			ERROR(gettext("Unknown option %s"), argv[-1]);
-			return NULL;
+		} else if (url_list) {
+			add_to_string_list(url_list, argv[-1], -1);
 		}
 	}
 
-	return empty_string_or_(location);
+	return 0;
+
+unknown_option:
+	usrerror(gettext("Unknown option %s"), argv[-1]);
+	return 1;
 }
 
-unsigned char *
-parse_options(int argc, unsigned char *argv[])
+int
+parse_options(int argc, unsigned char *argv[], struct list_head *url_list)
 {
-	return _parse_options(argc, argv, cmdline_options);
+	return parse_options_(argc, argv, cmdline_options, url_list);
 }
 
 
@@ -144,13 +157,13 @@ lookup_cmd(struct option *o, unsigned char ***argv, int *argc)
 #ifdef HAVE_HERROR
 		herror(gettext("error"));
 #else
-		ERROR(gettext("Host not found"));
+		usrerror(gettext("Host not found"));
 #endif
 		return "";
 	}
 
 	for (i = 0; i < addrno; i++) {
-#ifdef IPV6
+#ifdef CONFIG_IPV6
 		struct sockaddr_in6 addr = *((struct sockaddr_in6 *) &(addrs)[i]);
 		unsigned char p[INET6_ADDRSTRLEN];
 
@@ -170,17 +183,137 @@ lookup_cmd(struct option *o, unsigned char ***argv, int *argc)
 #endif
 	}
 
-	if (addrs) mem_free(addrs);
+	mem_free_if(addrs);
 
 	fflush(stdout);
 
 	return "";
 }
 
+#define skipback_whitespace(start, S) \
+	while ((start) < (S) && isspace((S)[-1])) (S)--;
+
+static unsigned char *
+remote_cmd(struct option *o, unsigned char ***argv, int *argc)
+{
+	struct {
+		unsigned char *name;
+		enum {
+			REMOTE_METHOD_OPENURL,
+			REMOTE_METHOD_PING,
+			REMOTE_METHOD_XFEDOCOMMAND,
+			REMOTE_METHOD_ADDBOOKMARK,
+			REMOTE_METHOD_NOT_SUPPORTED,
+		} type;
+	} remote_methods[] = {
+		{ "openURL",	  REMOTE_METHOD_OPENURL },
+		{ "ping",	  REMOTE_METHOD_PING },
+		{ "addBookmark",  REMOTE_METHOD_ADDBOOKMARK },
+		{ "xfeDoCommand", REMOTE_METHOD_XFEDOCOMMAND },
+		{ NULL,		  REMOTE_METHOD_NOT_SUPPORTED },
+	};
+	unsigned char *command, *arg, *argend = NULL;
+	int method, len = 0;
+
+	if (*argc < 1) return gettext("Parameter expected");
+
+	command = *(*argv);
+
+	while (isalpha(command[len]))
+		len++;
+
+	arg = strchr(&command[len], '(');
+	if (arg) {
+		arg++;
+		skip_space(arg);
+
+		argend = strchr(arg, ')');
+	}
+
+	if (!argend) {
+		/* Just open any passed URLs in new tabs */
+		remote_session_flags |= SES_REMOTE_NEW_TAB;
+		return NULL;
+	}
+
+	skipback_whitespace(arg, argend);
+
+	for (method = 0; remote_methods[method].name; method++) {
+		unsigned char *name = remote_methods[method].name;
+
+		if (!strlcasecmp(command, len, name, -1))
+			break;
+	}
+
+	switch (remote_methods[method].type) {
+	case REMOTE_METHOD_OPENURL:
+		if (arg == argend) {
+			/* Prompt for a URL with a dialog box */
+			remote_session_flags |= SES_REMOTE_PROMPT_URL;
+			break;
+		}
+
+		len = strcspn(arg, ",)");
+		if (arg[len] == ',') {
+			unsigned char *where = arg + len + 1;
+
+			if (strstr(where, "new-window")) {
+				remote_session_flags |= SES_REMOTE_NEW_WINDOW;
+
+			} else if (strstr(where, "new-tab")) {
+				remote_session_flags |= SES_REMOTE_NEW_TAB;
+
+			} else {
+				/* Bail out when getting unknown parameter */
+				/* TODO: new-screen */
+				break;
+			}
+
+		} else {
+			remote_session_flags |= SES_REMOTE_CURRENT_TAB;
+		}
+
+		while (len > 0 && isspace(arg[len - 1])) len--;
+		if (len) remote_url = memacpy(arg, len);
+
+		break;
+
+	case REMOTE_METHOD_XFEDOCOMMAND:
+		len = argend - arg;
+
+		if (!strlcasecmp(arg, len, "openBrowser", 11)) {
+			remote_session_flags = SES_REMOTE_NEW_WINDOW;
+		}
+		break;
+
+	case REMOTE_METHOD_PING:
+		remote_session_flags = SES_REMOTE_PING;
+		break;
+
+	case REMOTE_METHOD_ADDBOOKMARK:
+		if (arg == argend) break;
+		remote_url = memacpy(arg, argend - arg);
+		remote_session_flags = SES_REMOTE_ADD_BOOKMARK;
+		break;
+
+	case REMOTE_METHOD_NOT_SUPPORTED:
+		break;
+	}
+
+	/* If no flags was applied it can only mean we are dealing with
+	 * unknown method. */
+	if (!remote_session_flags)
+		return gettext("Remote method not supported");
+
+	(*argv)++; (*argc)--;	/* Consume next argument */
+
+	return NULL;
+}
+
 static unsigned char *
 version_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
-	printf(gettext("ELinks %s - Text WWW browser\n"), VERSION_STRING);
+	printf("%s\n", full_static_version);
 	fflush(stdout);
 	return "";
 }
@@ -215,8 +348,10 @@ print_full_help(struct option *tree, unsigned char *path)
 				      ? (unsigned char *) gettext(option->desc)
 				      : (unsigned char *) "N/A";
 
-		/* Don't print deprecated aliases and command line options */
-		if (type == OPT_ALIAS && tree != cmdline_options)
+		/* Don't print deprecated aliases (if we don't walk command
+		 * line options which use aliases for legitimate options). */
+		if ((type == OPT_ALIAS && tree != cmdline_options)
+		    || (option->flags & OPT_HIDDEN))
 			continue;
 
 		if (!capt && !strncasecmp(option->name, "_template_", 10))
@@ -244,7 +379,9 @@ print_full_help(struct option *tree, unsigned char *path)
 			case OPT_INT:
 			case OPT_LONG:
 				printf(gettext("(default: %ld)"),
-					(long) option->value.number);
+					type == OPT_LONG
+					? option->value.big_number
+					: (long) option->value.number);
 				break;
 
 			case OPT_STRING:
@@ -264,18 +401,13 @@ print_full_help(struct option *tree, unsigned char *path)
 
 			case OPT_COLOR:
 			{
-				unsigned char *name;
+				color_t color = option->value.color;
+				unsigned char hexcolor[8];
 
-				name = get_color_name(option->value.color);
-				if (name) {
-					printf(gettext("(default: %s)"), name);
-					mem_free(name);
-					break;
-				}
-				printf(gettext("(default: #%06lx)"),
-					(unsigned long) option->value.color);
-			}
+				printf(gettext("(default: %s)"),
+				       get_color_string(color, hexcolor));
 				break;
+			}
 
 			case OPT_COMMAND:
 				break;
@@ -283,7 +415,7 @@ print_full_help(struct option *tree, unsigned char *path)
 			case OPT_LANGUAGE:
 #ifdef ENABLE_NLS
 				printf(gettext("(default: \"%s\")"),
-					language_to_name(option->value.number));
+				       language_to_name(option->value.number));
 #endif
 				break;
 
@@ -355,6 +487,10 @@ print_short_help()
 					    : (unsigned char *) "";
 		int len = strlen(option->name);
 
+		/* Avoid printing compatibility options */
+		if (option->flags & OPT_HIDDEN)
+			continue;
+
 		/* When no caption is available the option name is 'stacked'
 		 * and the caption is shared with next options that has one. */
 		if (!option->capt) {
@@ -379,7 +515,8 @@ print_short_help()
 		len = (len < 0) ? !!(*help) : len;
 
 		align[len] = '\0';
-		printf("  -%s%s %s%s%s\n", info, option->name, help, align, capt);
+		printf("  -%s%s %s%s%s\n",
+		       info, option->name, help, align, capt);
 		align[len] = ' ';
 		if (saved) {
 			done_string(saved);
@@ -394,15 +531,19 @@ print_short_help()
 static unsigned char *
 printhelp_cmd(struct option *option, unsigned char ***argv, int *argc)
 {
-	version_cmd(NULL, NULL, NULL);
-	printf("\n");
+	unsigned char *lineend = strchr(full_static_version, '\n');
+
+	if (lineend) *lineend = '\0';
+
+	printf("%s\n\n", full_static_version);
 
 	if (!strcmp(option->name, "config-help")) {
-		printf(gettext("Configuration options:\n"));
+		printf("%s:\n", gettext("Configuration options"));
 		print_full_help(config_options, "");
 	} else {
-		printf(gettext("Usage: elinks [OPTION]... [URL]\n\n"));
-		printf(gettext("Options:\n"));
+		printf("%s\n\n%s:\n",
+		       gettext("Usage: elinks [OPTION]... [URL]"),
+		       gettext("Options"));
 		if (!strcmp(option->name, "long-help")) {
 			print_full_help(cmdline_options, "-");
 		} else {
@@ -414,6 +555,45 @@ printhelp_cmd(struct option *option, unsigned char ***argv, int *argc)
 	return "";
 }
 
+static unsigned char *
+redir_cmd(struct option *option, unsigned char ***argv, int *argc)
+{
+	unsigned char *target;
+
+	/* I can't get any dirtier. --pasky */
+
+	if (!strcmp(option->name, "confdir")) {
+		target = "config-dir";
+	} else if (!strcmp(option->name, "conffile")) {
+		target = "config-file";
+	} else if (!strcmp(option->name, "stdin")) {
+		static int complained;
+
+		if (complained)
+			return NULL;
+		complained = 1;
+
+		/* Emulate bool option, possibly eating following 0/1. */
+		if ((*argv)[0]
+		    && ((*argv)[0][0] == '0' || (*argv)[0][0] == '1')
+		    && !(*argv)[0][1])
+			(*argv)++, (*argc)--;
+		fprintf(stderr, "Warning: Deprecated option -stdin used!\n");
+		fprintf(stderr, "ELinks now determines the -stdin option value automatically.\n");
+		fprintf(stderr, "In the future versions ELinks will report error when you will\n");
+		fprintf(stderr, "continue to use this option.\a\n");
+		return NULL;
+
+	} else {
+		return gettext("Internal consistency error");
+	}
+
+	option = get_opt_rec(cmdline_options, target);
+	assert(option);
+	option_types[option->type].cmdline(option, argv, argc);
+	return NULL;
+}
+
 
 /**********************************************************************
  Options values
@@ -424,60 +604,77 @@ printhelp_cmd(struct option *option, unsigned char ***argv, int *argc)
 struct option_info cmdline_options_info[] = {
 	INIT_OPT_BOOL("", N_("Restrict to anonymous mode"),
 		"anonymous", 0, 0,
-		N_("Restrict ELinks so that it can run on an anonymous account.\n"
-		"No local file browsing, no downloads. Execution of viewers\n"
-		"is allowed, but user can't add or modify entries in\n"
-		"association table.")),
+		N_("Restricts ELinks so it can run on an anonymous account.\n"
+		"Local file browsing, downloads, and modification of options\n"
+		"will be disabled. Execution of viewers is allowed, but entries\n"
+		"in the association table can't be added or modified.")),
 
 	INIT_OPT_BOOL("", N_("Autosubmit first form"),
 		"auto-submit", 0, 0,
-		N_("Go and submit the first form you'll stumble upon.")),
+		N_("Automatically submit the first form in the given URLs.")),
 
-	INIT_OPT_INT("", N_("Clone session with given ID"),
-		"base-session", 0, 0, MAXINT, 0,
-		N_("ID of session (ELinks instance) which we want to clone.\n"
-		"This is internal ELinks option, you don't want to use it.")),
+	INIT_OPT_INT("", N_("Clone internal session with given ID"),
+		"base-session", 0, 0, INT_MAX, 0,
+		N_("Used internally when opening ELinks instances in new windows.\n"
+		"The ID maps to information that will be used when creating the\n"
+		"new instance. You don't want to use it.")),
 
-	INIT_OPT_STRING("", N_("Set config dir to given string"),
-		"confdir", 0, "",
-		N_("Set the config dir to the given path. ELinks will read\n"
-		"its config files and writes to it. If the path begins with\n"
-		"a '/' its used as an absolute path. Else it is assumed to\n"
-		"be relative to your HOME dir.")),
+	INIT_OPT_COMMAND("", NULL, "confdir", OPT_HIDDEN, redir_cmd, NULL),
 
-	INIT_OPT_STRING("", N_("Configuration file name"),
-		"conffile", 0, "elinks.conf",
-		N_("Name of the file with configuration, from which and to\n"
-		"which all the configuration shall be written. It should be\n"
-		"relative to confdir.")),
+	INIT_OPT_STRING("", N_("Name of directory with configuration file"),
+		"config-dir", 0, "",
+		N_("Path of the directory ELinks will read and write its\n"
+		"config and runtime state files to instead of ~/.elinks.\n"
+		"If the path does not begin with a '/' it is assumed to be\n"
+		"relative to your HOME directory.")),
 
-	INIT_OPT_ALIAS("", N_("MIME type to assume for documents"),
+	INIT_OPT_COMMAND("", NULL, "conffile", OPT_HIDDEN, redir_cmd, NULL),
+
+	INIT_OPT_STRING("", N_("Name of configuration file"),
+		"config-file", 0, "elinks.conf",
+		N_("Name of the configuration file that all configuration\n"
+		"options will be read from and written to. It should be\n"
+		"relative to config-dir.")),
+
+	INIT_OPT_COMMAND("", N_("Print help for configuration options"),
+		"config-help", 0, printhelp_cmd,
+		N_("Print help for configuration options and exit.")),
+
+	INIT_OPT_CMDALIAS("", N_("MIME type assumed for unknown document types"),
 		"default-mime-type", 0, "mime.default_type",
-		N_("Default MIME type to assume for documents of unknown type.")),
+		N_("The default MIME type used for documents of unknown type.")),
 
-	INIT_OPT_BOOL("", N_("Write formatted version of given URL to stdout"),
+	INIT_OPT_BOOL("", N_("Ignore user-defined keybindings"),
+		"default-keys", 0, 0,
+		N_("When set, all keybindings from configuration files will be\n"
+		"ignored. It forces use of default keybindings and will reset\n"
+		"user-defined ones on save.")),
+
+	INIT_OPT_BOOL("", N_("Print formatted versions of given URLs to stdout"),
 		"dump", 0, 0,
-		N_("Write a plain-text version of the given HTML document to\n"
-		"stdout.")),
+		N_("Print formatted plain-text versions of given URLs to stdout.")),
 
-	INIT_OPT_ALIAS("", N_("Codepage to use with -dump"),
+	INIT_OPT_CMDALIAS("", N_("Codepage to use with -dump"),
 		"dump-charset", 0, "document.dump.codepage",
-		N_("Codepage used in dump output.")),
+		N_("Codepage used when formatting dump output.")),
 
-	INIT_OPT_ALIAS("", N_("Width of document formatted with -dump"),
+	INIT_OPT_CMDALIAS("", N_("Width of document formatted with -dump"),
 		"dump-width", 0, "document.dump.width",
 		N_("Width of the dump output.")),
 
-	INIT_OPT_COMMAND("", N_("Evaluate given configuration option"),
+	INIT_OPT_COMMAND("", N_("Evaluate configuration file directive"),
 		"eval", 0, eval_cmd,
-		N_("Specify elinks.conf config options on the command-line:\n"
-		"  -eval 'set protocol.file.allow_special_files = 1'")),
+		N_("Specify configuration file directives on the command-line\n"
+		"which will be evaluated after all configuration files has been\n"
+		"read. Example usage:\n"
+		"\t-eval 'set protocol.file.allow_special_files = 1'")),
 
 	/* lynx compatibility */
-	INIT_OPT_COMMAND("", N_("Assume the file is HTML"),
+	INIT_OPT_COMMAND("", N_("Interpret documents of unknown types as HTML"),
 		"force-html", 0, forcehtml_cmd,
-		N_("This makes ELinks assume that the files it sees are HTML. This is\n"
-		"equivalent to -default-mime-type text/html.")),
+		N_("Makes ELinks assume documents of unknown types are HTML.\n"
+		"Useful when using ELinks as an external viewer from MUAs.\n"
+		"This is equivalent to -default-mime-type text/html.")),
 
 	/* XXX: -?, -h and -help share the same caption and should be kept in
 	 * the current order for usage help printing to be ok */
@@ -489,31 +686,63 @@ struct option_info cmdline_options_info[] = {
 		"help", 0, printhelp_cmd,
 		N_("Print usage help and exit.")),
 
+	INIT_OPT_BOOL("", N_("Only permit local connections"),
+		"localhost", 0, 0,
+		N_("Restricts ELinks to work offline and only connect to servers\n"
+		"with local addresses (ie. 127.0.0.1). No connections to remote\n"
+		"servers will be permitted.")),
+
 	INIT_OPT_COMMAND("", N_("Print detailed usage help and exit"),
 		"long-help", 0, printhelp_cmd,
 		N_("Print detailed usage help and exit.")),
 
-	INIT_OPT_COMMAND("", N_("Print help for configuration options"),
-		"config-help", 0, printhelp_cmd,
-		N_("Print help on configuration options and exit.")),
-
 	INIT_OPT_COMMAND("", N_("Look up specified host"),
 		"lookup", 0, lookup_cmd,
-		N_("Look up specified host.")),
+		N_("Look up specified host and print all DNS resolved IP addresses.")),
 
 	INIT_OPT_BOOL("", N_("Run as separate instance"),
 		"no-connect", 0, 0,
 		N_("Run ELinks as a separate instance instead of connecting to an\n"
 		"existing instance. Note that normally no runtime state files\n"
-		"(bookmarks, history and so on) are written to the disk when\n"
-		"this option is used. See also -touch-files.")),
+		"(bookmarks, history, etc.) are written to the disk when this\n"
+		"option is used. See also -touch-files.")),
 
-	INIT_OPT_BOOL("", N_("Don't use files in ~/.elinks"),
+	INIT_OPT_BOOL("", N_("Disable use of files in ~/.elinks"),
 		"no-home", 0, 0,
-		N_("Don't attempt to create and/or use home rc directory (~/.elinks).")),
+		N_("Disables creation and use of files in the user specific home\n"
+		"configuration directory (~/.elinks). It forces default configuration\n"
+		"values to be used and disables saving of runtime state files.")),
+
+	INIT_OPT_BOOL("", N_("Disable link numbering in dump output"),
+		"no-numbering", 0, 0,
+		N_("Prevents printing of link number in dump output.\n"
+		"Note that this really affects only -dump, nothing else.")),
+
+	INIT_OPT_BOOL("", N_("Disable printing of link references in dump output"),
+		"no-references", 0, 0,
+		N_("Prevents printing of references (URIs) of document links\n"
+		"in dump output.\n"
+		"Note that this really affects only -dump, nothing else.")),
+
+	INIT_OPT_COMMAND("", N_("Control an already running ELinks"),
+		"remote", 0, remote_cmd,
+		N_("Control a remote ELinks instance by passing commands to it.\n"
+		"The option takes an additional argument containing the method\n"
+		"which should be invoked and any parameters that should be passed\n"
+		"to it. For ease of use, the additional method argument can be\n"
+		"omitted in which case any URL arguments will be opened in new\n"
+		"tabs in the remote instance.\n"
+		"Following is a list of the supported methods:\n"
+		"\tping()                    : look for a remote instance\n"
+		"\topenURL()                 : prompt URL in current tab\n"
+		"\topenURL(URL)              : open URL in current tab\n"
+		"\topenURL(URL, new-tab)     : open URL in new tab\n"
+		"\topenURL(URL, new-window)  : open URL in new window\n"
+		"\taddBookmark(URL)          : bookmark URL\n"
+		"\txfeDoCommand(openBrowser) : open new window")),
 
 	INIT_OPT_INT("", N_("Connect to session ring with given ID"),
-		"session-ring", 0, 0, MAXINT, 0,
+		"session-ring", 0, 0, INT_MAX, 0,
 		N_("ID of session ring this ELinks session should connect to. ELinks\n"
 		"works in so-called session rings, whereby all instances of ELinks\n"
 		"are interconnected and share state (cache, bookmarks, cookies,\n"
@@ -529,29 +758,26 @@ struct option_info cmdline_options_info[] = {
 		"are written to the disk when this option is used. See also\n"
 		"-touch-files.")),
 
-	INIT_OPT_BOOL("", N_("Write the source of given URL to stdout"),
+	INIT_OPT_BOOL("", N_("Print the source of given URLs to stdout"),
 		"source", 0, 0,
-		N_("Write the given HTML document in source form to stdout.")),
+		N_("Print given URLs in source form to stdout.")),
 
-	INIT_OPT_BOOL("", N_("Read document from stdin"),
-		"stdin", 0, 0,
-		N_("Open stdin as an HTML document - this is fully equivalent to:\n"
-		" -eval 'set protocol.file.allow_special_files = 1' file:///dev/stdin\n"
-		"Use whichever suits you more ;-). Note that reading document from\n"
-		"stdin WORKS ONLY WHEN YOU USE -dump OR -source!! (I would like to\n"
-		"know why you would use -source -stdin, though ;-)")),
-
-	INIT_OPT_BOOL("", N_("Do not number links in dump output"),
-		"no-numbering", 0, 0,
-		N_("Prevents numbering of links (and showing their list at the end of\n"
-		"the dumped document) in the -dump output; this was the default behaviour\n"
-		"until 0.5pre12. Note that this really affects only --dump, nothing else.")),
+	INIT_OPT_COMMAND("", NULL, "stdin", OPT_HIDDEN, redir_cmd, NULL),
 
 	INIT_OPT_BOOL("", N_("Touch files in ~/.elinks when running with -no-connect/-session-ring"),
 		"touch-files", 0, 0,
-		N_("Set to 1 to have runtime state files (bookmarks, history, ...)\n"
-		"changed even when -no-connect or -session-ring is used; has no\n"
-		"effect if not used in connection with any of these options.")),
+		N_("When enabled, runtime state files (bookmarks, history, etc.) are\n"
+		"written to disk, even when -no-connect or -session-ring is used.\n"
+		"The option has no effect if not used in conjunction with any of\n"
+		"these options.")),
+
+	INIT_OPT_INT("", N_("Verbose level"),
+		"verbose", 0, 0, VERBOSE_LEVELS - 1, VERBOSE_WARNINGS,
+		N_("The verbose level controls what messages are shown at\n"
+		"start up and while running:\n"
+		"\t0 means only show serious errors\n"
+		"\t1 means show serious errors and warnings\n"
+		"\t2 means show all messages")),
 
 	INIT_OPT_COMMAND("", N_("Print version information and exit"),
 		"version", 0, version_cmd,
