@@ -1,11 +1,12 @@
 /* Config file manipulation */
-/* $Id: conf.c,v 1.121 2004/01/08 21:17:29 pasky Exp $ */
+/* $Id: conf.c,v 1.152.2.1 2005/01/29 00:45:18 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h> /* OS/2 needs this after sys/types.h */
@@ -63,7 +64,7 @@ static unsigned char *
 skip_white(unsigned char *start, int *line)
 {
 	while (*start) {
-		while (WHITECHAR(*start)) {
+		while (isspace(*start)) {
 			if (*start == '\n') {
 				(*line)++;
 			}
@@ -80,15 +81,6 @@ skip_white(unsigned char *start, int *line)
 	return start;
 }
 
-enum parse_error {
-	ERROR_NONE,
-	ERROR_PARSE,
-	ERROR_COMMAND,
-	ERROR_OPTION,
-	ERROR_VALUE,
-	ERROR_NOMEM,
-};
-
 /* Parse a command. Returns error code. */
 /* If dynamic string credentials are supplied, we will mirror the command at
  * the end of the string; however, we won't load the option value to the tree,
@@ -108,7 +100,7 @@ parse_set(struct option *opt_tree, unsigned char **file, int *line,
 
 	/* Option name */
 	optname = *file;
-	while (isA(**file) || **file == '*' || **file == '.') (*file)++;
+	while (isident(**file) || **file == '*' || **file == '.') (*file)++;
 
 	bin = **file;
 	**file = '\0';
@@ -132,7 +124,7 @@ parse_set(struct option *opt_tree, unsigned char **file, int *line,
 		struct option *opt;
 		unsigned char *val;
 
-		opt = get_opt_rec(opt_tree, optname);
+		opt = mirror ? get_opt_rec_real(opt_tree, optname) : get_opt_rec(opt_tree, optname);
 		mem_free(optname);
 
 		if (!opt || (opt->flags & OPT_HIDDEN))
@@ -141,15 +133,18 @@ parse_set(struct option *opt_tree, unsigned char **file, int *line,
 		if (!option_types[opt->type].read)
 			return ERROR_VALUE;
 
-		val = option_types[opt->type].read(opt, file);
-		if (!val) {
-			return ERROR_VALUE;
-		} else if (mirror) {
-			opt->flags |= OPT_WATERMARK;
+		val = option_types[opt->type].read(opt, file, line);
+		if (!val) return ERROR_VALUE;
+
+		if (mirror) {
+			if (opt->flags & OPT_DELETED)
+				opt->flags &= ~OPT_WATERMARK;
+			else
+				opt->flags |= OPT_WATERMARK;
 			if (option_types[opt->type].write) {
 				option_types[opt->type].write(opt, mirror);
 			}
-		} else if (!val || !option_types[opt->type].set
+		} else if (!option_types[opt->type].set
 			   || !option_types[opt->type].set(opt, val)) {
 			mem_free(val);
 			return ERROR_VALUE;
@@ -163,6 +158,55 @@ parse_set(struct option *opt_tree, unsigned char **file, int *line,
 	return ERROR_NONE;
 }
 
+static enum parse_error
+parse_unset(struct option *opt_tree, unsigned char **file, int *line,
+	    struct string *mirror)
+{
+	unsigned char *orig_pos = *file;
+	unsigned char *optname;
+	unsigned char bin;
+
+	/* XXX: This does not handle the autorewriting well and is mostly a
+	 * quick hack than anything now. --pasky */
+
+	*file = skip_white(*file, line);
+	if (!**file) return ERROR_PARSE;
+
+	/* Option name */
+	optname = *file;
+	while (isident(**file) || **file == '*' || **file == '.') (*file)++;
+
+	bin = **file;
+	**file = '\0';
+	optname = stracpy(optname);
+	if (!optname) return ERROR_NOMEM;
+	**file = bin;
+
+	/* Mirror what we have */
+	if (mirror) add_bytes_to_string(mirror, orig_pos, *file - orig_pos);
+
+	{
+		struct option *opt;
+
+		opt = get_opt_rec_real(opt_tree, optname);
+		mem_free(optname);
+
+		if (!opt || (opt->flags & OPT_HIDDEN))
+			return ERROR_OPTION;
+
+		if (!mirror) {
+			if (opt->flags & OPT_ALLOC) delete_option(opt);
+		} else {
+			if (opt->flags & OPT_DELETED)
+				opt->flags |= OPT_WATERMARK;
+			else
+				opt->flags &= ~OPT_WATERMARK;
+		}
+	}
+
+	return ERROR_NONE;
+}
+
 static enum parse_error
 parse_bind(struct option *opt_tree, unsigned char **file, int *line,
 	   struct string *mirror)
@@ -175,13 +219,13 @@ parse_bind(struct option *opt_tree, unsigned char **file, int *line,
 	if (!*file) return ERROR_PARSE;
 
 	/* Keymap */
-	keymap = option_types[OPT_STRING].read(NULL, file);
+	keymap = option_types[OPT_STRING].read(NULL, file, line);
 	*file = skip_white(*file, line);
 	if (!keymap || !**file)
 		return ERROR_OPTION;
 
 	/* Keystroke */
-	keystroke = option_types[OPT_STRING].read(NULL, file);
+	keystroke = option_types[OPT_STRING].read(NULL, file, line);
 	*file = skip_white(*file, line);
 	if (!keystroke || !**file) {
 		mem_free(keymap);
@@ -204,7 +248,7 @@ parse_bind(struct option *opt_tree, unsigned char **file, int *line,
 
 	/* Action */
 	next_pos = *file;
-	action = option_types[OPT_STRING].read(NULL, file);
+	action = option_types[OPT_STRING].read(NULL, file, line);
 	if (!action) {
 		mem_free(keymap);
 		return ERROR_VALUE;
@@ -215,14 +259,22 @@ parse_bind(struct option *opt_tree, unsigned char **file, int *line,
 		unsigned char *act_str = bind_act(keymap, keystroke);
 
 		if (act_str) {
-			add_bytes_to_string(mirror, orig_pos, next_pos - orig_pos);
+			add_bytes_to_string(mirror, orig_pos,
+					    next_pos - orig_pos);
 			add_to_string(mirror, act_str);
 			mem_free(act_str);
 		} else {
 			err = ERROR_VALUE;
 		}
 	} else {
-		err = bind_do(keymap, keystroke, action) ? ERROR_VALUE : ERROR_NONE;
+		/* We don't bother to bind() if -default-keys. */
+		if (!get_cmd_opt_bool("default-keys")
+		    && bind_do(keymap, keystroke, action)) {
+			/* bind_do() tried but failed. */
+			err = ERROR_VALUE;
+		} else {
+			err = ERROR_NONE;
+		}
 	}
 	mem_free(keymap); mem_free(keystroke); mem_free(action);
 	return err;
@@ -245,7 +297,7 @@ parse_include(struct option *opt_tree, unsigned char **file, int *line,
 	if (!*file) return ERROR_PARSE;
 
 	/* File name */
-	fname = option_types[OPT_STRING].read(NULL, file);
+	fname = option_types[OPT_STRING].read(NULL, file, line);
 	if (!fname)
 		return ERROR_VALUE;
 
@@ -282,12 +334,49 @@ struct parse_handler {
 
 static struct parse_handler parse_handlers[] = {
 	{ "set", parse_set },
+	{ "unset", parse_unset },
 	{ "bind", parse_bind },
 	{ "include", parse_include },
 	{ NULL, NULL }
 };
 
 
+enum parse_error
+parse_config_command(struct option *options, unsigned char **file, int *line,
+		     struct string *mirror)
+{
+	struct parse_handler *handler;
+
+	for (handler = parse_handlers; handler->command;
+	     handler++) {
+		int cmdlen = strlen(handler->command);
+
+		if (!strncmp(*file, handler->command, cmdlen)
+		    && isspace((*file)[cmdlen])) {
+			enum parse_error err;
+			struct string mirror2 = NULL_STRING;
+			struct string *m2 = NULL;
+
+			/* Mirror what we already have */
+			if (mirror && init_string(&mirror2)) {
+				m2 = &mirror2;
+				add_bytes_to_string(m2, *file, cmdlen);
+			}
+
+
+			*file += cmdlen;
+			err = handler->handler(options, file, line, m2);
+			if (!err && mirror && m2) {
+				add_string_to_string(mirror, m2);
+			}
+			if (m2)	done_string(m2);
+			return err;
+		}
+	}
+
+	return ERROR_COMMAND;
+}
+
 void
 parse_config_file(struct option *options, unsigned char *name,
 		  unsigned char *file, struct string *mirror)
@@ -295,7 +384,8 @@ parse_config_file(struct option *options, unsigned char *name,
 	int line = 1;
 	int error_occured = 0;
 	enum parse_error err = 0;
-	unsigned char error_msg[][80] = {
+	enum verbose_level verbose = get_cmd_opt_int("verbose");
+	unsigned char error_msg[][40] = {
 		"no error",
 		"parse error",
 		"unknown command",
@@ -311,62 +401,34 @@ parse_config_file(struct option *options, unsigned char *name,
 		file = skip_white(file, &line);
 
 		/* Mirror what we already have */
-		if (mirror) add_bytes_to_string(mirror, orig_pos, file - orig_pos);
+		if (mirror)
+			add_bytes_to_string(mirror, orig_pos, file - orig_pos);
 
 		/* Second chance to escape from the hell. */
 		if (!*file) break;
 
-		{
-			struct parse_handler *handler;
+		err = parse_config_command(options, &file, &line, mirror);
 
-			for (handler = parse_handlers; handler->command;
-			     handler++) {
-				int cmdlen = strlen(handler->command);
+		if (err == ERROR_COMMAND) {
+			orig_pos = file;
+			/* Jump over this crap we can't understand. */
+			while (!isspace(*file) && *file != '#' && *file)
+				file++;
 
-				if (!strncmp(file, handler->command, cmdlen)
-				    && WHITECHAR(file[cmdlen])) {
-					struct string mirror2 = NULL_STRING;
-					struct string *m2 = NULL;
-
-					/* Mirror what we already have */
-					if (mirror && init_string(&mirror2)) {
-						m2 = &mirror2;
-						add_bytes_to_string(m2,
-								    file, cmdlen);
-					}
-
-
-					file += cmdlen;
-					err = handler->handler(options,
-							       &file, &line,
-							       m2);
-					if (!err && mirror && m2) {
-						add_string_to_string(mirror, m2);
-					}
-					if (m2)	done_string(m2);
-					goto test_end;
-				}
-			}
+			/* Mirror what we already have */
+			if (mirror) add_bytes_to_string(mirror, orig_pos,
+							file - orig_pos);
 		}
 
-		err = ERROR_COMMAND;
-		orig_pos = file;
-		/* Jump over this crap we can't understand. */
-		while (!WHITECHAR(*file) && *file != '#' && *file)
-			file++;
-
-		/* Mirror what we already have */
-		if (mirror) add_bytes_to_string(mirror, orig_pos, file - orig_pos);
-
-test_end:
-
 		if (!mirror && err) {
 			/* TODO: Make this a macro and report error directly
 			 * as it's stumbled upon; line info may not be accurate
 			 * anymore now (?). --pasky */
-			fprintf(stderr, "%s:%d: %s\n",
-				name, line, error_msg[err]);
-			error_occured = 1;
+			if (verbose >= VERBOSE_WARNINGS) {
+				fprintf(stderr, "%s:%d: %s\n",
+					name, line, error_msg[err]);
+				error_occured = 1;
+			}
 			err = 0;
 		}
 	}
@@ -452,7 +514,8 @@ load_config_from(unsigned char *file, struct option *tree)
 void
 load_config(void)
 {
-	load_config_from(get_opt_str_tree(cmdline_options, "conffile"), config_options);
+	load_config_from(get_cmd_opt_str("config-file"),
+			 config_options);
 }
 
 
@@ -468,6 +531,14 @@ conf_i18n(unsigned char *s, int i18n)
 	return s;
 }
 
+static void
+add_indent_to_string(struct string *string, int depth)
+{
+	if (!depth) return;
+
+	add_xchar_to_string(string, ' ', depth * indentation);
+}
+
 static void
 smart_config_output_fn(struct string *string, struct option *option,
 		       unsigned char *path, int depth, int do_print_comment,
@@ -480,14 +551,15 @@ smart_config_output_fn(struct string *string, struct option *option,
 
 	/* XXX: OPT_LANGUAGE shouldn't have any bussiness here, but we're just
 	 * weird in that area. */
-	if (touching && !(option->flags & OPT_TOUCHED) && option->type != OPT_LANGUAGE)
+	if (touching && !(option->flags & OPT_TOUCHED)
+	    && option->type != OPT_LANGUAGE)
 		return;
 
 	switch (action) {
 		case 0:
 			if (!(comments & 1)) break;
-			if (depth)
-				add_xchar_to_string(string, ' ', depth * indentation);
+
+			add_indent_to_string(string, depth);
 
 			add_to_string(string, "## ");
 			if (path) {
@@ -508,15 +580,16 @@ smart_config_output_fn(struct string *string, struct option *option,
 
 			desc_i18n = conf_i18n(option->desc, i18n);
 
-			if (depth)
-				add_xchar_to_string(string, ' ', depth * indentation);
-			add_to_string(string, "# ");
+			if (!*desc_i18n) break;
+
+			add_indent_to_string(string, depth);
+			add_to_string(string, "#  ");
 			{
 				unsigned char *i = desc_i18n;
 				unsigned char *j = i;
 				unsigned char *last_space = NULL;
 				int config_width = 80;
-				int n = depth * indentation + 2;
+				int n = depth * indentation + 3;
 
 				for (; *i; i++, n++) {
 					if (*i == '\n') {
@@ -530,11 +603,10 @@ smart_config_output_fn(struct string *string, struct option *option,
 split:
 						add_bytes_to_string(string, j, last_space - j);
 						add_char_to_string(string, '\n');
-						if (depth)
-							add_xchar_to_string(string, ' ', depth * indentation);
-						add_to_string(string, "# ");
+						add_indent_to_string(string, depth);
+						add_to_string(string, "#  ");
 						j = last_space + 1;
-						n = depth * indentation + 1 + i - last_space;
+						n = depth * indentation + 2 + i - last_space;
 						last_space = NULL;
 					}
 				}
@@ -544,20 +616,25 @@ split:
 			break;
 
 		case 2:
-			if (depth)
-				add_xchar_to_string(string, ' ', depth * indentation);
+
+			add_indent_to_string(string, depth);
+			if (option->flags & OPT_DELETED) {
+				add_to_string(string, "un");
+			}
 			add_to_string(string, "set ");
 			if (path) {
 				add_to_string(string, path);
 				add_char_to_string(string, '.');
 			}
 			add_to_string(string, option->name);
-			add_to_string(string, " = ");
-			/* OPT_ALIAS won't ever. OPT_TREE won't reach action 2.
-			 * OPT_SPECIAL makes no sense in the configuration
-			 * context. */
-			assert(option_types[option->type].write);
-			option_types[option->type].write(option, string);
+			if (!(option->flags & OPT_DELETED)) {
+				add_to_string(string, " = ");
+				/* OPT_ALIAS won't ever. OPT_TREE won't reach action 2.
+				 * OPT_SPECIAL makes no sense in the configuration
+				 * context. */
+				assert(option_types[option->type].write);
+				option_types[option->type].write(option, string);
+			}
 			add_char_to_string(string, '\n');
 			if (do_print_comment) add_char_to_string(string, '\n');
 			break;
@@ -569,6 +646,21 @@ split:
 	}
 }
 
+
+static void
+add_cfg_header_to_string(struct string *string, unsigned char *text)
+{
+	int n = strlen(text) + 2;
+
+	int_bounds(&n, 10, 80);
+
+	add_to_string(string, "\n\n\n");
+	add_xchar_to_string(string, '#', n);
+	add_to_string(string, "\n# ");
+	add_to_string(string, text);
+	add_to_string(string, "#\n\n");
+}
+
 static unsigned char *
 create_config_string(unsigned char *prefix, unsigned char *name,
 		     struct option *options)
@@ -578,7 +670,7 @@ create_config_string(unsigned char *prefix, unsigned char *name,
 	struct string tmpstring;
 	int origlen;
 	int savestyle = get_opt_int("config.saving_style");
-	int i18n = get_opt_int("config.i18n");
+	int i18n = get_opt_bool("config.i18n");
 
 	if (!init_string(&config)) return NULL;
 
@@ -589,13 +681,16 @@ create_config_string(unsigned char *prefix, unsigned char *name,
 		touching = 0;
 	}
 
+	if (savestyle == 2) watermark_deleted_options(options->value.tree);
+
 	/* Scaring. */
 	if (savestyle == 2
 	    || (savestyle < 2
 		&& (load_config_file(prefix, name, options, &config)
 		    || !config.length))) {
-		/* At first line, and in english, write ELinks version, may be
-		 * of some help in future. Please keep that format for it. --Zas */
+		/* At first line, and in English, write ELinks version, may be
+		 * of some help in future. Please keep that format for it.
+		 * --Zas */
 		add_to_string(&config, "## ELinks " VERSION " configuration file\n\n");
 		assert(savestyle >= 0  && savestyle <= 2);
 		switch (savestyle) {
@@ -642,25 +737,20 @@ create_config_string(unsigned char *prefix, unsigned char *name,
 
 	if (!init_string(&tmpstring)) goto get_me_out;
 
-	add_to_string(&tmpstring, "\n\n\n");
-	add_to_string(&tmpstring, "#####################################\n");
-	add_to_string(&tmpstring, "# ");
-	add_to_string(&tmpstring, conf_i18n(N_("Automatically saved options\n"), i18n));
-	add_to_string(&tmpstring, "#\n\n");
+	add_cfg_header_to_string(&tmpstring,
+				 conf_i18n(N_("Automatically saved options\n"), i18n));
 
 	origlen = tmpstring.length;
-	smart_config_string(&tmpstring, 2, i18n, options->value.tree, NULL, 0, smart_config_output_fn);
+	smart_config_string(&tmpstring, 2, i18n, options->value.tree, NULL, 0,
+			    smart_config_output_fn);
 	if (tmpstring.length > origlen)
 		add_bytes_to_string(&config, tmpstring.source, tmpstring.length);
 	done_string(&tmpstring);
 
 	if (!init_string(&tmpstring)) goto get_me_out;
 
-	add_to_string(&tmpstring, "\n\n\n");
-	add_to_string(&tmpstring, "#####################################\n");
-	add_to_string(&tmpstring, "# ");
-	add_to_string(&tmpstring, conf_i18n(N_("Automatically saved keybindings\n"), i18n));
-	add_to_string(&tmpstring, "#\n\n");
+	add_cfg_header_to_string(&tmpstring,
+				 conf_i18n(N_("Automatically saved keybindings\n"), i18n));
 
 	origlen = tmpstring.length;
 	bind_config_string(&tmpstring);
@@ -674,30 +764,24 @@ get_me_out:
 	return config.source;
 }
 
-/* TODO: The error condition should be handled somewhere else. */
 static int
 write_config_file(unsigned char *prefix, unsigned char *name,
 		  struct option *options, struct terminal *term)
 {
 	int ret = -1;
 	struct secure_save_info *ssi;
-	unsigned char *config_file;
+	unsigned char *config_file = NULL;
 	unsigned char *cfg_str = create_config_string(prefix, name, options);
 	int prefixlen = strlen(prefix);
-	int prefix_has_slash, name_has_slash;
+	int prefix_has_slash = (prefixlen && dir_sep(prefix[prefixlen - 1]));
+	int name_has_slash = dir_sep(name[0]);
+	unsigned char *slash = name_has_slash || prefix_has_slash ? "" : "/";
 
 	if (!cfg_str) return -1;
 
-	name_has_slash = (name[0] && dir_sep(name[0]));
-	prefix_has_slash = (prefixlen && dir_sep(prefix[prefixlen - 1]));
-	if (name_has_slash || prefix_has_slash) {
-		if (name_has_slash && prefix_has_slash)
-			name++;
-		config_file = straconcat(prefix, name, NULL);
-	} else {
-		config_file = straconcat(prefix, "/", name, NULL);
-	}
+	if (name_has_slash && prefix_has_slash) name++;
 
+	config_file = straconcat(prefix, slash, name, NULL);
 	if (!config_file) goto free_cfg_str;
 
 	ssi = secure_open(config_file, 0177);
@@ -706,55 +790,8 @@ write_config_file(unsigned char *prefix, unsigned char *name,
 		ret = secure_close(ssi);
 	}
 
-	if (term && (secsave_errno != SS_ERR_NONE || ret)) {
-		unsigned char *errmsg = NULL;
-		unsigned char *strerr = _("Secure file error", term);
-
-		switch (secsave_errno) {
-			case SS_ERR_NONE: /* Impossible. */
-				break;
-			case SS_ERR_OPEN_READ:
-				strerr = _("Cannot read the file", term);
-	  			break;
-			case SS_ERR_STAT:
-				strerr = _("Cannot stat the file", term);
-	  			break;
-			case SS_ERR_ACCESS:
-				strerr = _("Cannot access the file", term);
-	  			break;
-			case SS_ERR_MKSTEMP:
-				strerr = _("Cannot create temp file", term);
-	  			break;
-			case SS_ERR_RENAME:
-				strerr = _("Cannot rename the file", term);
-	  			break;
-			case SS_ERR_DISABLED:
-				strerr = _("File saving disabled by option", term);
-				break;
-			case SS_ERR_OUT_OF_MEM:
-				strerr = _("Out of memory", term);
-				break;
-			case SS_ERR_OPEN_WRITE:
-				strerr = _("Cannot write the file", term);
-				break;
-			case SS_ERR_OTHER:
-				break;
-		}
-
-		if (ret > 0) errmsg = straconcat(strerr, " (", strerror(ret), ")", NULL);
-
-		if (errmsg) {
-			write_config_error(term, getml(config_file, errmsg, NULL),
-					   config_file, errmsg);
-		} else {
-			write_config_error(term, getml(config_file, NULL),
-					   config_file, strerr);
-		}
-
-		goto free_cfg_str;
-	}
-
-	write_config_success(term, getml(config_file, NULL), config_file);
+	write_config_dialog(term, config_file, secsave_errno, ret);
+	mem_free(config_file);
 
 free_cfg_str:
 	mem_free(cfg_str);
@@ -762,16 +799,17 @@ free_cfg_str:
 	return ret;
 }
 
-static int
-write_config_to(unsigned char *file, struct option *tree,
-		struct terminal *term)
-{
-	return write_config_file(elinks_home, file, tree, term);
-}
-
 int
 write_config(struct terminal *term)
 {
-	return write_config_to(get_opt_str_tree(cmdline_options, "conffile"),
-			       config_options, term);
+	assert(term);
+
+	if (!elinks_home) {
+		write_config_dialog(term, get_cmd_opt_str("config-file"),
+				    SS_ERR_DISABLED, 0);
+		return -1;
+	}
+
+	return write_config_file(elinks_home, get_cmd_opt_str("config-file"),
+				 config_options, term);
 }
