@@ -1,5 +1,5 @@
 /* Options variables manipulation core */
-/* $Id: options.c,v 1.431 2004/01/08 17:14:41 zas Exp $ */
+/* $Id: options.c,v 1.468.2.4 2005/04/05 21:08:40 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -10,27 +10,28 @@
 
 #include "elinks.h"
 
-#include "main.h" /* shrink_memory() */
-#include "bfu/listbox.h"
+#include "bfu/dialog.h"
+#include "cache/cache.h"
 #include "config/conf.h"
 #include "config/dialogs.h"
 #include "config/options.h"
 #include "config/opttypes.h"
-#include "cache/cache.h"
 #include "dialogs/status.h"
 #include "document/options.h"
 #include "globhist/globhist.h"
 #include "intl/charsets.h"
 #include "intl/gettext/libintl.h"
 #include "lowlevel/select.h"
+#include "main.h" /* shrink_memory() */
 #include "sched/session.h"
+#include "terminal/color.h"
 #include "terminal/screen.h"
 #include "terminal/terminal.h"
 #include "util/color.h"
 #include "util/error.h"
 #include "util/memory.h"
 #include "util/string.h"
-#include "viewer/text/view.h"
+#include "viewer/text/draw.h"
 
 
 /* TODO? In the past, covered by shadow and legends, remembered only by the
@@ -69,13 +70,13 @@ struct option *cmdline_options;
 static void add_opt_rec(struct option *, unsigned char *, struct option *);
 static void free_options_tree(struct list_head *, int recursive);
 
-#ifdef DEBUG
+#ifdef CONFIG_DEBUG
 /* Detect ending '.' (and some others) in options captions.
  * It will emit a message in debug mode only. --Zas */
 
-#define bad_punct(c) (c != ')' && c != '\'' && c != '"' && ispunct(c))
-	
-void
+#define bad_punct(c) (c != ')' && !isquote(c) && ispunct(c))
+
+static void
 check_caption(unsigned char *caption)
 {
 	int len;
@@ -102,8 +103,46 @@ check_caption(unsigned char *caption)
 }
 
 #undef bad_punct
+
+static void
+check_description(unsigned char *desc)
+{
+	int len;
+	unsigned char c;
+
+	if (!desc) return;
+
+	len = strlen(desc);
+	if (!len) return;
+
+	c = desc[len - 1];
+	if (isspace(c))
+		DBG("bad char at end of description [%s]", desc);
+
+#ifdef ENABLE_NLS
+	desc = gettext(desc);
+	len = strlen(desc);
+	if (!len) return;
+
+	if (ispunct(c) != ispunct(desc[len - 1]))
+		DBG("punctuation char possibly missing at end of i18n description [%s]", desc);
+
+	c = desc[len - 1];
+	if (isspace(c))
+		DBG("bad char at end of i18n description [%s]", desc);
+#endif
+}
+
+static void
+debug_check_option_syntax(struct option *option)
+{
+	if (!option) return;
+	check_caption(option->capt);
+	check_description(option->desc);
+}
+
 #else
-#define check_caption(caption)
+#define debug_check_option_syntax(option)
 #endif
 
 
@@ -176,9 +215,7 @@ get_opt_rec(struct option *tree, unsigned char *name_)
 			mem_free(aname);
 			return NULL;
 		}
-		if (option->name) mem_free(option->name);
-		option->name = stracpy(name);
-		if (option->box_item) option->box_item->text = option->name;
+		mem_free_set(&option->name, stracpy(name));
 
 		add_opt_rec(tree, "", option);
 
@@ -208,19 +245,25 @@ get_opt_rec_real(struct option *tree, unsigned char *name)
  * NULL. Note that you are supposed to use wrapper get_opt(). */
 union option_value *
 get_opt_(
-#ifdef DEBUG
-	 unsigned char *file, int line,
+#ifdef CONFIG_DEBUG
+	 unsigned char *file, int line, enum option_type option_type,
 #endif
 	 struct option *tree, unsigned char *name)
 {
 	struct option *opt = get_opt_rec(tree, name);
 
-#ifdef DEBUG
+#ifdef CONFIG_DEBUG
 	errfile = file;
 	errline = line;
 	if (!opt) elinks_internal("Attempted to fetch nonexisting option %s!", name);
 
 	/* Various sanity checks. */
+	if (option_type != opt->type)
+		DBG("get_opt_*(\"%s\") @ %s:%d: call with wrapper for %s for option of type %s",
+		    name, file, line,
+		    get_option_type_name(option_type),
+		    get_option_type_name(opt->type));
+
 	switch (opt->type) {
 	case OPT_TREE:
 		if (!opt->value.tree)
@@ -255,6 +298,86 @@ get_opt_(
 	return &opt->value;
 }
 
+static void
+add_opt_sort(struct option *tree, struct option *option, int abi)
+{
+	struct list_head *cat = tree->value.tree;
+	struct list_head *bcat = &tree->box_item->child;
+	struct option *pos;
+
+	/* The list is empty, just add it there. */
+	if (list_empty(*cat)) {
+		add_to_list(*cat, option);
+		if (abi) add_to_list(*bcat, option->box_item);
+
+	/* This fits as the last list entry, add it there. This
+	 * optimizes the most expensive BUT most common case ;-). */
+	} else if ((option->type != OPT_TREE
+		    || ((struct option *) cat->prev)->type == OPT_TREE)
+		   && strcmp(((struct option *) cat->prev)->name,
+			     option->name) <= 0) {
+append:
+		add_to_list_end(*cat, option);
+		if (abi) add_to_list_end(*bcat, option->box_item);
+
+	/* At the end of the list is tree and we are ordinary. That's
+	 * clear case then. */
+	} else if (option->type != OPT_TREE
+		   && ((struct option *) cat->prev)->type == OPT_TREE) {
+		goto append;
+
+	/* Scan the list linearly. This could be probably optimized ie.
+	 * to choose direction based on the first letter or so. */
+	} else {
+		struct listbox_item *bpos = (struct listbox_item *) bcat;
+
+		foreach (pos, *cat) {
+			/* First move the box item to the current position but
+			 * only if the position has not been marked as deleted
+			 * and actually has a box_item -- else we will end up
+			 * 'overflowing' and causing assertion failure. */
+			if (!(pos->flags & OPT_DELETED) && pos->box_item) {
+				bpos = bpos->next;
+				assert(bpos != (struct listbox_item *) bcat);
+			}
+
+			if ((option->type != OPT_TREE
+			     || pos->type == OPT_TREE)
+			    && strcmp(pos->name, option->name) <= 0)
+				continue;
+
+			/* Ordinary options always sort behind trees. */
+			if (option->type != OPT_TREE
+			    && pos->type == OPT_TREE)
+				continue;
+
+			/* The (struct option) add_at_pos() can mess up the
+			 * order so that we add the box_item to itself, so
+			 * better do it first. */
+
+			/* Always ensure that them _template_ options are
+			 * before anything else so a lonely autocreated option
+			 * (with _template_ options set to invisible) will be
+			 * connected with an upper corner (ascii: `-) instead
+			 * of a rotated T (ascii: +-) when displaying it. */
+			if (option->type == pos->type
+			    && *option->name <= '_'
+			    && !strcmp(pos->name, "_template_")) {
+				if (abi) add_at_pos(bpos, option->box_item);
+				add_at_pos(pos, option);
+				break;
+			}
+
+			if (abi) add_at_pos(bpos->prev, option->box_item);
+			add_at_pos(pos->prev, option);
+			break;
+		}
+
+		assert(pos != (struct option *) cat);
+		assert(bpos != (struct listbox_item *) bcat);
+	}
+}
+
 /* Add option to tree. */
 static void
 add_opt_rec(struct option *tree, unsigned char *path, struct option *option)
@@ -267,8 +390,10 @@ add_opt_rec(struct option *tree, unsigned char *path, struct option *option)
 	assertm(tree, "Missing option tree for '%s'", path);
 	if (!tree->value.tree) return;
 
+	object_nolock(option, "option");
+
 	if (option->box_item && option->name && !strcmp(option->name, "_template_"))
-		option->box_item->visible = get_opt_int("config.show_template");
+		option->box_item->visible = get_opt_bool("config.show_template");
 
 	if (tree->flags & OPT_AUTOCREATE && !option->desc) {
 		struct option *template = get_opt_rec(tree, "_template_");
@@ -277,6 +402,8 @@ add_opt_rec(struct option *tree, unsigned char *path, struct option *option)
 		option->desc = template->desc;
 	}
 
+	option->root = tree;
+
 	abi = (tree->box_item && option->box_item);
 
 	if (abi) {
@@ -286,69 +413,11 @@ add_opt_rec(struct option *tree, unsigned char *path, struct option *option)
 		 * NULL @next. */
 		if (tree->box_item->next) {
 			option->box_item->depth = tree->box_item->depth + 1;
-			option->box_item->root = tree->box_item;
 		}
 	}
 
 	if (tree->flags & OPT_SORT) {
-		struct list_head *cat = tree->value.tree;
-		struct list_head *bcat = &tree->box_item->child;
-		struct option *pos;
-
-		/* The list is empty, just add it there. */
-		if (list_empty(*cat)) {
-			add_to_list(*cat, option);
-			if (abi) add_to_list(*bcat, option->box_item);
-
-		/* This fits as the last list entry, add it there. This
-		 * optimizes the most expensive BUT most common case ;-). */
-		} else if ((option->type != OPT_TREE
-			    || ((struct option *) cat->prev)->type == OPT_TREE)
-			   && strcmp(((struct option *) cat->prev)->name,
-				     option->name) <= 0) {
-append:
-			add_to_list_end(*cat, option);
-			if (abi) add_to_list_end(*bcat, option->box_item);
-
-		/* At the end of the list is tree and we are ordinary. That's
-		 * clear case then. */
-		} else if (option->type != OPT_TREE
-			   && ((struct option *) cat->prev)->type == OPT_TREE) {
-			goto append;
-
-		/* Scan the list linearly. This could be probably optimized ie.
-		 * to choose direction based on the first letter or so. */
-		} else {
-			struct listbox_item *bpos = bcat->next;
-
-			assert(bpos != (struct listbox_item *) bcat);
-
-			foreach (pos, *cat) {
-				if ((option->type != OPT_TREE
-				     || pos->type == OPT_TREE)
-				    && strcmp(pos->name, option->name) <= 0) {
-next:
-					bpos = bpos->next;
-					assert(bpos != (struct listbox_item *) bcat);
-					continue;
-				}
-
-				/* Ordinary options always sort behind trees. */
-				if (option->type != OPT_TREE
-				    && pos->type == OPT_TREE)
-					goto next;
-
-				/* The (struct option) add_at_pos() can mess
-				 * up the order so that we add the box_item
-				 * to itself, so better do it first. */
-				if (abi) add_at_pos(bpos->prev, option->box_item);
-				add_at_pos(pos->prev, option);
-				break;
-			}
-
-			assert(pos != (struct option *) cat);
-			assert(bpos != (struct listbox_item *) bcat);
-		}
+		add_opt_sort(tree, option, abi);
 
 	} else {
 		add_to_list_end(*tree->value.tree, option);
@@ -361,14 +430,12 @@ next:
 static inline struct listbox_item *
 init_option_listbox_item(struct option *option)
 {
-	struct listbox_item *box = mem_calloc(1, sizeof(struct listbox_item));
+	struct listbox_item *box = mem_calloc(1, sizeof(*box));
 
 	if (!box) return NULL;
 
 	init_list(box->child);
 	box->visible = 1;
-	box->translated = 1;
-	box->text = option->capt ? option->capt : option->name;
 	box->udata = option;
 	box->type = (option->type == OPT_TREE) ? BI_FOLDER : BI_LEAF;
 
@@ -380,7 +447,7 @@ add_opt(struct option *tree, unsigned char *path, unsigned char *capt,
 	unsigned char *name, enum option_flags flags, enum option_type type,
 	int min, int max, void *value, unsigned char *desc)
 {
-	struct option *option = mem_calloc(1, sizeof(struct option));
+	struct option *option = mem_calloc(1, sizeof(*option));
 
 	if (!option) return NULL;
 
@@ -396,7 +463,7 @@ add_opt(struct option *tree, unsigned char *path, unsigned char *capt,
 	option->capt = capt;
 	option->desc = desc;
 
-	check_caption(option->capt);
+	debug_check_option_syntax(option);
 
 	if (option->type != OPT_ALIAS
 	    && ((tree->flags & OPT_LISTBOX) || (option->flags & OPT_LISTBOX))) {
@@ -434,7 +501,8 @@ add_opt(struct option *tree, unsigned char *path, unsigned char *capt,
 			option->value.number = (long) value;
 			break;
 		case OPT_COLOR:
-			decode_color(value, &option->value.color);
+			decode_color(value, strlen((unsigned char *) value),
+					&option->value.color);
 			break;
 		case OPT_COMMAND:
 			option->value.command = value;
@@ -454,7 +522,7 @@ add_opt(struct option *tree, unsigned char *path, unsigned char *capt,
  * completely filled (struct option *) have long name and functions which take
  * only option specs have short name. */
 
-void
+static void
 delete_option_do(struct option *option, int recursive)
 {
 	if (option->next) del_from_list(option);
@@ -465,8 +533,7 @@ delete_option_do(struct option *option, int recursive)
 
 	switch (option->type) {
 		case OPT_STRING:
-			if (option->value.string)
-				mem_free(option->value.string);
+			mem_free_if(option->value.string);
 			break;
 		case OPT_TREE:
 			if (!option->value.tree) break;
@@ -491,16 +558,36 @@ delete_option_do(struct option *option, int recursive)
 		done_listbox_item(&option_browser, option->box_item);
 
 	if (option->flags & OPT_ALLOC) {
-		if (option->name) mem_free(option->name);
+		mem_free_if(option->name);
 		mem_free(option);
 	} else if (!option->capt) {
 		/* We are probably dealing with a built-in autocreated option
-		 * that will be attempted to be deleted when shutting down.
-		 * Clear it so nothing will be done later. */
-		memset(option, 0, sizeof(struct option));
+		 * that will be attempted to be deleted when shutting down. */
+		/* Clear it so nothing will be done later. */
+		memset(option, 0, sizeof(*option));
 	}
 }
 
+void
+mark_option_as_deleted(struct option *option)
+{
+	if (option->type == OPT_TREE) {
+		struct option *unmarked;
+
+		assert(option->value.tree);
+
+		foreach (unmarked, *option->value.tree)
+			mark_option_as_deleted(unmarked);
+	}
+
+	if (option->box_item) {
+		done_listbox_item(&option_browser, option->box_item);
+		option->box_item = NULL;
+	}
+
+	option->flags |= (OPT_TOUCHED | OPT_DELETED);
+}
+
 void
 delete_option(struct option *option)
 {
@@ -510,7 +597,7 @@ delete_option(struct option *option)
 struct option *
 copy_option(struct option *template)
 {
-	struct option *option = mem_calloc(1, sizeof(struct option));
+	struct option *option = mem_calloc(1, sizeof(*option));
 
 	if (!option) return NULL;
 
@@ -543,7 +630,7 @@ copy_option(struct option *template)
 struct list_head *
 init_options_tree(void)
 {
-	struct list_head *ptr = mem_alloc(sizeof(struct list_head));
+	struct list_head *ptr = mem_alloc(sizeof(*ptr));
 
 	if (ptr) init_list(*ptr);
 	return ptr;
@@ -555,31 +642,20 @@ register_autocreated_options(void)
 {
 	/* TODO: Use table-driven initialization. --jonas */
 	get_opt_int("terminal.linux.type") = 2;
-	get_opt_bool("terminal.linux.colors") = 1;
+	get_opt_int("terminal.linux.colors") = 1;
 	get_opt_bool("terminal.linux.m11_hack") = 1;
 	get_opt_int("terminal.vt100.type") = 1;
 	get_opt_int("terminal.vt110.type") = 1;
 	get_opt_int("terminal.xterm.type") = 1;
-	get_opt_int("terminal.xterm.underline") = 1;
+	get_opt_bool("terminal.xterm.underline") = 1;
 	get_opt_int("terminal.xterm-color.type") = 1;
-	get_opt_bool("terminal.xterm-color.colors") = 1;
-	get_opt_int("terminal.xterm-color.underline") = 1;
+	get_opt_int("terminal.xterm-color.colors") = COLOR_MODE_16;
+	get_opt_bool("terminal.xterm-color.underline") = 1;
+#ifdef CONFIG_256_COLORS
 	get_opt_int("terminal.xterm-256color.type") = 1;
-	get_opt_bool("terminal.xterm-256color.colors") = 2;
-	get_opt_int("terminal.xterm-256color.underline") = 1;
-
-	strcpy(get_opt_str("protocol.user.mailto.unix"), DEFAULT_AC_OPT_MAILTO);
-	strcpy(get_opt_str("protocol.user.mailto.unix-xwin"),DEFAULT_AC_OPT_MAILTO);
-	strcpy(get_opt_str("protocol.user.telnet.unix"), DEFAULT_AC_OPT_TELNET);
-	strcpy(get_opt_str("protocol.user.telnet.unix-xwin"), DEFAULT_AC_OPT_TELNET);
-	strcpy(get_opt_str("protocol.user.tn3270.unix"), DEFAULT_AC_OPT_TN3270);
-	strcpy(get_opt_str("protocol.user.tn3270.unix-xwin"), DEFAULT_AC_OPT_TN3270);
-	strcpy(get_opt_str("protocol.user.gopher.unix"), DEFAULT_AC_OPT_GOPHER);
-	strcpy(get_opt_str("protocol.user.gopher.unix-xwin"), DEFAULT_AC_OPT_GOPHER);
-	strcpy(get_opt_str("protocol.user.news.unix"), DEFAULT_AC_OPT_NEWS);
-	strcpy(get_opt_str("protocol.user.news.unix-xwin"), DEFAULT_AC_OPT_NEWS);
-	strcpy(get_opt_str("protocol.user.irc.unix"), DEFAULT_AC_OPT_IRC);
-	strcpy(get_opt_str("protocol.user.irc.unix-xwin"), DEFAULT_AC_OPT_IRC);
+	get_opt_int("terminal.xterm-256color.colors") = COLOR_MODE_256;
+	get_opt_bool("terminal.xterm-256color.underline") = 1;
+#endif
 }
 
 static struct option_info config_options_info[];
@@ -645,6 +721,19 @@ unmark_options_tree(struct list_head *tree)
 	}
 }
 
+void
+watermark_deleted_options(struct list_head *tree)
+{
+	struct option *option;
+
+	foreach (option, *tree) {
+		if (option->flags & OPT_DELETED)
+			option->flags |= OPT_WATERMARK;
+		else if (option->type == OPT_TREE)
+			watermark_deleted_options(option->value.tree);
+	}
+}
+
 static int
 check_nonempty_tree(struct list_head *options)
 {
@@ -676,7 +765,7 @@ smart_config_string(struct string *str, int print_comment, int i18n,
 		if (option->flags & OPT_HIDDEN ||
 		    option->flags & OPT_WATERMARK ||
 		    option->type == OPT_ALIAS ||
-		    (option->box_item && !option->box_item->visible) /* _template_ */)
+		    !strcmp(option->name, "_template_"))
 			continue;
 
 		/* Is there anything to be printed anyway? */
@@ -779,6 +868,13 @@ change_hook_html(struct session *ses, struct option *current, struct option *cha
 	return 0;
 }
 
+static int
+change_hook_insert_mode(struct session *ses, struct option *current, struct option *changed)
+{
+	update_status();
+	return 0;
+}
+
 static int
 change_hook_active_link(struct session *ses, struct option *current, struct option *changed)
 {
@@ -832,6 +928,13 @@ update_visibility(struct list_head *tree, int show)
 	}
 }
 
+void
+update_options_visibility(void)
+{
+	update_visibility(config_options->value.tree,
+			  get_opt_bool("config.show_template"));
+}
+
 static int
 change_hook_stemplate(struct session *ses, struct option *current, struct option *changed)
 {
@@ -852,6 +955,8 @@ static struct change_hook_info change_hooks[] = {
 	{ "config.show_template",	change_hook_stemplate },
 	{ "connection",			change_hook_connection },
 	{ "document.browse",		change_hook_html },
+	{ "document.browse.forms.insert_mode",
+					change_hook_insert_mode },
 	{ "document.browse.links.active_link",
 					change_hook_active_link },
 	{ "document.cache",		change_hook_cache },
@@ -865,6 +970,45 @@ static struct change_hook_info change_hooks[] = {
 	{ NULL,				NULL },
 };
 
+int
+commit_option_values(struct option_resolver *resolvers,
+		     struct option *root, union option_value *values, int size)
+{
+	int touched = 0;
+	int i;
+
+	assert(resolvers && root && values && size);
+
+	for (i = 0; i < size; i++) {
+		unsigned char *name = resolvers[i].name;
+		struct option *option = get_opt_rec(root, name);
+		int id = resolvers[i].id;
+
+		if (memcmp(&option->value, &values[id], sizeof(union option_value))) {
+			option->value = values[id];
+			option->flags |= OPT_TOUCHED;
+			touched++;
+		}
+	}
+
+	return touched;
+}
+
+void
+checkout_option_values(struct option_resolver *resolvers,
+		       struct option *root,
+		       union option_value *values, int size)
+{
+	int i;
+
+	for (i = 0; i < size; i++) {
+		unsigned char *name = resolvers[i].name;
+		struct option *option = get_opt_rec(root, name);
+		int id = resolvers[i].id;
+
+		values[id] = option->value;
+	}
+}
 
 /**********************************************************************
  Options values
@@ -881,7 +1025,7 @@ register_options(struct option_info info[], struct option *tree)
 		struct option *option = &info[i].option;
 		unsigned char *string;
 
-		check_caption(option->capt);
+		debug_check_option_syntax(option);
 
 		if (option->type != OPT_ALIAS
 		    && ((tree->flags & OPT_LISTBOX)
@@ -913,7 +1057,8 @@ register_options(struct option_info info[], struct option *tree)
 			case OPT_COLOR:
 				string = option->value.string;
 				assert(string);
-				decode_color(string, &option->value.color);
+				decode_color(string, strlen(string),
+						&option->value.color);
 				break;
 			case OPT_CODEPAGE:
 				string = option->value.string;
