@@ -1,12 +1,10 @@
 /* Internal cookies implementation */
-/* $Id: cookies.c,v 1.118 2004/01/03 13:12:16 jonas Exp $ */
+/* $Id: cookies.c,v 1.192.2.3 2005/04/06 09:28:22 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#ifdef CONFIG_COOKIES
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -18,9 +16,11 @@
 
 #include "elinks.h"
 
-/* #define COOKIES_DEBUG */
+#if 0
+#define DEBUG_COOKIES
+#endif
 
-#include "bfu/msgbox.h"
+#include "bfu/dialog.h"
 #include "cookies/cookies.h"
 #include "cookies/dialogs.h"
 #include "cookies/parser.h"
@@ -29,14 +29,14 @@
 #include "intl/gettext/libintl.h"
 #include "lowlevel/home.h"
 #include "modules/module.h"
-#include "protocol/http/date.h"
-#include "protocol/http/header.h"
+#include "protocol/date.h"
+#include "protocol/header.h"
 #include "protocol/protocol.h"
 #include "protocol/uri.h"
 #include "sched/session.h"
 #include "terminal/terminal.h"
 #include "util/conv.h"
-#ifdef COOKIES_DEBUG
+#ifdef DEBUG_COOKIES
 #include "util/error.h"
 #endif
 #include "util/file.h"
@@ -51,10 +51,7 @@
 
 static int cookies_nosave = 0;
 
-static unsigned int cookie_id = 0;
-
 static INIT_LIST_HEAD(cookies);
-static INIT_LIST_HEAD(cookie_queries);
 
 struct c_domain {
 	LIST_HEAD(struct c_domain);
@@ -64,14 +61,7 @@ struct c_domain {
 
 static INIT_LIST_HEAD(c_domains);
 
-struct c_server {
-	LIST_HEAD(struct c_server);
-
-	int accept;
-	unsigned char server[1]; /* Must be at end of struct. */
-};
-
-static INIT_LIST_HEAD(c_servers);
+static INIT_LIST_HEAD(cookie_servers);
 
 static int cookies_dirty = 0;
 
@@ -104,23 +94,27 @@ static struct option_info cookies_options[] = {
 		"max_age", 0, -1, 10000, -1,
 		N_("Cookie maximum age (in days):\n"
 		"-1 is use cookie's expiration date if any\n"
-		"0  is force expiration at the end of session, ignoring cookie's expiration date\n"
-		"1+ is use cookie's expiration date, but limit age to the given number of days")),
+		"0  is force expiration at the end of session, ignoring cookie's\n"
+		"   expiration date\n"
+		"1+ is use cookie's expiration date, but limit age to the given\n"
+		"   number of days")),
 
 	INIT_OPT_BOOL("cookies", N_("Paranoid security"),
 		"paranoid_security", 0, 0,
 		N_("When enabled, we'll require three dots in cookies domain for all\n"
-		"non-international domains (instead of just two dots). Please see\n"
-		"code (cookies.c:check_domain_security()) for explanation.")),
+		"non-international domains (instead of just two dots). Some countries\n"
+		"have generic second level domains (eg. .com.pl, .co.uk) and allowing\n"
+		"sites to set cookies for these generic domains could potentially be\n"
+		"very bad. Note, it is off by default as it breaks a lot of sites.")),
 
 	INIT_OPT_BOOL("cookies", N_("Saving"),
 		"save", 0, 1,
-		N_("Load/save cookies from/to disk?")),
+		N_("Whether cookies should be loaded from and save to disk.")),
 
 	INIT_OPT_BOOL("cookies", N_("Resaving"),
 		"resave", 0, 1,
 		N_("Save cookies after each change in cookies list? No effect when\n"
-		"cookies_save is off.")),
+		"cookie saving (cookies.save) is off.")),
 
 	NULL_OPTION_INFO,
 };
@@ -132,26 +126,88 @@ static struct option_info cookies_options[] = {
 #define get_cookies_save()		get_opt_cookies(COOKIES_SAVE).number
 #define get_cookies_resave()		get_opt_cookies(COOKIES_RESAVE).number
 
-static void accept_cookie(struct cookie *);
+static struct cookie_server *
+get_cookie_server(unsigned char *host, int hostlen)
+{
+	struct cookie_server *sort_spot = NULL;
+	struct cookie_server *cs;
 
+	foreach (cs, cookie_servers) {
+		/* XXX: We must count with cases like "x.co" vs "x.co.uk"
+		 * below! */
+		int cslen = strlen(cs->host);
+		int cmp = strncasecmp(cs->host, host, hostlen);
+
+		if (!sort_spot && (cmp > 0 || (cmp == 0 && cslen > hostlen))) {
+			/* This is the first @cs with name greater than @host,
+			 * our dream sort spot! */
+			sort_spot = cs->prev;
+		}
+
+		if (cmp || cslen != hostlen)
+			continue;
+
+		object_lock(cs);
+		return cs;
+	}
+
+	cs = mem_calloc(1, sizeof(*cs) + hostlen);
+	if (!cs) return NULL;
+
+	memcpy(cs->host, host, hostlen);
+	object_nolock(cs, "cookie_server");
+
+	cs->box_item = add_listbox_folder(&cookie_browser, NULL, cs);
+
+	object_lock(cs);
+
+	if (!sort_spot) {
+		/* No sort spot found, therefore this sorts at the end. */
+		add_to_list_end(cookie_servers, cs);
+		del_from_list(cs->box_item);
+		add_to_list_end(cookie_browser.root.child, cs->box_item);
+	} else {
+		/* Sort spot found, sort after it. */
+		add_at_pos(sort_spot, cs);
+		if (sort_spot != (struct cookie_server *) &cookie_servers) {
+			del_from_list(cs->box_item);
+			add_at_pos(sort_spot->box_item, cs->box_item);
+		} /* else we are already at the top anyway. */
+	}
+
+	return cs;
+}
+
+static void
+done_cookie_server(struct cookie_server *cs)
+{
+	object_unlock(cs);
+	if (is_object_used(cs)) return;
+
+	if (cs->box_item) done_listbox_item(&cookie_browser, cs->box_item);
+	del_from_list(cs);
+	mem_free(cs);
+}
 
 void
 free_cookie(struct cookie *c)
 {
 	if (c->box_item) done_listbox_item(&cookie_browser, c->box_item);
-	if (c->name) mem_free(c->name);
-	if (c->value) mem_free(c->value);
-	if (c->server) mem_free(c->server);
-	if (c->path) mem_free(c->path);
-	if (c->domain) mem_free(c->domain);
+	if (c->server) done_cookie_server(c->server);
+	mem_free_if(c->name);
+	mem_free_if(c->value);
+	mem_free_if(c->path);
+	mem_free_if(c->domain);
 	mem_free(c);
 }
 
 
+/* Check whether cookie's domain matches server.
+ * It returns 1 if ok, 0 else. */
 static int
-check_domain_security(unsigned char *domain, unsigned char *server, int server_len)
+is_domain_security_ok(unsigned char *domain, unsigned char *server, int server_len)
 {
-	register int i, j;
+	int i;
 	int domain_len;
 	int need_dots;
 
@@ -164,19 +220,25 @@ check_domain_security(unsigned char *domain, unsigned char *server, int server_l
 
 	if (domain_len > server_len) return 0;
 
-	if (!strncasecmp(domain, server, server_len)) {
-		/* We should probably allow domains which are same as servers.
-		 * --<rono@sentuny.com.au> */
-		/* Mozilla does it as well ;))) and I can't figure out any
-		 * security risk. --pasky */
+	/* Ensure that the domain is atleast a substring of the server before
+	 * continuing. */
+	if (strncasecmp(domain, server + server_len - domain_len, domain_len))
 		return 0;
-	}
 
-	for (i = server_len - domain_len, j = 0; domain[j]; i++, j++)
-		if (upcase(server[i]) != upcase(domain[j]))
-			return 0;
+	/* Allow domains which are same as servers. --<rono@sentuny.com.au> */
+	/* Mozilla does it as well ;))) and I can't figure out any security
+	 * risk. --pasky */
+	if (server_len == domain_len)
+		return 1;
 
-	/* Also test if domain is secure enough.. */
+	/* Check whether the server is an IP address, and require an exact host
+	 * match for the cookie, so any chance of IP address funkiness is
+	 * eliminated (e.g. the alias 127.1 domain-matching 99.54.127.1). Idea
+	 * from mozilla. (bug 562) */
+	if (is_ip_address(server, server_len))
+		return 0;
+
+	/* Also test if domain is secure en ugh.. */
 
 	need_dots = 1;
 
@@ -212,37 +274,34 @@ check_domain_security(unsigned char *domain, unsigned char *server, int server_l
 	return 1;
 }
 
-static void accept_cookie_dialog(struct session *ses);
-
 void
 set_cookie(struct uri *uri, unsigned char *str)
 {
-	unsigned char *date, *secure;
+	unsigned char *secure, *path;
 	struct cookie *cookie;
-	struct c_server *cs;
 	struct cookie_str cstr;
+	int max_age;
 
 	if (get_cookies_accept_policy() == COOKIES_ACCEPT_NONE)
 		return;
 
-#ifdef COOKIES_DEBUG
-	DBG("set_cookie -> (%s) %s", struri(*uri), str);
+#ifdef DEBUG_COOKIES
+	DBG("set_cookie -> (%s) %s", struri(uri), str);
 #endif
 
-	cstr.str = str;
-	if (!parse_cookie_str(&cstr)) return;
+	if (!parse_cookie_str(&cstr, str)) return;
 
-	cookie = mem_calloc(1, sizeof(struct cookie));
+	cookie = mem_calloc(1, sizeof(*cookie));
 	if (!cookie) return;
 
-	object_nolock(cookie); /* Debugging purpose. */
+	object_nolock(cookie, "cookie"); /* Debugging purpose. */
 
 	/* Fill main fields */
 
 	cookie->name = memacpy(str, cstr.nam_end - str);
 	cookie->value = memacpy(cstr.val_start, cstr.val_end - cstr.val_start);
-	cookie->server = memacpy(uri->host, uri->hostlen);
-	cookie->domain = parse_http_header_param(str, "domain");
+	cookie->server = get_cookie_server(uri->host, uri->hostlen);
+	cookie->domain = parse_header_param(str, "domain");
 	if (!cookie->domain) cookie->domain = memacpy(uri->host, uri->hostlen);
 
 	/* Now check that all is well */
@@ -254,64 +313,66 @@ set_cookie(struct uri *uri, unsigned char *str)
 		return;
 	}
 
-	/* Get expiration date */
+#if 0
+	/* We don't actually set ->accept at the moment. But I have kept it
+	 * since it will maybe help to fix bug 77 - Support for more
+	 * finegrained control upon accepting of cookies. */
+	if (!cookie->server->accept) {
+#ifdef DEBUG_COOKIES
+		DBG("Dropped.");
+#endif
+		free_cookie(cookie);
+		return;
+	}
+#endif
 
-	date = parse_http_header_param(str, "expires");
-	if (date) {
-		cookie->expires = parse_http_date(date); /* Convert date to seconds. */
+	/* Set cookie expiration if needed.
+	 * Cookie expires at end of session by default,
+	 * set to 0 by calloc().
+	 *
+	 * max_age:
+	 * -1 is use cookie's expiration date if any
+	 * 0  is force expiration at the end of session,
+	 *    ignoring cookie's expiration date
+	 * 1+ is use cookie's expiration date,
+	 *    but limit age to the given number of days.
+	 */
 
-		if (cookie->expires) {
-			int max_age = get_cookies_max_age(); /* Max. age in days */
+	max_age = get_cookies_max_age();
+	if (max_age) {
+		unsigned char *date = parse_header_param(str, "expires");
 
-			if (max_age == 0) cookie->expires = 0; /* Always expires at session end. */
-			else if (max_age > 0) {
-				ttime deadline = time(NULL) + max_age*24*3600 /* days->seconds.*/;
+		if (date) {
+			ttime expires = parse_date(&date, NULL, 0, 1); /* Convert date to seconds. */
 
-				if (cookie->expires > deadline) /* Over-aged cookie ? */
-					cookie->expires = deadline;
+			mem_free(date);
+
+			if (expires) {
+				if (max_age > 0) {
+					int seconds = max_age*24*3600;
+					ttime deadline = time(NULL) + seconds;
+
+					if (expires > deadline) /* Over-aged cookie ? */
+						expires = deadline;
+				}
+
+				cookie->expires = expires;
 			}
 		}
-
-#if 0
-		/* I decided not to expire such cookies (possibly ones with
-		 * date we can't parse properly), but instead just keep their
-		 * expire value at zero, thus making them expire on ELinks'
-		 * quit. --pasky */
-		if (!cookie->expires) {
-			/* We use zero for cookies which expire with
-			 * browser shutdown. */
-			cookie->expires = (ttime) 1;
-		}
-#endif
-		mem_free(date);
-
-	} else {
-		cookie->expires = (ttime) 0;
 	}
 
-	cookie->path = parse_http_header_param(str, "path");
-	if (!cookie->path) {
+	path = parse_header_param(str, "path");
+	if (!path) {
 		unsigned char *path_end;
-		struct string path;
 
-		if (!init_string(&path)) {
+		path = get_uri_string(uri, URI_PATH);
+		if (!path) {
 			free_cookie(cookie);
 			return;
 		}
 
-		add_char_to_string(&path, '/');
-		add_bytes_to_string(&path, uri->data, uri->datalen);
-
-		cookie->path = path.source;
-		for (path_end = cookie->path; *path_end; path_end++) {
-			if (end_of_dir(*path_end)) {
-				*path_end = '\0';
-				break;
-			}
-		}
-
-		for (path_end = cookie->path + strlen(cookie->path) - 1;
-		     path_end >= cookie->path; path_end--) {
+		for (path_end = path + strlen(path) - 1;
+		     path_end >= path; path_end--) {
 			if (*path_end == '/') {
 				path_end[1] = '\0';
 				break;
@@ -319,112 +380,104 @@ set_cookie(struct uri *uri, unsigned char *str)
 		}
 
 	} else {
-		if (!cookie->path[0]
-		    || cookie->path[strlen(cookie->path) - 1] != '/')
-			add_to_strn(&cookie->path, "/");
+		if (!path[0]
+		    || path[strlen(path) - 1] != '/')
+			add_to_strn(&path, "/");
 
-		if (cookie->path[0] != '/') {
-			add_to_strn(&cookie->path, "x");
-			memmove(cookie->path + 1, cookie->path,
-				strlen(cookie->path) - 1);
-			cookie->path[0] = '/';
+		if (path[0] != '/') {
+			add_to_strn(&path, "x");
+			memmove(path + 1, path, strlen(path) - 1);
+			path[0] = '/';
 		}
 	}
+	cookie->path = path;
 
 	if (cookie->domain[0] == '.')
 		memmove(cookie->domain, cookie->domain + 1,
 			strlen(cookie->domain));
 
-	secure = parse_http_header_param(str, "secure");
+	/* cookie->secure is set to 0 by default by calloc(). */
+	secure = parse_header_param(str, "secure");
 	if (secure) {
 		cookie->secure = 1;
 		mem_free(secure);
-	} else {
-		cookie->secure = 0;
 	}
 
-#ifdef COOKIES_DEBUG
+#ifdef DEBUG_COOKIES
 	{
-		unsigned char *server = memacpy(uri->host, uri->hostlen);
-
-		DBG("Got cookie %s = %s from %s (%s), domain %s, "
-		      "expires at %d, secure %d\n", cookie->name,
-		      cookie->value, cookie->server, server, cookie->domain,
+		DBG("Got cookie %s = %s from %s, domain %s, "
+		      "expires at %d, secure %d", cookie->name,
+		      cookie->value, cookie->server->host, cookie->domain,
 		      cookie->expires, cookie->secure);
-		if (server) mem_free(server);
 	}
 #endif
 
-	if (!check_domain_security(cookie->domain, uri->host, uri->hostlen)) {
-#ifdef COOKIES_DEBUG
-		DBG("Domain security violated: %s vs %*s", cookie->domain,
-				uri->host, uri->hostlen);
+	if (!is_domain_security_ok(cookie->domain, uri->host, uri->hostlen)) {
+#ifdef DEBUG_COOKIES
+		DBG("Domain security violated: %s vs %.*s", cookie->domain,
+		    uri->hostlen, uri->host);
 #endif
 		mem_free(cookie->domain);
 		cookie->domain = memacpy(uri->host, uri->hostlen);
 	}
 
-	cookie->id = cookie_id++;
-
-	foreach (cs, c_servers) {
-		if (strlcasecmp(cs->server, -1, uri->host, uri->hostlen))
-			continue;
-
-		if (cs->accept)	goto ok;
-
-#ifdef COOKIES_DEBUG
-		DBG("Dropped.");
-#endif
-		free_cookie(cookie);
-		return;
-	}
-
 	/* We have already check COOKIES_ACCEPT_NONE */
 	if (get_cookies_accept_policy() == COOKIES_ACCEPT_ASK) {
 		add_to_list(cookie_queries, cookie);
-		add_questions_entry(accept_cookie_dialog);
+		add_questions_entry(accept_cookie_dialog, cookie);
 		return;
 	}
 
-ok:
 	accept_cookie(cookie);
-
-	return;
 }
 
-
 static void
-accept_cookie(struct cookie *c)
+del_from_cookie_list(struct cookie *cookie)
+{
+	del_from_list(cookie);
+	free_cookie(cookie);
+}
+
+void
+accept_cookie(struct cookie *cookie)
 {
 	struct c_domain *cd;
-	struct cookie *d, *e;
+	struct listbox_item *root = cookie->server->box_item;
 	int domain_len;
 
-	c->box_item = add_listbox_items(&cookie_browser, c, 1, c->server, c->name, NULL);
+	if (root)
+		cookie->box_item = add_listbox_leaf(&cookie_browser, root, cookie);
 
-	foreach (d, cookies) {
-		if (strcasecmp(d->name, c->name)
-		    || strcasecmp(d->domain, c->domain))
-			continue;
-		e = d;
-		d = d->prev;
-		del_from_list(e);
-		free_cookie(e);
+	/* Do not weed out duplicates when loading the cookie file. It doesn't
+	 * scale at all, being O(N^2) and taking about 2s with my 500 cookies
+	 * (so if you don't notice that 100ms with your 100 cookies, that's
+	 * not an argument). --pasky */
+	if (!cookies_nosave) {
+		struct cookie *c, *next;
+
+		foreachsafe (c, next, cookies) {
+			if (strcasecmp(c->name, cookie->name)
+			    || strcasecmp(c->domain, cookie->domain))
+				continue;
+
+			del_from_cookie_list(c);
+		}
 	}
 
-	add_to_list(cookies, c);
+	add_to_list(cookies, cookie);
 	cookies_dirty = 1;
 
+	/* XXX: This crunches CPU too. --pasky */
 	foreach (cd, c_domains)
-		if (!strcasecmp(cd->domain, c->domain))
+		if (!strcasecmp(cd->domain, cookie->domain))
 			return;
 
-	domain_len = strlen(c->domain);
+	domain_len = strlen(cookie->domain);
 	/* One byte is reserved for domain in struct c_domain. */
-	cd = mem_alloc(sizeof(struct c_domain) + domain_len);
+	cd = mem_alloc(sizeof(*cd) + domain_len);
 	if (!cd) return;
 
-	memcpy(cd->domain, c->domain, domain_len + 1);
+	memcpy(cd->domain, cookie->domain, domain_len + 1);
 	add_to_list(c_domains, cd);
 
 	if (get_cookies_save() && get_cookies_resave())
@@ -432,6 +485,8 @@ accept_cookie(struct cookie *c)
 }
 
 #if 0
+static unsigned int cookie_id = 0;
+
 static void
 delete_cookie(struct cookie *c)
 {
@@ -462,7 +517,7 @@ end:
 static struct
 cookie *find_cookie_id(void *idp)
 {
-	int id = (int)idp;
+	int id = (int) idp;
 	struct cookie *c;
 
 	foreach (c, cookies)
@@ -487,25 +542,9 @@ reject_cookie(void *idp)
 static void
 cookie_default(void *idp, int a)
 {
-	struct c_server *s;
 	struct cookie *c = find_cookie_id(idp);
-	int server_len;
 
-	if (!c) return;
-
-	foreach (s, c_servers)
-		if (!strcasecmp(s->server, c->server))
-			goto found;
-
-	server_len = strlen(c->server);
-	/* One byte is reserved for server in struct c_server. */
-	s = mem_alloc(sizeof(struct c_server) + server_len);
-	if (s) {
-		memcpy(s->server, c->server, server_len + 1);
-		add_to_list(c_servers, s);
-found:
-		s->accept = a;
-	}
+	if (c) c->server->accept = a;
 }
 
 
@@ -524,69 +563,41 @@ accept_cookie_never(void *idp)
 }
 #endif
 
-static void
-accept_cookie_dialog(struct session *ses)
-{
-	struct cookie *cookie = cookie_queries.next;
-	unsigned char *expires = NULL;
-
-	assert(ses);
-
-	if (list_empty(cookie_queries)) return;
-
-	del_from_list(cookie);
-
-#ifdef HAVE_STRFTIME
-	if (cookie->expires) {
-		struct tm *when_local = localtime(&cookie->expires);
-		unsigned char str[13];
-		int wr = strftime(str, sizeof(str), "%b %e %H:%M", when_local);
-
-		if (wr > 0)
-			expires = memacpy(str, wr);
-	}
-#endif
-
-	msg_box(ses->tab->term, getml(expires, NULL), MSGBOX_FREE_TEXT,
-		N_("Accept cookie?"), AL_LEFT,
-		msg_text(ses->tab->term, N_("Do you want to accept a cookie "
-		"from %s?\n\n"
-		"Name: %s\n"
-		"Value: %s\n"
-		"Domain: %s\n"
-		"Expires: %s\n"
-		"Secure: %s\n"),
-		cookie->server, cookie->name, cookie->value,
-		cookie->domain,
-		expires ? expires : _("at quit time",  ses->tab->term),
-		_(cookie->secure ? N_("yes") : N_("no"), ses->tab->term)),
-		cookie, 2,
-		N_("Accept"), accept_cookie, B_ENTER,
-		N_("Reject"), free_cookie, B_ESC);
-}
-
-
+/* Check whether domain is matching server
+ * Ie.
+ * example.com matches www.example.com/
+ * example.com doesn't match www.example.com.org/
+ * example.com doesn't match www.example.comm/
+ * example.com doesn't match example.co
+ */
 static int
-is_in_domain(unsigned char *d, unsigned char *s, int sl)
+is_in_domain(unsigned char *domain, unsigned char *server, int server_len)
 {
-	int dl = strlen(d);
+	int domain_len = strlen(domain);
+	int len;
 
-	if (dl > sl) return 0;
-	if (dl == sl) return !strncasecmp(d, s, sl);
-	if (s[sl - dl - 1] != '.') return 0;
+	if (domain_len > server_len)
+		return 0;
 
-	return !strncasecmp(d, s + sl - dl, dl);
+	if (domain_len == server_len)
+		return !strncasecmp(domain, server, server_len);
+
+	len = server_len - domain_len;
+	if (server[len - 1] != '.')
+		return 0;
+
+	return !strncasecmp(domain, server + len, domain_len);
 }
 
 
 static inline int
-is_path_prefix(unsigned char *d, unsigned char *s, int sl)
+is_path_prefix(unsigned char *d, unsigned char *s)
 {
 	int dl = strlen(d);
 
 	/* TODO: strlcmp()? --pasky */
 
-	if (dl > sl) return 0;
+	if (dl > strlen(s)) return 0;
 
 	return !memcmp(d, s, dl);
 }
@@ -599,9 +610,8 @@ struct string *
 send_cookies(struct uri *uri)
 {
 	struct c_domain *cd;
-	struct cookie *c, *d;
-	unsigned char *data = NULL;
-	int datalen = uri->datalen + 1;
+	struct cookie *c, *next;
+	unsigned char *path = NULL;
 	static struct string header;
 
 	if (!uri->host || !uri->data)
@@ -609,28 +619,25 @@ send_cookies(struct uri *uri)
 
 	foreach (cd, c_domains)
 		if (is_in_domain(cd->domain, uri->host, uri->hostlen)) {
-			data = uri->data - 1;
+			path = get_uri_string(uri, URI_PATH);
 			break;
 		}
 
-	if (!data) return NULL;
+	if (!path) return NULL;
 
 	init_string(&header);
 
-	foreach (c, cookies) {
+	foreachsafe (c, next, cookies) {
 		if (!is_in_domain(c->domain, uri->host, uri->hostlen)
-		    || !is_path_prefix(c->path, data, datalen))
+		    || !is_path_prefix(c->path, path))
 			continue;
 
 		if (is_expired(c->expires)) {
-#ifdef COOKIES_DEBUG
-			DBG("Cookie %s=%s (exp %d) expired.\n",
-			      c->name, c->value, c->expires);
+#ifdef DEBUG_COOKIES
+			DBG("Cookie %s=%s (exp %d) expired.",
+			    c->name, c->value, c->expires);
 #endif
-			d = c;
-			c = c->prev;
-			del_from_list(d);
-			free_cookie(d);
+			del_from_cookie_list(c);
 
 			cookies_dirty = 1;
 			continue;
@@ -640,29 +647,33 @@ send_cookies(struct uri *uri)
 		if (c->secure && uri->protocol != PROTOCOL_HTTPS)
 			continue;
 
-		if (header.length) {
+		if (header.length)
 			add_to_string(&header, "; ");
-		}
 
 		add_to_string(&header, c->name);
 		add_char_to_string(&header, '=');
 		add_to_string(&header, c->value);
-#ifdef COOKIES_DEBUG
+#ifdef DEBUG_COOKIES
 		DBG("Cookie: %s=%s", c->name, c->value);
 #endif
 	}
 
+	mem_free(path);
+
 	if (cookies_dirty && get_cookies_save() && get_cookies_resave())
 		save_cookies();
 
-	if (!header.length)
+	if (!header.length) {
 		done_string(&header);
+		return NULL;
+	}
 
-	return header.length ? &header : NULL;
+	return &header;
 }
 
 static void done_cookies(struct module *module);
 
+
 void
 load_cookies(void) {
 	/* Buffer size is set to be enough to read long lines that
@@ -687,68 +698,66 @@ load_cookies(void) {
 	if (elinks_home) mem_free(cookfile);
 	if (!fp) return;
 
-	while (safe_fgets(in_buffer, 6 * MAX_STR_LEN, fp)) {
+	/* XXX: We don't want to overwrite the cookies file
+	 * periodically to our death. */
+	cookies_nosave = 1;
+
+	while (fgets(in_buffer, 6 * MAX_STR_LEN, fp)) {
 		struct cookie *cookie;
 		unsigned char *p, *q = in_buffer;
-		enum { NAME = 0, VALUE, SERVER, PATH, DOMAIN, MEMBERS } member;
-		unsigned char *members[MEMBERS];
-		ttime expires = 0;
+		enum { NAME = 0, VALUE, SERVER, PATH, DOMAIN, EXPIRES, SECURE, MEMBERS } member;
+		struct {
+			unsigned char *pos;
+			int len;
+		} members[MEMBERS];
+		ttime expires;
 
-		memset(members, 0, sizeof(unsigned char *) * MEMBERS);
-
-		/* First read in and allocate all members. */
-		for (member = NAME; member < MEMBERS; member++, q = p) {
+		/* First find all members. */
+		for (member = NAME; member < MEMBERS; member++, q = ++p) {
 			p = strchr(q, '\t');
-			if (!p) break;
-			*p++ = '\0';
-			members[member] = stracpy(q);
-			if (!members[member]) break;
-		}
-
-		/* Finally get a hold of the expire field. */
-		p = (member == MEMBERS ? strchr(q, '\t') : NULL);
-
-		if (p) {
-			expires = atol(q);
-			if (is_dead(expires)) {
-				p = NULL;
-				cookies_dirty = 1;
+			if (!p) {
+				if (member + 1 != MEMBERS) break; /* last field ? */
+				p = strchr(q, '\n');
+				if (!p) break;
 			}
+
+			members[member].pos = q;
+			members[member].len = p - q;
 		}
 
-		/* Prepare cookie if all members and fields was read. */
-		cookie = (p ? mem_calloc(1, sizeof(struct cookie)) : NULL);
+		if (member != MEMBERS) continue;	/* Invalid line. */
 
-		if (!cookie) {
-			/* Something went wrong so clean up. */
-			for (member = NAME; member < MEMBERS; member++)
-				if (members[member])
-					mem_free(members[member]);
+		/* Skip expired cookies if any. */
+		expires = str_to_ttime(members[EXPIRES].pos);
+		if (is_dead(expires)) {
+			cookies_dirty = 1;
 			continue;
 		}
 
-		cookie->name	= members[NAME];
-		cookie->value	= members[VALUE];
-		cookie->server	= members[SERVER];
-		cookie->path	= members[PATH];
-		cookie->domain	= members[DOMAIN];
+		/* Prepare cookie if all members and fields was read. */
+		cookie = mem_calloc(1, sizeof(*cookie));
+		if (!cookie) continue;
+
+		cookie->server  = get_cookie_server(members[SERVER].pos, members[SERVER].len);
+		cookie->name	= memacpy(members[NAME].pos, members[NAME].len);
+		cookie->value	= memacpy(members[VALUE].pos, members[VALUE].len);
+		cookie->path	= memacpy(members[PATH].pos, members[PATH].len);
+		cookie->domain	= memacpy(members[DOMAIN].pos, members[DOMAIN].len);
+
+		/* Check whether all fields were correctly allocated. */
+		if (!cookie->server || !cookie->name || !cookie->value
+		    || !cookie->path || !cookie->domain) {
+			free_cookie(cookie);
+			continue;
+		}
 
-		*p++ = '\0';
 		cookie->expires = expires;
+		cookie->secure  = !!atoi(members[SECURE].pos);
 
-		/* Drop ending '\n'. */
-		if (*p) p[strlen(p) - 1] = '\0';
-		cookie->secure = atoi(p);
-
-		cookie->id = cookie_id++;
-
-		/* XXX: We don't want to overwrite the cookies file
-		 * periodically to our death. */
-		cookies_nosave = 1;
 		accept_cookie(cookie);
-		cookies_nosave = 0;
 	}
 
+	cookies_nosave = 0;
 	fclose(fp);
 }
 
@@ -759,7 +768,7 @@ save_cookies(void) {
 	struct secure_save_info *ssi;
 
 	if (cookies_nosave || !elinks_home || !cookies_dirty
-	    || get_opt_int_tree(cmdline_options, "anonymous"))
+	    || get_cmd_opt_bool("anonymous"))
 		return;
 
 	cookfile = straconcat(elinks_home, COOKIES_FILENAME, NULL);
@@ -773,7 +782,7 @@ save_cookies(void) {
 		if (is_dead(c->expires)) continue;
 		if (secure_fprintf(ssi, "%s\t%s\t%s\t%s\t%s\t%ld\t%d\n",
 				   c->name, c->value,
-				   empty_string_or_(c->server),
+				   c->server->host,
 				   empty_string_or_(c->path),
 				   empty_string_or_(c->domain),
 				   c->expires, c->secure) < 0)
@@ -790,6 +799,16 @@ init_cookies(struct module *module)
 		load_cookies();
 }
 
+static void
+free_cookies_list(struct list_head *list)
+{
+	while (!list_empty(*list)) {
+		struct cookie *cookie = list->next;
+
+		del_from_list(cookie);
+		free_cookie(cookie);
+	}
+}
 
 static void
 done_cookies(struct module *module)
@@ -799,19 +818,8 @@ done_cookies(struct module *module)
 	if (!cookies_nosave && get_cookies_save())
 		save_cookies();
 
-	while (!list_empty(cookies)) {
-		struct cookie *cookie = cookies.next;
-
-		del_from_list(cookie);
-		free_cookie(cookie);
-	}
-
-	while (!list_empty(cookie_queries)) {
-		struct cookie *cookie = cookie_queries.next;
-
-		del_from_list(cookie);
-		free_cookie(cookie);
-	}
+	free_cookies_list(&cookies);
+	free_cookies_list(&cookie_queries);
 }
 
 struct module cookies_module = struct_module(
@@ -823,5 +831,3 @@ struct module cookies_module = struct_module(
 	/* init: */		init_cookies,
 	/* done: */		done_cookies
 );
-
-#endif /* CONFIG_COOKIES */
