@@ -1,12 +1,13 @@
-/* $Id: formhist.h,v 1.24 2003/11/26 19:39:31 zas Exp $ */
+/* $Id: formhist.h,v 1.28 2004/12/18 01:42:19 pasky Exp $ */
 
 #ifndef EL__FORMHIST_FORMHIST_H
 #define EL__FORMHIST_FORMHIST_H
 
+#include "document/forms.h"
 #include "modules/module.h"
 #include "sched/session.h"
 #include "util/lists.h"
-#include "viewer/text/form.h"
+#include "util/object.h"
 
 struct formhist_data {
 	LIST_HEAD(struct formhist_data);
@@ -17,10 +18,10 @@ struct formhist_data {
 	/* This is indeed maintained by formhist.c, not dialogs.c; much easier
 	 * and simpler. */
 	struct listbox_item *box_item;
-	int refcount;
+	struct object object;
 
 	/* Whether to save this form or not. */
-	int dontsave:1;
+	unsigned int dontsave:1;
 
 	/* <action> URI for this form. Must be at end of struct. */
 	unsigned char url[1];
@@ -34,7 +35,7 @@ unsigned char *form_type2str(enum form_type num);
  * value if present, NULL upon an error. */
 unsigned char *get_form_history_value(unsigned char *url, unsigned char *name);
 
-void memorize_form(struct session *ses, struct list_head *submit, struct form_control *frm);
+void memorize_form(struct session *ses, struct list_head *submit, struct form *forminfo);
 
 int save_forms_to_file(void);
 void free_form(struct formhist_data *form);
