@@ -16,13 +16,13 @@
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
 #ifdef HAVE_CONFIG_H
-#include <config.h>
+#include "config.h"
 #endif
 
 #include <stdlib.h>		/* Just for NULL.  */
 
-#include "gettextP.h"
-#include "libgnuintl.h"
+#include "intl/gettext/gettextP.h"
+#include "intl/gettext/libintl.h"
 
 
 /* Look up MSGID in the current default message catalog for the current
