@@ -1,11 +1,11 @@
-/* $Id: select.h,v 1.8 2003/12/01 14:01:26 pasky Exp $ */
+/* $Id: select.h,v 1.10 2004/04/14 19:46:56 jonas Exp $ */
 
 #ifndef EL__LOWLEVEL_SELECT_H
 #define EL__LOWLEVEL_SELECT_H
 
+#include "util/lists.h"
 #include "util/ttime.h"
 
-extern int terminate;
 extern struct list_head bottom_halves;
 
 long select_info(int);
