@@ -1,9 +1,9 @@
-/* $Id: dialogs.h,v 1.3 2003/12/31 12:53:40 pasky Exp $ */
+/* $Id: dialogs.h,v 1.4 2004/06/22 21:49:49 jonas Exp $ */
 
 #ifndef EL__MIME_DIALOGS_H
 #define EL__MIME_DIALOGS_H
 
-#include "terminal/terminal.h"
+struct terminal;
 
 void menu_add_ext(struct terminal *, void *, void *);
 void menu_del_ext(struct terminal *, void *, void *);
