@@ -1,11 +1,12 @@
-/* $Id: codes.h,v 1.3 2003/06/21 14:30:05 pasky Exp $ */
+/* $Id: codes.h,v 1.5 2004/06/28 02:27:20 jonas Exp $ */
 
 #ifndef EL__PROTOCOL_HTTP_CODES_H
 #define EL__PROTOCOL_HTTP_CODES_H
 
+struct connection;
+
 /* HTTP response codes device. */
 
-unsigned char *http_code_to_string(int code);
-unsigned char *http_error_document(int code);
+void http_error_document(struct connection *conn, int code);
 
 #endif /* EL__PROTOCOL_HTTP_CODES_H */
