@@ -1,5 +1,5 @@
-/* Connections managment */
-/* $Id: connection.c,v 1.135 2003/12/21 14:51:20 zas Exp $ */
+/* Connections management */
+/* $Id: connection.c,v 1.210.4.3 2005/04/05 21:08:43 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -13,8 +13,8 @@
 
 #include "elinks.h"
 
-#include "config/options.h"
 #include "cache/cache.h"
+#include "config/options.h"
 #include "document/document.h"
 #include "encoding/encoding.h"
 #include "intl/gettext/libintl.h"
@@ -27,7 +27,6 @@
 #include "sched/connection.h"
 #include "sched/session.h"
 #include "ssl/ssl.h"
-#include "util/base64.h"
 #include "util/error.h"
 #include "util/memory.h"
 #include "util/object.h"
@@ -39,16 +38,19 @@
 struct keepalive_connection {
 	LIST_HEAD(struct keepalive_connection);
 
-	protocol_handler *protocol;
+	/* XXX: This is just the URI of the connection that registered the
+	 * keepalive connection so only rely on the protocol, user, password,
+	 * host and port part. */
+	struct uri *uri;
+
+	/* Function called when the keepalive has timed out or is deleted */
+	void (*done)(struct connection *);
 
 	ttime timeout;
 	ttime add_time;
 
-	int port;
-	int pf;
+	unsigned int protocol_family:1; /* 0 == PF_INET, 1 == PF_INET6 */
 	int socket;
-
-	unsigned char host[1]; /* Keep last */
 };
 
 
@@ -65,13 +67,9 @@ static INIT_LIST_HEAD(host_connections);
 static INIT_LIST_HEAD(keepalive_connections);
 
 /* Prototypes */
-static void send_connection_info(struct connection *conn);
+static void notify_connection_callbacks(struct connection *conn);
 static void check_keepalive_connections(void);
 
-/* See connection_state description. */
-#define is_in_result_state(cstate)	(cstate < 0)
-#define is_in_progress_state(cstate)	(cstate >= 0)
-
 static /* inline */ enum connection_priority
 get_priority(struct connection *conn)
 {
@@ -100,10 +98,11 @@ connect_info(int type)
 			break;
 		case INFO_CONNECTING:
 			foreach (conn, queue)
-				info += (conn->state > S_WAIT && conn->state < S_TRANS);
+				info += is_in_connecting_state(conn->state);
 			break;
 		case INFO_TRANSFER:
-			foreach (conn, queue) info += (conn->state == S_TRANS);
+			foreach (conn, queue)
+				info += is_in_transfering_state(conn->state);
 			break;
 		case INFO_KEEP:
 			foreach (keep_conn, keepalive_connections) info++;
@@ -136,20 +135,21 @@ connection_disappeared(struct connection *conn)
 struct host_connection {
 	LIST_HEAD(struct host_connection);
 
-	int connections;
-	unsigned char host[1]; /* Keep last */
+	/* XXX: This is just the URI of the connection that registered the
+	 * host connection so only rely on the host part. */
+	struct uri *uri;
+	struct object object;
 };
 
 static struct host_connection *
 get_host_connection(struct connection *conn)
 {
-	unsigned char *host = conn->uri.host;
-	int hostlen = conn->uri.hostlen;
 	struct host_connection *host_conn;
 
-	if (!host) return NULL;
+	if (!conn->uri->host) return NULL;
+
 	foreach (host_conn, host_connections)
-		if (!strlcmp(host_conn->host, -1, host, hostlen))
+		if (compare_uri(host_conn->uri, conn->uri, URI_HOST))
 			return host_conn;
 
 	return NULL;
@@ -162,14 +162,15 @@ add_host_connection(struct connection *conn)
 {
 	struct host_connection *host_conn = get_host_connection(conn);
 
-	if (!host_conn && conn->uri.host) {
-		host_conn = mem_calloc(1, sizeof(struct host_connection) + conn->uri.hostlen);
+	if (!host_conn && conn->uri->host) {
+		host_conn = mem_calloc(1, sizeof(*host_conn));
 		if (!host_conn) return 0;
 
-		memcpy(host_conn->host, conn->uri.host, conn->uri.hostlen);
+		host_conn->uri = get_uri_reference(conn->uri);
+		object_nolock(host_conn, "host_connection");
 		add_to_list(host_connections, host_conn);
 	}
-	if (host_conn) host_conn->connections++;
+	if (host_conn) object_lock(host_conn);
 
 	return 1;
 }
@@ -182,17 +183,18 @@ done_host_connection(struct connection *conn)
 
 	if (!host_conn) return;
 
-	host_conn->connections--;
-	if (host_conn->connections > 0) return;
+	object_unlock(host_conn);
+	if (is_object_used(host_conn)) return;
 
 	del_from_list(host_conn);
+	done_uri(host_conn->uri);
 	mem_free(host_conn);
 }
 
 
 static void sort_queue();
 
-#ifdef DEBUG
+#ifdef CONFIG_DEBUG
 static void
 check_queue_bugs(void)
 {
@@ -206,8 +208,9 @@ check_queue_bugs(void)
 		cc += conn->running;
 
 		assertm(priority >= prev_priority, "queue is not sorted");
-		assertm(conn->state >= 0, "interrupted connection on queue "
-			"(conn %s, state %d)", struri(conn->uri), conn->state);
+		assertm(is_in_progress_state(conn->state),
+			"interrupted connection on queue (conn %s, state %d)",
+			struri(conn->uri), conn->state);
 		prev_priority = priority;
 	}
 
@@ -221,109 +224,115 @@ check_queue_bugs(void)
 
 
 static struct connection *
-init_connection(unsigned char *url, unsigned char *ref_url, int start,
-		enum cache_mode cache_mode, enum connection_priority priority)
+init_connection(struct uri *uri, struct uri *proxied_uri, struct uri *referrer,
+		int start, enum cache_mode cache_mode,
+		enum connection_priority priority)
 {
-	struct connection *conn = mem_calloc(1, sizeof(struct connection));
+	struct connection *conn = mem_calloc(1, sizeof(*conn));
 
 	if (!conn) return NULL;
 
-	if (!parse_uri(&conn->uri, url)
-	    || (VALID_PROTOCOL(conn->uri.protocol)
-		&& get_protocol_need_slash_after_host(conn->uri.protocol)
-		&& !conn->uri.hostlen)) {
-		/* Alert small hack to signal parse uri failure. */
-		*url = 0;
-		mem_free(conn);
-		return NULL;
-	}
+	assert(proxied_uri->protocol != PROTOCOL_PROXY);
 
+	/* load_uri() gets the URI from get_proxy() which grabs a reference for
+	 * us. */
+	conn->uri = uri;
+	conn->proxied_uri = proxied_uri;
 	conn->id = connection_id++;
-	conn->ref_url = ref_url;
-	conn->pri[priority] =  1;
+	conn->pri[priority] = 1;
 	conn->cache_mode = cache_mode;
-	conn->socket = conn->data_socket = -1;
+	conn->socket.fd = conn->data_socket.fd = -1;
 	conn->content_encoding = ENCODING_NONE;
 	conn->stream_pipes[0] = conn->stream_pipes[1] = -1;
 	conn->cgi_pipes[0] = conn->cgi_pipes[1] = -1;
 	init_list(conn->downloads);
 	conn->est_length = -1;
-	conn->prg.start = start;
-	conn->prg.timer = -1;
+	conn->progress.start = start;
+	conn->progress.timer = -1;
 	conn->timer = -1;
 
+	if (referrer) {
+		/* Don't set referrer when it is the file protocol and the URI
+		 * being loaded is not. This means CGI scripts will have it
+		 * available while preventing information about the local
+		 * system from being leaked to external servers. */
+		if (referrer->protocol != PROTOCOL_FILE
+		    || uri->protocol == PROTOCOL_FILE)
+			conn->referrer = get_uri_reference(referrer);
+	}
+
 	return conn;
-};
+}
 
 static void stat_timer(struct connection *conn);
 
 static void
-update_remaining_info(struct connection *conn)
+update_progress(struct connection *conn)
 {
-	struct remaining_info *prg = &conn->prg;
-	ttime a = get_time() - prg->last_time;
+	struct progress *progress = &conn->progress;
+	ttime a = get_time() - progress->last_time;
 
-	prg->loaded = conn->received;
-	prg->size = conn->est_length;
-	prg->pos = conn->from;
-	if (prg->size < prg->pos && prg->size != -1)
-		prg->size = conn->from;
+	progress->loaded = conn->received;
+	progress->size = conn->est_length;
+	progress->pos = conn->from;
+	if (progress->size < progress->pos && progress->size != -1)
+		progress->size = conn->from;
 
-	prg->dis_b += a;
-	while (prg->dis_b >= SPD_DISP_TIME * CURRENT_SPD_SEC) {
-		prg->cur_loaded -= prg->data_in_secs[0];
-		memmove(prg->data_in_secs, prg->data_in_secs + 1,
-			sizeof(int) * (CURRENT_SPD_SEC - 1));
-		prg->data_in_secs[CURRENT_SPD_SEC - 1] = 0;
-		prg->dis_b -= SPD_DISP_TIME;
+	progress->dis_b += a;
+	while (progress->dis_b >= SPD_DISP_TIME * CURRENT_SPD_SEC) {
+		progress->cur_loaded -= progress->data_in_secs[0];
+		memmove(progress->data_in_secs, progress->data_in_secs + 1,
+			sizeof(*progress->data_in_secs) * (CURRENT_SPD_SEC - 1));
+		progress->data_in_secs[CURRENT_SPD_SEC - 1] = 0;
+		progress->dis_b -= SPD_DISP_TIME;
 	}
 
-	prg->data_in_secs[CURRENT_SPD_SEC - 1] += prg->loaded - prg->last_loaded;
-	prg->cur_loaded += prg->loaded - prg->last_loaded;
-	prg->last_loaded = prg->loaded;
-	prg->last_time += a;
-	prg->elapsed += a;
-	prg->timer = install_timer(SPD_DISP_TIME, (void (*)(void *)) stat_timer, conn);
+	progress->data_in_secs[CURRENT_SPD_SEC - 1] += progress->loaded - progress->last_loaded;
+	progress->cur_loaded += progress->loaded - progress->last_loaded;
+	progress->last_loaded = progress->loaded;
+	progress->last_time += a;
+	progress->elapsed += a;
+	progress->timer = install_timer(SPD_DISP_TIME, (void (*)(void *)) stat_timer, conn);
 }
 
 static void
 stat_timer(struct connection *conn)
 {
-	update_remaining_info(conn);
-	send_connection_info(conn);
+	update_progress(conn);
+	notify_connection_callbacks(conn);
 }
 
 void
 set_connection_state(struct connection *conn, enum connection_state state)
 {
 	struct download *download;
-	struct remaining_info *prg = &conn->prg;
+	struct progress *progress = &conn->progress;
 
 	if (is_in_result_state(conn->state) && is_in_progress_state(state))
 		conn->prev_error = conn->state;
 
 	conn->state = state;
 	if (conn->state == S_TRANS) {
-		if (prg->timer == -1) {
-			if (!prg->valid) {
-				int tmp = prg->start;
-				int tmp2 = prg->seek;
+		if (progress->timer == -1) {
+			if (!progress->valid) {
+				int tmp = progress->start;
+				int tmp2 = progress->seek;
 
-				memset(prg, 0, sizeof(struct remaining_info));
-				prg->start = tmp;
-				prg->seek = tmp2;
-				prg->valid = 1;
+				memset(progress, 0, sizeof(*progress));
+				progress->start = tmp;
+				progress->seek = tmp2;
+				progress->valid = 1;
 			}
-			prg->last_time = get_time();
-			prg->last_loaded = prg->loaded;
-			update_remaining_info(conn);
+			progress->last_time = get_time();
+			progress->last_loaded = progress->loaded;
+			update_progress(conn);
 			if (connection_disappeared(conn))
 				return;
 		}
 
-	} else if (prg->timer != -1) {
-		kill_timer(prg->timer);
-		prg->timer = -1;
+	} else if (progress->timer != -1) {
+		kill_timer(progress->timer);
+		progress->timer = -1;
 	}
 
 	foreach (download, conn->downloads) {
@@ -331,7 +340,7 @@ set_connection_state(struct connection *conn, enum connection_state state)
 		download->prev_error = conn->prev_error;
 	}
 
-	if (is_in_progress_state(state)) send_connection_info(conn);
+	if (is_in_progress_state(state)) notify_connection_callbacks(conn);
 }
 
 static void
@@ -342,11 +351,13 @@ free_connection_data(struct connection *conn)
 	 * at least active_connections underflows along the way. --pasky */
 	conn->running = 0;
 
-	if (conn->socket != -1) set_handlers(conn->socket, NULL, NULL, NULL, NULL);
-	if (conn->data_socket != -1) set_handlers(conn->data_socket, NULL, NULL, NULL, NULL);
+	if (conn->socket.fd != -1)
+		set_handlers(conn->socket.fd, NULL, NULL, NULL, NULL);
+	if (conn->data_socket.fd != -1)
+		set_handlers(conn->data_socket.fd, NULL, NULL, NULL, NULL);
 	close_socket(NULL, &conn->data_socket);
 
-	/* XXX: See also protocol/http/http.c:uncompress_shutdown(). */
+	/* XXX: See also protocol/http/http.c:decompress_shutdown(). */
 	if (conn->stream) {
 		close_encoded(conn->stream);
 		conn->stream = NULL;
@@ -365,18 +376,14 @@ free_connection_data(struct connection *conn)
 		kill_dns_request(&conn->dnsquery);
 	}
 	if (conn->conn_info) {
-		if (conn->conn_info->addr) mem_free(conn->conn_info->addr);
-		mem_free(conn->conn_info);
-		conn->conn_info = NULL;
-	}
-	if (conn->buffer) {
-		mem_free(conn->buffer);
-		conn->buffer = NULL;
-	}
-	if (conn->info) {
-		mem_free(conn->info);
-		conn->info = NULL;
+		/* No callbacks should be made */
+		conn->conn_info->done = NULL;
+		done_connection_info(conn);
 	}
+
+	mem_free_set(&conn->buffer, NULL);
+	mem_free_set(&conn->info, NULL);
+
 	if (conn->timer != -1) {
 		kill_timer(conn->timer);
 		conn->timer = -1;
@@ -391,16 +398,15 @@ free_connection_data(struct connection *conn)
 }
 
 void
-send_connection_info(struct connection *conn)
+notify_connection_callbacks(struct connection *conn)
 {
 	enum connection_state state = conn->state;
-	struct download *download = conn->downloads.next;
+	struct download *download, *next;
 
-	while ((void *)download != &conn->downloads) {
-		download->ce = conn->cache;
-		download = download->next;
-		if (download->prev->end)
-			download->prev->end(download->prev, download->prev->data);
+	foreachsafe (download, next, conn->downloads) {
+		download->cached = conn->cached;
+		if (download->callback)
+			download->callback(download, download->data);
 		if (is_in_progress_state(state) && connection_disappeared(conn))
 			return;
 	}
@@ -410,42 +416,85 @@ static void
 done_connection(struct connection *conn)
 {
 	del_from_list(conn);
-	send_connection_info(conn);
-	/* TODO: Some free_uri(). */
-	mem_free(struri(conn->uri));
+	notify_connection_callbacks(conn);
+	if (conn->referrer) done_uri(conn->referrer);
+	done_uri(conn->uri);
+	done_uri(conn->proxied_uri);
 	mem_free(conn);
 	check_queue_bugs();
 }
 
+static inline void add_to_queue(struct connection *conn);
+
+/* Returns zero if no callback was done and the keepalive connection should be
+ * deleted or non-zero if the keepalive connection should not be deleted. */
+static int
+do_keepalive_connection_callback(struct keepalive_connection *keep_conn)
+{
+	struct uri *proxied_uri = get_proxied_uri(keep_conn->uri);
+	struct uri *proxy_uri   = get_proxy_uri(keep_conn->uri, NULL);
+
+	if (proxied_uri && proxy_uri) {
+		struct connection *conn;
+
+		conn = init_connection(proxy_uri, proxied_uri, NULL, 0,
+				       CACHE_MODE_NEVER, PRI_CANCEL);
+
+		if (conn) {
+			void (*done)(struct connection *) = keep_conn->done;
+
+			add_to_queue(conn);
+
+			/* Get the keepalive info and let it clean up */
+			if (!has_keepalive_connection(conn)
+			    || !add_host_connection(conn)) {
+				free_connection_data(conn);
+				done_connection(conn);
+				return 0;
+			}
+
+			active_connections++;
+			conn->running = 1;
+			done(conn);
+			return 1;
+		}
+	}
+
+	if (proxied_uri) done_uri(proxied_uri);
+	if (proxy_uri) done_uri(proxy_uri);
+
+	return 0;
+}
 
 static inline void
 done_keepalive_connection(struct keepalive_connection *keep_conn)
 {
+	if (keep_conn->done && do_keepalive_connection_callback(keep_conn))
+		return;
+
 	del_from_list(keep_conn);
 	if (keep_conn->socket != -1) close(keep_conn->socket);
+	done_uri(keep_conn->uri);
 	mem_free(keep_conn);
 }
 
 static struct keepalive_connection *
-init_keepalive_connection(struct connection *conn, ttime timeout)
+init_keepalive_connection(struct connection *conn, ttime timeout,
+			  void (*done)(struct connection *))
 {
 	struct keepalive_connection *keep_conn;
-	struct uri *uri = &conn->uri;
-	unsigned char *host = uri->user ? uri->user : uri->host;
-	int hostlen  = (uri->port ? uri->port + uri->portlen - host
-				  : uri->host + uri->hostlen - host);
+	struct uri *uri = conn->uri;
 
 	assert(uri->host);
 	if_assert_failed return NULL;
 
-	keep_conn = mem_calloc(1, sizeof(struct keepalive_connection) + hostlen);
+	keep_conn = mem_calloc(1, sizeof(*keep_conn));
 	if (!keep_conn) return NULL;
 
-	memcpy(keep_conn->host, host, hostlen);
-	keep_conn->port = get_uri_port(uri);
-	keep_conn->protocol = get_protocol_handler(uri->protocol);
-	keep_conn->pf = conn->pf;
-	keep_conn->socket = conn->socket;
+	keep_conn->uri = get_uri_reference(uri);
+	keep_conn->done = done;
+	keep_conn->protocol_family = conn->protocol_family;
+	keep_conn->socket = conn->socket.fd;
 	keep_conn->timeout = timeout;
 	keep_conn->add_time = get_time();
 
@@ -456,19 +505,11 @@ static struct keepalive_connection *
 get_keepalive_connection(struct connection *conn)
 {
 	struct keepalive_connection *keep_conn;
-	struct uri *uri = &conn->uri;
-	protocol_handler *handler = get_protocol_handler(uri->protocol);
-	int port = get_uri_port(uri);
-	unsigned char *host = uri->user ? uri->user : uri->host;
-	int hostlen  = (uri->port ? uri->port + uri->portlen - host
-				  : uri->host + uri->hostlen - host);
 
-	if (!uri->host) return NULL;
+	if (!conn->uri->host) return NULL;
 
 	foreach (keep_conn, keepalive_connections)
-		if (keep_conn->protocol == handler
-		    && keep_conn->port == port
-		    && !strlcmp(keep_conn->host, -1, host, hostlen))
+		if (compare_uri(keep_conn->uri, conn->uri, URI_KEEPALIVE))
 			return keep_conn;
 
 	return NULL;
@@ -481,34 +522,44 @@ has_keepalive_connection(struct connection *conn)
 
 	if (!keep_conn) return 0;
 
-	conn->socket = keep_conn->socket;
-	conn->pf = keep_conn->pf;
+	conn->socket.fd = keep_conn->socket;
+	conn->protocol_family = keep_conn->protocol_family;
 
-	/* Mark that the socket should not be closed */
+	/* Mark that the socket should not be closed and the callback should be
+	 * ignored. */
 	keep_conn->socket = -1;
+	keep_conn->done = NULL;
 	done_keepalive_connection(keep_conn);
 
 	return 1;
 }
 
 void
-add_keepalive_connection(struct connection *conn, ttime timeout)
+add_keepalive_connection(struct connection *conn, ttime timeout,
+			 void (*done)(struct connection *))
 {
 	struct keepalive_connection *keep_conn;
 
-	free_connection_data(conn);
-	assertm(conn->socket != -1, "keepalive connection not connected");
+	assertm(conn->socket.fd != -1, "keepalive connection not connected");
 	if_assert_failed goto done;
 
-	keep_conn = init_keepalive_connection(conn, timeout);
-	if (keep_conn)
+	keep_conn = init_keepalive_connection(conn, timeout, done);
+	if (keep_conn) {
 		add_to_list(keepalive_connections, keep_conn);
-	else
-		close(conn->socket);
+
+	} else if (done) {
+		/* It will take just a little more time */
+		done(conn);
+		return;
+
+	} else {
+		close(conn->socket.fd);
+	}
 
 done:
+	free_connection_data(conn);
 	done_connection(conn);
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half((void (*)(void *)) check_queue, NULL);
 }
 
 static void
@@ -521,7 +572,7 @@ keepalive_timer(void *x)
 void
 check_keepalive_connections(void)
 {
-	struct keepalive_connection *keep_conn;
+	struct keepalive_connection *keep_conn, *next;
 	ttime ct = get_time();
 	int p = 0;
 
@@ -530,11 +581,10 @@ check_keepalive_connections(void)
 		keepalive_timeout = -1;
 	}
 
-	foreach (keep_conn, keepalive_connections) {
+	foreachsafe (keep_conn, next, keepalive_connections) {
 		if (can_read(keep_conn->socket)
 		    || ct - keep_conn->add_time > keep_conn->timeout) {
-			keep_conn = keep_conn->prev;
-			done_keepalive_connection(keep_conn->next);
+			done_keepalive_connection(keep_conn);
 		} else {
 			p++;
 		}
@@ -584,7 +634,7 @@ sort_queue(void)
 
 		swp = 0;
 		foreach (conn, queue) {
-			if ((void *)conn->next == &queue) break;
+			if (!list_has_next(queue, conn)) break;
 
 			if (get_priority(conn->next) < get_priority(conn)) {
 				struct connection *c = conn->next;
@@ -600,10 +650,6 @@ sort_queue(void)
 static void
 interrupt_connection(struct connection *conn)
 {
-#ifdef HAVE_SSL
-	if (conn->ssl) done_ssl_connection(conn);
-#endif
-
 	close_socket(conn, &conn->socket);
 	free_connection_data(conn);
 }
@@ -618,7 +664,7 @@ suspend_connection(struct connection *conn)
 static void
 run_connection(struct connection *conn)
 {
-	protocol_handler *func = get_protocol_handler(conn->uri.protocol);
+	protocol_handler *func = get_protocol_handler(conn->uri->protocol);
 
 	assert(func);
 
@@ -633,6 +679,7 @@ run_connection(struct connection *conn)
 
 	active_connections++;
 	conn->running = 1;
+
 	func(conn);
 }
 
@@ -642,10 +689,10 @@ retry_connection(struct connection *conn)
 	int max_tries = get_opt_int("connection.retries");
 
 	interrupt_connection(conn);
-	if (conn->uri.post || !max_tries || ++conn->tries >= max_tries) {
-		/*send_connection_info(conn);*/
+	if (conn->uri->post || !max_tries || ++conn->tries >= max_tries) {
+		/* notify_connection_callbacks(conn); */
 		done_connection(conn);
-		register_bottom_half((void (*)(void *))check_queue, NULL);
+		register_bottom_half((void (*)(void *)) check_queue, NULL);
 	} else {
 		conn->prev_error = conn->state;
 		run_connection(conn);
@@ -656,9 +703,9 @@ void
 abort_connection(struct connection *conn)
 {
 	if (conn->running) interrupt_connection(conn);
-	/* send_connection_info(conn); */
+	/* notify_connection_callbacks(conn); */
 	done_connection(conn);
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half((void (*)(void *)) check_queue, NULL);
 }
 
 /* Set certain state on a connection and then abort the connection. */
@@ -678,7 +725,7 @@ retry_conn_with_state(struct connection *conn, enum connection_state state)
 }
 
 static int
-try_to_suspend_connection(struct connection *conn, unsigned char *host)
+try_to_suspend_connection(struct connection *conn, struct uri *uri)
 {
 	enum connection_priority priority = get_priority(conn);
 	struct connection *c;
@@ -686,8 +733,8 @@ try_to_suspend_connection(struct connection *conn, unsigned char *host)
 	foreachback (c, queue) {
 		if (get_priority(c) <= priority) return -1;
 		if (c->state == S_WAIT) continue;
-		if (c->uri.post && get_priority(c) < PRI_CANCEL) continue;
-		if (host && strlcmp(host, -1, c->uri.host, c->uri.hostlen)) continue;
+		if (c->uri->post && get_priority(c) < PRI_CANCEL) continue;
+		if (uri && !compare_uri(uri, c->uri, URI_HOST)) continue;
 		suspend_connection(c);
 		return 0;
 	}
@@ -700,8 +747,8 @@ try_connection(struct connection *conn, int max_conns_to_host, int max_conns)
 {
 	struct host_connection *host_conn = get_host_connection(conn);
 
-	if (host_conn && host_conn->connections >= max_conns_to_host)
-		return try_to_suspend_connection(conn, host_conn->host) ? 0 : -1;
+	if (host_conn && get_object_refcount(host_conn) >= max_conns_to_host)
+		return try_to_suspend_connection(conn, host_conn->uri) ? 0 : -1;
 
 	if (active_connections >= max_conns)
 		return try_to_suspend_connection(conn, NULL) ? 0 : -1;
@@ -722,12 +769,11 @@ again:
 	check_queue_bugs();
 	check_keepalive_connections();
 
-	while (conn != (struct connection *)&queue) {
+	while (conn != (struct connection *) &queue) {
 		struct connection *c;
 		enum connection_priority pri = get_priority(conn);
 
-		/* No way to reduce code redundancy here ? --Zas */
-		for (c = conn; c != (struct connection *)&queue && get_priority(c) == pri;) {
+		for (c = conn; c != (struct connection *) &queue && get_priority(c) == pri;) {
 			struct connection *cc = c;
 
 			c = c->next;
@@ -736,7 +782,7 @@ again:
 				goto again;
 		}
 
-		for (c = conn; c != (struct connection *)&queue && get_priority(c) == pri;) {
+		for (c = conn; c != (struct connection *) &queue && get_priority(c) == pri;) {
 			struct connection *cc = c;
 
 			c = c->next;
@@ -761,22 +807,23 @@ again2:
 }
 
 int
-load_url(unsigned char *url, unsigned char *ref_url, struct download *download,
+load_uri(struct uri *uri, struct uri *referrer, struct download *download,
 	 enum connection_priority pri, enum cache_mode cache_mode, int start)
 {
-	struct cache_entry *ce = NULL;
+	struct cache_entry *cached;
 	struct connection *conn;
-	unsigned char *u;
+	struct uri *proxy_uri, *proxied_uri;
+	enum connection_state connection_state = S_OK;
 
 	if (download) {
 		download->conn = NULL;
-		download->ce = NULL;
+		download->cached = NULL;
 		download->pri = pri;
 		download->state = S_OUT_OF_MEM;
 		download->prev_error = 0;
 	}
 
-#ifdef DEBUG
+#ifdef CONFIG_DEBUG
 	foreach (conn, queue) {
 		struct download *assigned;
 
@@ -784,7 +831,7 @@ load_url(unsigned char *url, unsigned char *ref_url, struct download *download,
 			assertm(assigned != download, "Download assigned to '%s'", struri(conn->uri));
 			if_assert_failed {
 				download->state = S_INTERNAL;
-				if (download->end) download->end(download, download->data);
+				if (download->callback) download->callback(download, download->data);
 				return 0;
 			}
 			/* No recovery path should be necessary. */
@@ -792,83 +839,96 @@ load_url(unsigned char *url, unsigned char *ref_url, struct download *download,
 	}
 #endif
 
-	if (cache_mode <= CACHE_MODE_NORMAL
-	    && (ce = find_in_cache(url))
-	    && !ce->incomplete) {
-		if (!is_object_used(ce) &&
-		    ((ce->cache_mode == CACHE_MODE_NEVER && cache_mode != CACHE_MODE_ALWAYS)
-		     || (ce->redirect && !get_opt_int("document.cache.cache_redirects")))) {
-			delete_cache_entry(ce);
-			ce = NULL;
-		} else {
-			if (download) {
-				download->ce = ce;
-				download->state = S_OK;
-			/* XXX: This doesn't work since sometimes stat->prg is
-			 * undefined and contains random memory locations. It's
-			 * not supposed to point on anything here since stat
-			 * has no connection attached. Downloads resuming will
-			 * probably break in some cases without this, though.
+	cached = get_validated_cache_entry(uri, cache_mode);
+	if (cached) {
+		if (download) {
+			download->cached = cached;
+			download->state = S_OK;
+			/* XXX:
+			 * This doesn't work since sometimes |download->progress|
+			 * is undefined and contains random memory locations.
+			 * It's not supposed to point on anything here since
+			 * |download| has no connection attached.
+			 * Downloads resuming will probably break in some
+			 * cases without this, though.
 			 * FIXME: Needs more investigation. --pasky */
-			/* if (stat->prg) stat->prg->start = start; */
-				if (download->end)
-					download->end(download, download->data);
-			}
-			return 0;
+			/* if (download->progress) download->progress->start = start; */
+			if (download->callback)
+				download->callback(download, download->data);
 		}
+		return 0;
 	}
 
-	u = get_proxy(url);
-	if (!u) {
-		if (download) download->end(download, download->data);
+	proxied_uri = get_proxied_uri(uri);
+	proxy_uri   = get_proxy_uri(uri, &connection_state);
+
+	if (!proxy_uri
+	    || !proxied_uri
+	    || (get_protocol_need_slash_after_host(proxy_uri->protocol)
+		&& !proxy_uri->hostlen)) {
+
+		if (download) {
+			if (connection_state == S_OK) {
+				connection_state = proxy_uri && proxied_uri
+						 ? S_BAD_URL : S_OUT_OF_MEM;
+			}
+
+			download->state = connection_state;
+			download->callback(download, download->data);
+		}
+		if (proxy_uri) done_uri(proxy_uri);
+		if (proxied_uri) done_uri(proxied_uri);
 		return -1;
 	}
 
 	foreach (conn, queue) {
-		if (conn->detached || strcmp(struri(conn->uri), u))
+		if (conn->detached
+		    || !compare_uri(conn->uri, proxy_uri, 0))
 			continue;
 
-		mem_free(u);
+		done_uri(proxy_uri);
+		done_uri(proxied_uri);
 
 		if (get_priority(conn) > pri) {
 			del_from_list(conn);
 			conn->pri[pri]++;
 			add_to_queue(conn);
-			register_bottom_half((void (*)(void *))check_queue, NULL);
+			register_bottom_half((void (*)(void *)) check_queue, NULL);
 		} else {
 			conn->pri[pri]++;
 		}
 
 		if (download) {
-			download->prg = &conn->prg;
+			download->progress = &conn->progress;
 			download->conn = conn;
-			download->ce = conn->cache;
+			download->cached = conn->cached;
 			add_to_list(conn->downloads, download);
+			/* This is likely to call download->callback() now! */
 			set_connection_state(conn, conn->state);
 		}
 		check_queue_bugs();
 		return 0;
 	}
 
-	conn = init_connection(u, ref_url, start, cache_mode, pri);
+	conn = init_connection(proxy_uri, proxied_uri, referrer, start, cache_mode, pri);
 	if (!conn) {
 		if (download) {
-			/* Zero length uri signals parse uri failure */
-			download->state = (!*u ? S_BAD_URL : S_OUT_OF_MEM);
-			download->end(download, download->data);
+			download->state = S_OUT_OF_MEM;
+			download->callback(download, download->data);
 		}
-		mem_free(u);
+		if (proxy_uri) done_uri(proxy_uri);
+		if (proxied_uri) done_uri(proxied_uri);
 		return -1;
 	}
 
-	if (cache_mode < CACHE_MODE_FORCE_RELOAD && ce && !list_empty(ce->frag)
-	    && !((struct fragment *) ce->frag.next)->offset)
-		conn->from = ((struct fragment *) ce->frag.next)->length;
+	if (cache_mode < CACHE_MODE_FORCE_RELOAD && cached && !list_empty(cached->frag)
+	    && !((struct fragment *) cached->frag.next)->offset)
+		conn->from = ((struct fragment *) cached->frag.next)->length;
 
 	if (download) {
-		download->prg = &conn->prg;
+		download->progress = &conn->progress;
 		download->conn = conn;
-		download->ce = NULL;
+		download->cached = NULL;
 		add_to_list(conn->downloads, download);
 	}
 
@@ -877,7 +937,7 @@ load_url(unsigned char *url, unsigned char *ref_url, struct download *download,
 
 	check_queue_bugs();
 
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half((void (*)(void *)) check_queue, NULL);
 	return 0;
 }
 
@@ -894,10 +954,10 @@ change_connection(struct download *old, struct download *new,
 
 	if (is_in_result_state(old->state)) {
 		if (new) {
-			new->ce = old->ce;
+			new->cached = old->cached;
 			new->state = old->state;
 			new->prev_error = old->prev_error;
-			if (new->end) new->end(new, new->data);
+			if (new->callback) new->callback(new, new->data);
 		}
 		return;
 	}
@@ -915,13 +975,13 @@ change_connection(struct download *old, struct download *new,
 	old->state = S_INTERRUPTED;
 
 	if (new) {
-		new->prg = &conn->prg;
+		new->progress = &conn->progress;
 		add_to_list(conn->downloads, new);
 		new->state = conn->state;
 		new->prev_error = conn->prev_error;
 		new->pri = newpri;
 		new->conn = conn;
-		new->ce = conn->cache;
+		new->cached = conn->cached;
 
 	} else if (conn->detached || interrupt) {
 		abort_conn_with_state(conn, S_INTERRUPTED);
@@ -930,7 +990,7 @@ change_connection(struct download *old, struct download *new,
 	sort_queue();
 	check_queue_bugs();
 
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half((void (*)(void *)) check_queue, NULL);
 }
 
 /* This will remove 'pos' bytes from the start of the cache for the specified
@@ -946,7 +1006,7 @@ detach_connection(struct download *download, int pos)
 		int total_len;
 		int i, total_pri = 0;
 
-		if (!conn->cache)
+		if (!conn->cached)
 			return;
 
 		total_len = (conn->est_length == -1) ? conn->from
@@ -967,7 +1027,7 @@ detach_connection(struct download *download, int pos)
 		/* Pre-clean cache. */
 		shrink_format_cache(0);
 
-		if (total_pri != 1 || is_object_used(conn->cache)) {
+		if (total_pri != 1 || is_object_used(conn->cached)) {
 			/* We're too important, or someone uses our cache
 			 * entry. */
 			return;
@@ -976,12 +1036,12 @@ detach_connection(struct download *download, int pos)
 		/* DBG("detached"); */
 
 		/* We aren't valid cache entry anymore. */
-		conn->cache->valid = 0;
+		conn->cached->valid = 0;
 		conn->detached = 1;
 	}
 
 	/* Strip the entry. */
-	free_entry_to(conn->cache, pos);
+	free_entry_to(conn->cached, pos);
 }
 
 static void
@@ -1017,7 +1077,7 @@ set_connection_timeout(struct connection *conn)
 	conn->timer = install_timer((conn->unrestartable
 				     ? get_opt_int("connection.unrestartable_receive_timeout")
 				     : get_opt_int("connection.receive_timeout"))
-				    * 500, (void (*)(void *))connection_timeout_1, conn);
+				    * 500, (void (*)(void *)) connection_timeout_1, conn);
 }
 
 
@@ -1025,8 +1085,7 @@ void
 abort_all_connections(void)
 {
 	while (!list_empty(queue)) {
-		set_connection_state(queue.next, S_INTERRUPTED);
-		abort_connection(queue.next);
+		abort_conn_with_state(queue.next, S_INTERRUPTED);
 	}
 
 	abort_all_keepalive_connections();
