@@ -1,5 +1,5 @@
 /* Locations handling */
-/* $Id: location.c,v 1.8 2003/10/29 15:25:38 zas Exp $ */
+/* $Id: location.c,v 1.13.6.1 2005/04/05 21:08:43 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -22,8 +22,7 @@ copy_location(struct location *dst, struct location *src)
 
 	init_list(dst->frames);
 	foreachback (frame, src->frames) {
-		/* One byte is reserved in struct vs. */
-		new_frame = mem_alloc(sizeof(struct frame) + frame->vs.url_len);
+		new_frame = mem_calloc(1, sizeof(*new_frame));
 		if (new_frame) {
 			new_frame->name = stracpy(frame->name);
 			if (!new_frame->name) {
@@ -44,10 +43,10 @@ destroy_location(struct location *loc)
 	struct frame *frame;
 
 	foreach (frame, loc->frames) {
-		destroy_vs(&frame->vs);
+		destroy_vs(&frame->vs, 1);
 		mem_free(frame->name);
 	}
 	free_list(loc->frames);
-	destroy_vs(&loc->vs);
+	destroy_vs(&loc->vs, 1);
 	mem_free(loc);
 }
