@@ -1,12 +1,10 @@
 /* Lua interface (scripting engine) */
-/* $Id: core.c,v 1.139 2003/12/21 14:13:20 zas Exp $ */
+/* $Id: core.c,v 1.173.4.9 2005/04/06 11:12:56 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#ifdef HAVE_LUA
-
 #include <setjmp.h>
 #include <signal.h>
 #include <stdio.h>
@@ -20,19 +18,16 @@
 
 #include "elinks.h"
 
-#include "bfu/button.h"
 #include "bfu/dialog.h"
-#include "bfu/inpfield.h"
-#include "bfu/inphist.h"
-#include "bfu/msgbox.h"
-#include "bfu/style.h"
-#include "bfu/text.h"
-#include "config/kbdbind.h"
 #include "cache/cache.h"
+#include "config/kbdbind.h"
+#include "config/options.h"
+#include "config/opttypes.h"
 #include "document/document.h"
 #include "document/renderer.h"
 #include "document/view.h"
 #include "intl/gettext/libintl.h"
+#include "intl/charsets.h"
 #include "lowlevel/home.h"
 #include "lowlevel/signals.h"
 #include "modules/module.h"
@@ -45,6 +40,7 @@
 #include "scripting/scripting.h"
 #include "terminal/terminal.h"
 #include "util/color.h"
+#include "util/file.h"
 #include "util/memory.h"
 #include "util/string.h"
 #include "viewer/dump/dump.h"
@@ -63,8 +59,6 @@ static sigjmp_buf errjmp;
 #define L	lua_state
 #define LS	lua_State *S
 
-typedef	unsigned char uchar;
-
 static void handle_standard_lua_returns(unsigned char *from);
 
 
@@ -75,7 +69,7 @@ static void handle_standard_lua_returns(unsigned char *from);
 static int
 l_alert(LS)
 {
-	alert_lua_error((uchar *) lua_tostring(S, 1));
+	alert_lua_error((unsigned char *) lua_tostring(S, 1));
 	return 0;
 }
 
@@ -84,7 +78,7 @@ l_current_url(LS)
 {
 	if (lua_ses && have_location(lua_ses)) {
 		struct view_state *vs = &cur_loc(lua_ses)->vs;
-		unsigned char *url = get_no_post_url(vs->url, NULL);
+		unsigned char *url = get_uri_string(vs->uri, URI_ORIGINAL);
 
 		if (url) {
 			lua_pushstring(S, url);
@@ -100,7 +94,7 @@ l_current_url(LS)
 static int
 l_current_link(LS)
 {
-	struct link *link = get_current_link(lua_ses);
+	struct link *link = get_current_session_link(lua_ses);
 
 	if (link) {
 		lua_pushstring(S, link->where);
@@ -120,7 +114,7 @@ l_current_title(LS)
 		unsigned char *clean_title = stracpy(doc_view->document->title);
 
 		if (clean_title) {
-			register int i = 0;
+			int i = 0;
 
 			while (clean_title[i]) {
 				if (clean_title[i] < ' '
@@ -143,9 +137,8 @@ static int
 l_current_document(LS)
 {
 	if (lua_ses) {
-		unsigned char *url = cur_loc(lua_ses)->vs.url;
-		struct cache_entry *ce = url ? find_in_cache(url) : NULL;
-		struct fragment *f = ce ? ce->frag.next : NULL;
+		struct cache_entry *cached = find_in_cache(cur_loc(lua_ses)->vs.uri);
+		struct fragment *f = cached ? cached->frag.next : NULL;
 
 		if (f && f->length) {
 			lua_pushlstring(S, f->data, f->length);
@@ -173,7 +166,7 @@ l_current_document_formatted(LS)
 	if (width > 0) {
 		old_width = lua_ses->tab->term->width;
 		lua_ses->tab->term->width = width;
-		render_document_frames(lua_ses);
+		render_document_frames(lua_ses, 0);
 	}
 
 	if (init_string(&buffer)) {
@@ -184,7 +177,7 @@ l_current_document_formatted(LS)
 
 	if (width > 0) {
 		lua_ses->tab->term->width = old_width;
-		render_document_frames(lua_ses);
+		render_document_frames(lua_ses, 0);
 	}
 	return 1;
 
@@ -207,7 +200,7 @@ l_pipe_read(LS)
 
 	while (!feof(fp)) {
 		unsigned char buf[1024];
-		int l = fread(buf, 1, sizeof buf, fp);
+		int l = fread(buf, 1, sizeof(buf), fp);
 
 		if (l > 0) {
 			s = mem_realloc(s, len + l);
@@ -222,11 +215,11 @@ l_pipe_read(LS)
 	pclose(fp);
 
 	lua_pushlstring(S, s, len);
-	if (s) mem_free(s);
+	mem_free_if(s);
 	return 1;
 
 lua_error:
-	if (s) mem_free(s);
+	mem_free_if(s);
 	lua_pushnil(S);
 	return 1;
 }
@@ -235,7 +228,7 @@ static int
 l_execute(LS)
 {
 	if (lua_isstring(S, 1)) {
-		exec_on_terminal(lua_ses->tab->term, (uchar *)lua_tostring(S, 1), "", 0);
+		exec_on_terminal(lua_ses->tab->term, (unsigned char *) lua_tostring(S, 1), "", 0);
 		lua_pushnumber(S, 0);
 		return 1;
 	}
@@ -279,21 +272,21 @@ static enum evhook_status
 run_lua_func(va_list ap, void *data)
 {
 	struct session *ses = va_arg(ap, struct session *);
-	int func_ref = (int)data;
+	int func_ref = (long) data;
 	int err;
 
 	if (func_ref == LUA_NOREF) {
 		alert_lua_error("key bound to nothing (internal error)");
-		return EHS_NEXT;
+		return EVENT_HOOK_STATUS_NEXT;
 	}
 
 	lua_getref(L, func_ref);
-	if (prepare_lua(ses)) return EHS_NEXT;
+	if (prepare_lua(ses)) return EVENT_HOOK_STATUS_NEXT;
 	err = lua_call(L, 0, 2);
 	finish_lua();
 	if (!err) handle_standard_lua_returns("keyboard function");
 
-	return EHS_NEXT;
+	return EVENT_HOOK_STATUS_NEXT;
 }
 
 static int
@@ -316,12 +309,12 @@ l_bind_key(LS)
 	ref = lua_ref(S, 1);
 	add_format_to_string(&event_name, "lua-run-func %i", ref);
 	event_id = register_event(event_name.source);
-	event_id = register_event_hook(event_id, run_lua_func, 0, (void *)ref);
+	event_id = register_event_hook(event_id, run_lua_func, 0, (void *) (long) ref);
 	done_string(&event_name);
 	if (event_id == EVENT_NONE) goto lua_error;
 
-	err = bind_scripting_func((uchar *)lua_tostring(S, 1),
-			    (uchar *)lua_tostring(S, 2), event_id);
+	err = bind_scripting_func((unsigned char *) lua_tostring(S, 1),
+			    (unsigned char *) lua_tostring(S, 2), event_id);
 	if (err) {
 		lua_unref(S, ref);
 		alert_lua_error2("error in bind_key: ", err);
@@ -349,8 +342,9 @@ struct lua_dlg_data {
 };
 
 static void
-dialog_run_lua(struct lua_dlg_data *data)
+dialog_run_lua(void *data_)
 {
+	struct lua_dlg_data *data = data_;
 	lua_State *s = data->state;
 	int err;
 
@@ -379,14 +373,17 @@ l_edit_bookmark_dialog(LS)
 	}
 
 #define L_EDIT_BMK_WIDGETS_COUNT 5
-	dlg = calloc_dialog(L_EDIT_BMK_WIDGETS_COUNT, sizeof(struct lua_dlg_data));
+	dlg = calloc_dialog(L_EDIT_BMK_WIDGETS_COUNT, sizeof(*data));
 	if (!dlg) return 0;
 
 	data = (struct lua_dlg_data *) get_dialog_offset(dlg, L_EDIT_BMK_WIDGETS_COUNT);
 	data->state = S;
-	safe_strncpy(data->cat, (uchar *)lua_tostring(S, 1), MAX_STR_LEN-1);
-	safe_strncpy(data->name, (uchar *)lua_tostring(S, 2), MAX_STR_LEN-1);
-	safe_strncpy(data->url, (uchar *)lua_tostring(S, 3), MAX_STR_LEN-1);
+	safe_strncpy(data->cat, (unsigned char *) lua_tostring(S, 1),
+		     MAX_STR_LEN-1);
+	safe_strncpy(data->name, (unsigned char *) lua_tostring(S, 2),
+		     MAX_STR_LEN-1);
+	safe_strncpy(data->url, (unsigned char *) lua_tostring(S, 3),
+		     MAX_STR_LEN-1);
 	lua_pushvalue(S, 4);
 	data->func_ref = lua_ref(S, 1);
 
@@ -398,8 +395,8 @@ l_edit_bookmark_dialog(LS)
 	add_dlg_field(dlg, _("Name", term), 0, 0, NULL, MAX_STR_LEN, data->name, NULL);
 	add_dlg_field(dlg, _("URL", term), 0, 0, NULL, MAX_STR_LEN, data->url, NULL);
 
-	add_dlg_ok_button(dlg, B_ENTER, _("OK", lua_ses->tab->term), dialog_run_lua, data);
-	add_dlg_button(dlg, B_ESC, cancel_dialog, _("Cancel", lua_ses->tab->term), NULL);
+	add_dlg_ok_button(dlg, _("~OK", term), B_ENTER, dialog_run_lua, data);
+	add_dlg_button(dlg, _("~Cancel", term), B_ESC, cancel_dialog, NULL);
 
 	add_dlg_end(dlg, L_EDIT_BMK_WIDGETS_COUNT);
 
@@ -425,8 +422,9 @@ struct lua_xdialog_data {
 };
 
 static void
-xdialog_run_lua(struct lua_xdialog_data *data)
+xdialog_run_lua(void *data_)
 {
+	struct lua_xdialog_data *data = data_;
 	lua_State *s = data->state;
 	int err;
 	int i;
@@ -457,14 +455,15 @@ l_xdialog(LS)
 	for (i = 1; i < nargs; i++) if (!lua_isstring(S, i)) goto lua_error;
 	if (!lua_isfunction(S, nargs)) goto lua_error;
 
-	dlg = calloc_dialog(nitems, sizeof(struct lua_xdialog_data));
+	dlg = calloc_dialog(nitems, sizeof(*data));
 	if (!dlg) return 0;
 
 	data = (struct lua_xdialog_data *) get_dialog_offset(dlg, nitems);
 	data->state = S;
 	data->nfields = nfields;
 	for (i = 0; i < nfields; i++)
-		safe_strncpy(data->fields[i], (uchar *)lua_tostring(S, i+1),
+		safe_strncpy(data->fields[i],
+			     (unsigned char *) lua_tostring(S, i+1),
 			     MAX_STR_LEN-1);
 	lua_pushvalue(S, nargs);
 	data->func_ref = lua_ref(S, 1);
@@ -473,14 +472,12 @@ l_xdialog(LS)
 	dlg->layouter = generic_dialog_layouter;
 	dlg->layout.maximize_width = 1;
 
-	i = 0;
-	while (dlg->widgets_size < nfields) {
+	for (i = 0; i < nfields; i++)
 		add_dlg_field(dlg, _("Name", term), 0, 0, NULL, MAX_STR_LEN,
-			      data->fields[i++], NULL);
-	}
+			      data->fields[i], NULL);
 
-	add_dlg_ok_button(dlg, B_ENTER, _("OK", term), xdialog_run_lua, data);
-	add_dlg_button(dlg, B_ESC, cancel_dialog, _("Cancel", term), NULL);
+	add_dlg_ok_button(dlg, _("~OK", term), B_ENTER, xdialog_run_lua, data);
+	add_dlg_button(dlg, _("~Cancel", term), B_ESC, cancel_dialog, NULL);
 
 	add_dlg_end(dlg, nitems);
 
@@ -496,6 +493,135 @@ lua_error:
 
 /* End xdialog bit. */
 
+/* Set/get option */
+
+static int
+l_set_option(LS)
+{
+	int nargs;
+	struct option *opt, *current;
+	const char *name;
+
+	nargs = lua_gettop(S);
+	if (nargs != 2)
+		goto lua_error;
+
+	/* Get option record */
+	name = lua_tostring(S, 1);
+	opt = get_opt_rec(config_options, (unsigned char *) name);
+	if (opt == NULL)
+		goto lua_error;
+
+	/* Set option */
+	switch (opt->type) {
+	case OPT_BOOL:
+	case OPT_INT:
+	case OPT_LONG:
+	{
+		int value;
+
+		value = lua_tonumber(S, 2);
+		option_types[opt->type].set(opt, (unsigned char *) (&value));
+		break;
+	}
+	case OPT_STRING:
+	case OPT_CODEPAGE:
+	case OPT_LANGUAGE:
+	case OPT_COLOR:
+		option_types[opt->type].set(opt, (unsigned char *) lua_tostring(S, 2));
+		break;
+	default:
+		goto lua_error;
+	}
+
+	/* Call hook */
+	current = opt;
+	while (current && (!current->change_hook ||
+		!current->change_hook(lua_ses, current, opt))) {
+		if (current->root)
+			current = current->root;
+		else
+			break;
+	}
+	return 1;
+
+lua_error:
+	lua_pushnil(S);
+	return 1;
+}
+
+static int
+l_get_option(LS)
+{
+	int nargs;
+	struct option *opt;
+	const char *name;
+
+	/* Get option record */
+	nargs = lua_gettop(S);
+	if (nargs != 1)
+		goto lua_error;
+	name = lua_tostring(S, 1);
+	opt = get_opt_rec(config_options, (unsigned char *) name);
+	if (opt == NULL)
+		goto lua_error;
+
+	/* Convert to an appropriate Lua type */
+	switch (opt->type) {
+	case OPT_BOOL:
+		lua_pushboolean(S, opt->value.number);
+		break;
+	case OPT_INT:
+	case OPT_LONG:
+		lua_pushnumber(S, opt->value.number);
+		break;
+	case OPT_STRING:
+		lua_pushstring(S, opt->value.string);
+		break;
+	case OPT_CODEPAGE:
+	{
+		unsigned char *cp_name;
+
+		cp_name = get_cp_mime_name(opt->value.number);
+		lua_pushstring(S, cp_name);
+		break;
+	}
+	case OPT_LANGUAGE:
+	{
+		unsigned char *lang;
+
+#ifdef ENABLE_NLS
+		lang = language_to_name(current_language);
+#else
+		lang = "System";
+#endif
+		lua_pushstring(S, lang);
+		break;
+	}
+	case OPT_COLOR:
+	{
+		color_t color;
+		unsigned char hexcolor[8];
+		unsigned char *strcolor;
+
+		color = opt->value.color;
+		strcolor = get_color_string(color, hexcolor);
+		lua_pushstring(S, strcolor);
+		break;
+	}
+	case OPT_COMMAND:
+	default:
+		goto lua_error;
+	}
+
+	return 1;
+
+lua_error:
+	lua_pushnil(S);
+	return 1;
+}
+
+/* End of set/get option */
 
 /* Initialisation */
 
@@ -503,11 +629,20 @@ static void
 do_hooks_file(LS, unsigned char *prefix, unsigned char *filename)
 {
 	unsigned char *file = straconcat(prefix, "/", filename, NULL);
-	int oldtop = lua_gettop(S);
 
-	lua_dofile(S, file);
+	if (!file) return;
+
+	/* Test file existence to avoid Lua error reporting (under version 5.x)
+	 * Fixes debian bug #231760 ('dbug 231760' using URI rewrite) */
+	if (file_can_read(file)) {
+		int oldtop = lua_gettop(S);
+
+		if (lua_dofile(S, file) != 0)
+			sleep(3); /* Let some time to see error messages. */
+		lua_settop(S, oldtop);
+	}
+
 	mem_free(file);
-	lua_settop(S, oldtop);
 }
 
 static void
@@ -515,6 +650,7 @@ init_lua(struct module *module)
 {
 	L = lua_open(0);
 	lua_baselibopen(L);
+	lua_mathlibopen(L);
 	lua_strlibopen(L);
 	lua_register(L, LUA_ALERT, l_alert);
 	lua_register(L, "current_url", l_current_url);
@@ -526,10 +662,13 @@ init_lua(struct module *module)
 	lua_register(L, "bind_key", l_bind_key);
 	lua_register(L, "edit_bookmark_dialog", l_edit_bookmark_dialog);
 	lua_register(L, "xdialog", l_xdialog);
+	lua_register(L, "set_option", l_set_option);
+	lua_register(L, "get_option", l_get_option);
 
 	lua_dostring(L, "function set_elinks_home(s) elinks_home = s end");
 	lua_getglobal(L, "set_elinks_home");
-	lua_pushstring(L, elinks_home ? elinks_home : (unsigned char *)CONFDIR);
+	lua_pushstring(L, elinks_home ? elinks_home
+				      : (unsigned char *) CONFDIR);
 	lua_call(L, 1, 0);
 
 	do_hooks_file(L, CONFDIR, LUA_HOOKS_FILENAME);
@@ -562,19 +701,16 @@ prepare_lua(struct session *ses)
 	lua_ses = ses;
 	errterm = lua_ses ? lua_ses->tab->term : NULL;
 	/* XXX this uses the wrong term, I think */
-	install_signal_handler(SIGINT, (void (*)(void *))handle_sigint, NULL, 1);
+	install_signal_handler(SIGINT, (void (*)(void *)) handle_sigint, NULL, 1);
 
 	return sigsetjmp(errjmp, 1);
 }
 
-void
-sig_ctrl_c(struct terminal *t);
-
 void
 finish_lua(void)
 {
 	/* XXX should save previous handler instead of assuming this one */
-	install_signal_handler(SIGINT, (void (*)(void *))sig_ctrl_c, errterm, 0);
+	install_signal_handler(SIGINT, (void (*)(void *)) sig_ctrl_c, errterm, 0);
 }
 
 
@@ -584,15 +720,13 @@ void
 alert_lua_error(unsigned char *msg)
 {
 	if (errterm) {
-		msg_box(errterm, NULL, MSGBOX_NO_INTL,
-			_("Lua Error", errterm), AL_LEFT,
-			msg,
-			NULL, 1,
-			_("OK", errterm), NULL, B_ENTER | B_ESC);
+		info_box(errterm, MSGBOX_NO_TEXT_INTL,
+			N_("Lua Error"), ALIGN_LEFT,
+			msg);
 		return;
 	}
 
-	ERROR("Lua: %s", msg);
+	usrerror("Lua: %s", msg);
 	sleep(3);
 }
 
@@ -634,7 +768,7 @@ handle_ret_eval(struct session *ses)
 static void
 handle_ret_run(struct session *ses)
 {
-	unsigned char *cmd = (uchar *)lua_tostring(L, -1);
+	unsigned char *cmd = (unsigned char *) lua_tostring(L, -1);
 
 	if (cmd) {
 		exec_on_terminal(ses->tab->term, cmd, "", 1);
@@ -647,10 +781,10 @@ handle_ret_run(struct session *ses)
 static void
 handle_ret_goto_url(struct session *ses)
 {
-	unsigned char *url = (uchar *)lua_tostring(L, -1);
+	unsigned char *url = (unsigned char *) lua_tostring(L, -1);
 
 	if (url) {
-		goto_url(ses, url);
+		goto_url_with_hook(ses, url);
 		return;
 	}
 
@@ -681,12 +815,7 @@ handle_standard_lua_returns(unsigned char *from)
 
 /* Console stuff. */
 
-static struct input_history lua_console_history = {
-	/* items: */	{ D_LIST_HEAD(lua_console_history.entries) },
-	/* size: */	0,
-	/* dirty: */	0,
-	/* nosave: */	0,
-};
+static INIT_INPUT_HISTORY(lua_console_history);
 
 static void
 lua_console(struct session *ses, unsigned char *expr)
@@ -712,15 +841,15 @@ dialog_lua_console(va_list ap, void *data)
 {
 	struct session *ses = va_arg(ap, struct session *);
 
-	if (get_opt_int_tree(cmdline_options, "anonymous"))
-		return EHS_NEXT;
+	if (get_cmd_opt_bool("anonymous"))
+		return EVENT_HOOK_STATUS_NEXT;
 
-	input_field(ses->tab->term, NULL, 1,
-		    N_("Lua Console"), N_("Enter expression"),
-		    N_("OK"), N_("Cancel"), ses, &lua_console_history,
-		    MAX_STR_LEN, "", 0, 0, NULL,
-		    (void (*)(void *, unsigned char *)) lua_console, NULL);
-	return EHS_NEXT;
+	input_dialog(ses->tab->term, NULL,
+		     N_("Lua Console"), N_("Enter expression"),
+		     ses, &lua_console_history,
+		     MAX_STR_LEN, "", 0, 0, NULL,
+		     (void (*)(void *, unsigned char *)) lua_console, NULL);
+	return EVENT_HOOK_STATUS_NEXT;
 }
 
 static void
@@ -733,7 +862,7 @@ enum evhook_status
 free_lua_console_history(va_list ap, void *data)
 {
 	free_lua_console_history_entries();
-	return EHS_NEXT;
+	return EVENT_HOOK_STATUS_NEXT;
 }
 
 
@@ -746,5 +875,3 @@ struct module lua_scripting_module = struct_module(
 	/* init: */		init_lua,
 	/* done: */		cleanup_lua
 );
-
-#endif
