@@ -1,12 +1,10 @@
 /* General scripting system functionality */
-/* $Id: scripting.c,v 1.9 2003/10/27 01:12:15 jonas Exp $ */
+/* $Id: scripting.c,v 1.12.6.3 2005/01/29 02:06:28 jonas Exp $ */
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#ifdef HAVE_SCRIPTING
-
 #include "elinks.h"
 
 #include "intl/gettext/libintl.h"
@@ -18,14 +16,22 @@
 /* Backends dynamic area: */
 
 #include "scripting/guile/core.h"
-#include "scripting/lua/core.h"
+#include "scripting/lua/lua.h"
+#include "scripting/perl/perl.h"
+#include "scripting/ruby/ruby.h"
 
 static struct module *scripting_modules[] = {
-#ifdef HAVE_LUA
+#ifdef CONFIG_LUA
 	&lua_scripting_module,
 #endif
-#ifdef HAVE_GUILE
+#ifdef CONFIG_GUILE
 	&guile_scripting_module,
+#endif
+#ifdef CONFIG_PERL
+	&perl_scripting_module,
+#endif
+#ifdef CONFIG_RUBY
+	&ruby_scripting_module,
 #endif
 	NULL,
 };
@@ -39,5 +45,3 @@ struct module scripting_module = struct_module(
 	/* init: */		NULL,
 	/* done: */		NULL
 );
-
-#endif
