@@ -1,13 +1,18 @@
-/* $Id: memdebug.h,v 1.3 2003/05/08 23:03:08 zas Exp $ */
+/* $Id: memdebug.h,v 1.5 2004/10/13 15:34:47 zas Exp $ */
 
 #ifndef EL__UTIL_MEMDEBUG_H
 #define EL__UTIL_MEMDEBUG_H
 
-#ifdef LEAK_DEBUG
+#ifdef DEBUG_MEMLEAK
 
 /* TODO: Another file? */
 
-extern long mem_amount;
+struct mem_stats {
+	long true_amount;
+	long amount;
+};
+
+extern struct mem_stats mem_stats;
 
 void *debug_mem_alloc(unsigned char *, int, size_t);
 void *debug_mem_calloc(unsigned char *, int, size_t, size_t);
@@ -19,6 +24,6 @@ void check_memory_leaks(void);
 
 #else
 #define set_mem_comment(p, c, l)
-#endif /* LEAK_DEBUG */
+#endif /* DEBUG_MEMLEAK */
 
 #endif
